%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Copyright (c) 2000 by Erik Johansson.  All Rights Reserved 
%% ====================================================================
%%  Filename : 	hipe.hrl
%%  Purpose  :  To define some usefull macros for debugging
%%               and error reports. 
%%  Notes    :  
%%              
%%              
%%              
%%  History  :	* 2000-11-03 Erik Johansson (happi@csd.uu.se): 
%%               Created.
%%  CVS      :
%%              $Author: happi $
%%              $Date: 2001/08/03 10:37:59 $
%%              $Revision: 1.4 $
%% ====================================================================
%%
%% Functions:
%%  version/0   - returns the version number of HiPE as a tuple.
%%  msg/2       - Works like io:format but prepends 
%%                ?MSGTAG to the message.
%%                If LOGGING is defined then error_logger is used.
%%
%% Defines:
%%  warning_msg   - Prints a tagged warning.
%%  error_msg     - Logs a tagged error.
%%  debug_msg     - Prints a tagged msg if DEBUG is defined.
%%  IF_DEBUG(A,B) - Excutes A if DEBUG is defined B otherwise.    
%%  EXIT          - Exits with added module and line info.
%%  ASSERT        - Exits if the expresion does not evaluate to true.
%%  VERBOSE_ASSSERT - A message is printed even when an asertion is true.
%%  TIME_STMNT(Stmnt, String, FreeVar) - Times the statemnet Stmnt
%%                   if TIMING is on. The execution time is bound to
%%                   FreeVar. String is printed after the execution
%%                   followed by the excution time in s and a newline.
%%
%%
%% Flags:
%%  DEBUG           - Turns on debugging.
%%  VERBOSE         - More is printed...
%%  LOGGING         - Messages are logged with erl_logger.
%%  DO_ASSERT       - Assertions are turned on.
%%  TIMING          - Turn on timing.
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-define(version(),{0,99,0}).
-define(msgtagmap(M), 
	case M of 
	  hipe -> "Compiler ";
	  hipe_main -> "Compiler ";
	  hipe_update_catches -> "Compiler ";
	  hipe_sparc_loader -> "Loader   ";
	  hipe_beam_to_icode -> "BEAM Dis.";
	  _ ->  atom_to_list(M)
	end).

-define(MSGTAG, 
	(fun () ->
	  
	  {Major, Minor, Increment} = ?version(),
	  "[HiPE " ++ ?msgtagmap(?MODULE) ++
	    "(v " ++ integer_to_list(Major) ++ "."
	    ++ integer_to_list(Minor) ++ "."
	    ++ integer_to_list(Increment) ++ ")] "
	 end)()).

-ifdef(LOGGING).
-define(msg(Msg, Args),
   error_logger:info_msg(?MSGTAG ++ Msg, Args)).
-define(untagged_msg(Msg, Args),
	error_logger:info_msg(Msg, Args)).
-else.
-define(msg(Msg, Args),
   io:format(?MSGTAG ++ Msg, Args)).
-define(untagged_msg(Msg, Args),
   io:format(Msg, Args)).
-endif.


-ifdef(DEBUG).
-define(debug_msg(Msg,Data), ?msg(Msg,Data)).
-define(debug_untagged_msg(Msg,Data), ?untagged_msg(Msg,Data)).
-define(IF_DEBUG(DebugAction,NoDebugAction),DebugAction).
-define(IF_DEBUG_LEVEL(Level,DebugAction,NoDebugAction),
	if(Level =< ?DEBUG) -> DebugAction; true -> NoDebugAction end ).

-else.
-define(debug_msg(Msg,Data), no_debug).
-define(debug_untagged_msg(Msg,Data), no_debug).
-define(IF_DEBUG(DebugAction,NoDebugAction),NoDebugAction).
-define(IF_DEBUG_LEVEL(Level,DebugAction,NoDebugAction),NoDebugAction).
-endif.




-ifdef(VERBOSE).
-define(EXIT(Reason),exit({?MODULE,?LINE,Reason})).
-else.
-define(EXIT(Reason),
 	?msg("EXITED with reason ~w @~w:~w\n",
 	    [Reason,?MODULE,?LINE]),
 	exit({?MODULE,?LINE,Reason})).
-endif.


-ifdef(TIMING).
-define(TIME_STMNT(STMNT,Msg,Timer),
	Timer = hipe_timing:start_timer(),
	STMNT,
	?msg(Msg ++ "~.2f s\n",[hipe_timing:stop_timer(Timer)/1000])).
-else.
-define(TIME_STMNT(STMNT,Msg,Timer),STMNT).
-endif.

-ifdef(DO_ASSERT).
-define(VERBOSE_ASSERT(X),
	case X of 
	  true ->
	    io:format("Assertion ok ~w ~w\n",[?MODULE,?LINE]),
	    true;
	  __ASSVAL_R -> 
	    io:format("Assertion failed ~w ~w: ~p\n",
		      [?MODULE,?LINE, __ASSVAL_R]),
	    ?EXIT(assertion_failed)
	end).
-define(ASSERT(X),
	case X of 
	  true -> true;
	  _ -> ?EXIT(assertion_failed)
	end).
-else.
-define(ASSERT(X),true).
-define(VERBOSE_ASSERT(X),true).
-endif.




-define(error_msg(Msg, Args),
	error_logger:error_msg(
	  ?MSGTAG ++
	  "ERROR:~s:~w: " ++ Msg,
	  [?FILE,?LINE|Args])).

-define(warning_msg(Msg, Args),
	?msg("WARNING:~s:~w: " ++ Msg, [?FILE,?LINE|Args])).


% Use this to display info, save stuff and so on.
% Vars cannot be exported from Action
-define(when_option(__Opt,__Opts,__Action),
	case property_lists:get_bool(__Opt,__Opts) of
	    true -> __Action; false -> ok end).

-define(start_timer(Text),
	fun(__Timers) ->
	    fun(__Space) ->
		fun( {__Total,__Last}) ->
		    put(hipe_timers,[__Total|__Timers]),
		    ?msg("[@~6w]"++__Space ++ "> ~s~n",[__Total,Text])
		end (erlang:statistics(runtime))
	    end (lists:duplicate(length(__Timers),$|))
	end (case get(hipe_timers) of
	       undefined -> [];
	       _ -> get(hipe_timers)
	     end)).

-define(stop_timer(Text),
	fun([StartTime|Timers]) -> 
	    fun(Space) ->
		fun( {__Total,__Last}) ->
		    put(hipe_timers,Timers),
		    ?msg("[@~6w]"++Space ++ "< ~s: ~w~n",
			 [__Total, Text, __Total-StartTime])
		end(erlang:statistics(runtime))
	    end(lists:duplicate(length(Timers),$|));
	   (_) ->
	    fun( {__Total,__Last}) ->
		put(hipe_timers, []),
		?msg("[@~6w]< ~s: ~w~n",
		     [__Total, Text, __Total])
	    end(erlang:statistics(runtime))
	
	end(get(hipe_timers))).

-define(option_time(Stmnt, Text, Options),
	?when_option(time, Options, ?start_timer(Text)),
	fun(R) ->
	    ?when_option(time, Options, ?stop_timer(Text)),
	    R
	end(Stmnt)).

-define(option_start_time(Text,Options),
	?when_option(time, Options, ?start_timer(Text))).

-define(option_stop_time(Text,Options),
	?when_option(time, Options, ?stop_timer(Text))).
		     
