%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(dets).

%% Disk based linear hashing lookup dictionary.

%% Public.
-export([all/0,
         close/1,
         delete/2,
         delete_all_objects/1,
         delete_object/2,
         first/1,
         foldl/3,
         foldr/3,
         from_ets/2,
         info/1,
         info/2,
         init_table/2,
         insert/2,
         lookup/2,
         match/1,
         match/2,
         match/3,
         match_delete/2,
         match_object/1,
         match_object/2,
         match_object/3,
         member/2,
         next/2,
         open_file/1,
         open_file/2,
         pid2name/1,
         safe_fixtable/2,
         select/1,
         select/2,
         select/3,
         select_delete/2,
         slot/2,
         sync/1,
         to_ets/2,
         traverse/2,
         update_counter/3]).

%% Server export.
-export([stop/0]).

%% To be used by mnesia only.
-export([do_match/2, fixtable/2]).

%% Internal.
-export([do_open_file/2, do_open_file/11]).

%% Debug.
-export([file_info/1,
	 fsck/1,
	 get_head_field/2,
	 inspect_chain/2,
	 view/1,
	 where/2,
	 verbose/0,
	 verbose/1
	]).

%% Not documented, or not ready for publication.
-export([lookup_keys/2]).

-compile({inline, [{sz2pos,1},{to_object,2}]}).
-compile({inline, [{skip_bytes,8}, {make_object,4}, {get_segp,1}]}).
-compile({inline, [{wl_lookup,5}]}).
-compile({inline, [{einval,2},{badarg,2}]}).

-compile({inline, [{bplus_mk_leaf,1}, {bplus_get_size,1},
		   {bplus_get_tree,2}, {bplus_get_lkey,2},
		   {bplus_get_rkey,2}]}).

-include_lib("kernel/include/file.hrl").

-define(SERVER_NAME, dets).

%%  This is the implementation of the mnesia file storage. Each (non
%%  ram-copy) table is maintained in a corresponding .DAT file. The
%%  dat file is organized as a segmented linear hashlist. The head of
%%  the file with the split indicator, size etc is held in ram by the
%%  server at all times.

%%  The layout of the file is :
%%
%%   bytes   decsription
%%  ---------------------- File header
%%    4      FreelistsPointer
%%    4      Cookie
%%    4      ClosedProperly (pos=8)
%%    4      Type (pos=12)
%%    4      Version (pos=16)
%%    4      M
%%    4      Next
%%    4      KeyPos
%%    4      NoObjects
%%    4      N
%%  ------------------ end of file header
%%    4*8192 SegmentArray
%%  ------------------
%%    4*256  First segment
%%  ----------------------------- This is BASE.
%%    ???    Objects (free and alive)
%%    4*256  Second segment (2 kB now, due to a bug)
%%    ???    Objects (free and alive)
%%    ... more objects and segments ...
%%  -----------------------------
%%    ???    Free lists
%%  -----------------------------
%%    4      File size, in bytes. 

%%  The first slot (0) in the segment array always points to the
%%  pre-allocated first segment.
%%  Before we can find an object we must find the slot where the
%%  object resides. Each slot is a (possibly empty) list (or chain) of
%%  objects that hash to the same slot. If the value stored in the
%%  slot is zero, the slot chain is empty. If the slot value is
%%  non-zero, the value points to a position in the file where the
%%  chain starts. Each object in a chain has the following layout:
%%
%%   bytes  decsription
%%  --------------------
%%    4     Pointer to the next object of the chain.
%%    4     Size of allocated area in bytes (often greater than object size)
%%    4     Status  (FREE or ACTIVE)
%%    ??    Binary representing the object
%%
%%  The status field is used while repairing a file (but not next or size).
%%
%%|---------------|
%%|      head     |
%%|       	  |
%%|               |
%%|_______________|
%%|		  |------|
%%|___seg ptr1____|      |
%%|		  |      |
%%|__ seg ptr 2___|      |
%%|               |      |    segment 1
%%|	....	  |      V _____________
%%			 |		|
%%			 |		|
%%			 |___slot 0 ____|
%%                       |              |
%%                       |___slot 1 ____|-----|
%%			 |		|     |
%%			 |   .....	|     |  1:st obj in slot 1
%%					      V  segment 1
%%						|-----------|
%%						|  next     |
%%						|___________|
%%						|  size     |
%%						|___________|
%%						|  status   |
%%						|___________|
%%						|	    |
%%						|           |
%%						|   obj     |
%%						|           |

%%  The method of hashing is the so called linear hashing algorithm
%%  with segments. 
%%
%%  Linear hashing:
%%
%%         - n indicates next bucket to split (initially zero); 
%%         - m is the size of the hash table 
%%  
%%         - to insert: 
%%                - hash = key mod m 
%%                - if hash < n then hash = key mod 2m 
%%                - when the number of objects exceeds the initial size
%%                  of the hash table, each insertion of an object
%%                  causes bucket n to be split:
%%                      - add a new bucket to the end of the table 
%%                      - redistribute the contents of bucket n 
%%                        using hash = key mod 2m 
%%                      - increment n 
%%                      - if n = m then m = 2m, n = 0 
%%         - to search: 
%%                hash = key mod m 
%%                if hash < n then hash = key mod 2m 
%%                do linear scan of the bucket 
%%  

%%%
%%% File header
%%%

-define(HEADSZ, 40).          % The size of the file header, in bytes.
-define(SEGSZ, 256).          % Size of a segment, in words.
-define(SEGARRSZ, 8192).      % Maximal number of segments.
-define(MAGIC, 16#0abcdef).   % dets cookie, won't ever change.
-define(SEGADDR(SegN), (?HEADSZ + (4 * (SegN)))).
-define(BASE, ?SEGADDR((?SEGSZ + ?SEGARRSZ))).
-define(MAXOBJS, (?SEGSZ * ?SEGARRSZ)). % 2 M objects

%% BIG is used for hashing. BIG must be greater than the maximum
%% number of slots, currently MAXOBJS.
-define(BIG, 16#ffffff).

%% Hard coded positions into the file header:
-define(FREELIST_POS, 0).
-define(CLOSED_PROPERLY_POS, 8).
-define(FILE_FORMAT_VERSION_POS, 16).
-define(D_POS, 20).
-define(NO_OBJECTS_POS, (?D_POS + 12)).

%% Type.
-define(SET, 1).
-define(BAG, 2).
-define(DUPLICATE_BAG, 3).

%% The version of a dets file is indicated by the ClosedProperly
%% field. Version 6 was used in the R1A release, and version 7 in the
%% R1B release up to and including the R3B01 release. Both version 6
%% and version 7 indicate properly closed files by the value
%% CLOSED_PROPERLY.
%%
%% The current version, 8, has three sub-versions:
%%
%% - 8(a), indicated by the value CLOSED_PROPERLY (same as in versions 6 
%%         and 7), introduced in R3B02;
%% - 8(b), indicated by the value CLOSED_PROPERLY2(_NEED_COMPACTING),
%%         introduced in R5A and used up to and including R6A;
%% - 8(c), indicated by the value CLOSED_PROPERLY_NEW_HASH(_NEED_COMPACTING),
%%         in use since R6B.
%%
%% The difference between the 8(a) and the 8(b) versions is the format
%% used for free lists saved on dets files.
%% The 8(c) version uses a different hashing algorithm, erlang:phash
%% (former versions use erlang:hash).
%% Version 8(b) files are only converted to version 8(c) if repair is
%% done, so we need compatability with 8(b) for a _long_ time.

-define(NOT_PROPERLY_CLOSED,0).
-define(CLOSED_PROPERLY,1).
-define(CLOSED_PROPERLY2,2).
-define(CLOSED_PROPERLY2_NEED_COMPACTING,3).
-define(CLOSED_PROPERLY_NEW_HASH,4).
-define(CLOSED_PROPERLY_NEW_HASH_NEED_COMPACTING,5).

-define(FILE_FORMAT_VERSION, 8).
-define(CAN_BUMP_BY_REPAIR, [6, 7]).
-define(CAN_CONVERT_FREELIST, [8]).

%%%
%%% Object header (next, size, status).
%%%

-define(OHDSZ, 12).         % The size of the object header, in bytes.
-define(STATUS_POS, 8).     % Position of the status field.
%% Status values.
-define(FREE, 16#3abcdef).
-define(ACTIVE, 16#12345678).

%% The size of each object is a multiple of 16.
%% BUMP is used when repairing files.
-define(BUMP, 16).
-define(CHUNK_SIZE, 8192).

-define(ReadAhead, 512).

-define(DEFAULT_MIN_NO_SLOTS, 256).
-define(DEFAULT_AUTOSAVE, 3). % minutes
-define(DEFAULT_CACHE, {3000, 2000}). % {delay, size}

%% Definitions for the buddy allocator.
-define(MAXBUD, 32).             % 2 GB is maximum file size
-define(MAXFREELISTS, 50000000). % Bytes reserved for the free lists (at end).

-define(POW(X), (1 bsl (X))).

%% Record holding the file header and more.
-record(head,  {
	  m,               % size
	  next,            % next position for growth (segm mgmt only)
	  fptr,            % the file descriptor
	  no_objects,      % number of objects in table,
	  n,               % split indicator
	  type,            % set | bag | duplicate_bag
	  keypos,          % default is 1 as for ets
	  freelists,       % tuple of free lists of buddies
	                   % if fixed = true, then a pair of freelists
	  freelists_p,     % cached FreelistsPointer
	  auto_save,       % Integer | infinity 
	  update_mode,     % saved | dirty | {error, Reason}
	  fixed = false,   % bool(). Is a fixed table?
	  hash_bif,        % hash bif used for this file (phash or hash)
	  est_no_objects,  % estimated number of objects (default or integer)
	  cache,           % cache(). Write cache.

	  filename,             % name of the file being used
	  access = read_write,  % read | read_write
	  ram_file = false,     % true | false
	  name                  % the name of the table
	 }).

%% If a file error occurs on a working dets file, update_mode is set
%% to the error tuple. When in 'error' mode, the free lists are not
%% written, and a repair is forced next time the file is opened.

-record(fileheader, {
	  freelist,
	  cookie,
	  closed_properly,
	  type,
	  version,
	  m,
	  next,
	  keypos,
	  no_objects,
	  trailer,
	  eof,
	  n
	}).

-record(dets_cont, {
         what,          % object | bindings | select
	 no_objs,       % requested number of objects: default | integer() > 0
	 pos,           % file position, where to read next chunk, or eof
	 bin,           % small chunk not consumed, ends at 'pos'
	 alloc,         % the part of the file not yet scanned, mostly a binary
	 tab,
         arg
	 }).

%% Write Cache.
-record(cache, {
	 tbl,     % ets table holding the write cache
	 wrtime,  % last write or update time
	 delay,   % in milliseconds
	 size     % threshold size of ets table, in words
	 }).

-define(PATTERN_TO_OBJECT_MATCH_SPEC(Pat), [{Pat,[],['$_']}]).
-define(PATTERN_TO_BINDINGS_MATCH_SPEC(Pat), [{Pat,[],['$$']}]).

%%-define(DEBUGM(X, Y), io:format(X, Y)).
-define(DEBUGM(X, Y), true).

%%-define(DEBUGF(X,Y), io:format(X, Y)).
-define(DEBUGF(X,Y), void).

%%-define(PROFILE(C), C).
-define(PROFILE(C), void).

%% Open an already existing file, no arguments
%% -> {ok, head()} | throw(Error)
fopen(Fname) ->
    case file:read_file_info(Fname) of
	{ok, _} ->
	    Tab = make_ref(),
	    Acc = read_write,
	    Ram = false, 
	    {ok, Fd, FH} = read_file_header(Fname, Acc, Ram),
	    case check_file_header(FH, Fd) of
		{error, not_closed} ->
		    io:format(user,"dets: file ~p not properly closed, "
			      "repairing ...~n", [Fname]),
                    case fsck(Fd, Tab, Fname, FH, ?BUMP) of
                        ok ->
                            fopen(Fname);
                        Error ->
                            throw(Error)
                    end;
		{ok, Head, ExtraInfo} ->
		    open_final(Head, Fname, Acc, Ram, ?DEFAULT_CACHE, 
			       Tab, ExtraInfo);
		{error, Reason} ->
		    file:close(Fd),
		    throw({error, {Reason, Fname}})
	    end;
	Error ->
	    file_error(Fname, Error)
    end.

%% Open and possibly create and initialize a file
%% -> {ok, head()} | throw(Error)
fopen(Tab, Fname, Type, Kp, Rep, Est, Ram, CacheSz, Auto, Acc) ->
    case filelib:is_file(Fname) of
	true ->
	    fopen_existing_file(Tab, Fname, Type, Kp, Rep, Est, Ram, CacheSz, 
				Auto, Acc);
	false when Acc == read ->
	    throw({error, {file_error, Fname, enoent}});
	false ->
	    fopen_init_file(Tab, Fname, Type, Kp, Est, Ram, CacheSz, Auto)
    end.

fopen_existing_file(Tab, Fname, Type, Kp, Rep, Est, Ram, CacheSz, Auto, Acc) ->
    {ok, Fd, FH} = read_file_header(Fname, Acc, Ram),
    Do = case check_file_header(FH, Fd) of
	     _ when Rep == force ->
		 M = ", repair forced.",
		 {repair, M, ?BUMP};
	     {ok, _Head, need_compacting} when Rep == true ->
		 %% The file needs to be compacted due
		 %% to a very big and fragmented free_list.
		 M = " is now compacted ...",
		 {repair, M, ?BUMP};
	     {ok, Head, ExtraInfo} ->
		 {final, Head, ExtraInfo};
	     {error, Reason} when Acc == read ->
		 file:close(Fd),
		 throw({error, {Reason, Fname}});
	     {error, not_closed} when Rep == true ->
		 %% Have to repair the file
		 M = " not properly closed, repairing ...",
		 {repair, M, ?BUMP};
	     {error, not_closed} when Rep == false ->
		 file:close(Fd),
		 throw({error, {needs_repair, Fname}});
	     {error, version_bump} when Rep == true ->
		 M = " old version, upgrading ...",
		 {repair, M, 1};
	     {error, Reason} ->
		 file:close(Fd),
		 throw({error, {Reason, Fname}})
	 end,
    case Do of
	_ when FH#fileheader.type =/= Type ->
	    file:close(Fd),
	    throw({error, {type_mismatch, Fname}});
	_ when FH#fileheader.keypos =/= Kp ->
	    file:close(Fd),
	    throw({error, {keypos_mismatch, Fname}});
	{repair, Mess, Bump} ->
	    io:format(user, "dets: file ~p~s~n", [Tab, Mess]),
            case fsck(Fd, Tab, Fname, FH, Bump) of
                ok ->
                    fopen(Tab, Fname, Type, Kp, false, Est, Ram, 
                          CacheSz, Auto, Acc);
                Error ->
                    throw(Error)
            end;
	{final, H, EI} ->
	    H1 = H#head{auto_save = Auto, est_no_objects = Est},
	    open_final(H1, Fname, Acc, Ram, CacheSz, Tab, EI)
    end.

%% -> {ok, head()} | throw(Error)
open_final(Head, Fname, Acc, Ram, CacheSz, Tab, ExtraInfo) ->
    Fd = Head#head.fptr,
    cache_segps(Fd, Head#head.filename, Head#head.next),
    Ftab = case Acc of
	       read_write -> init_freelist(Head, ExtraInfo);
	       read -> false
	   end,
    check_growth(Head),
    NewHead = Head#head{freelists = Ftab,
			access = Acc,
			ram_file = Ram,
			filename = Fname,
			name = Tab,
		        cache = new_cache(CacheSz)},
    {ok, NewHead}.

%% -> {ok, head()} | throw(Error)
fopen_init_file(Tab, Fname, Type, Kp, Est, Ram, CacheSz, Auto) ->
    FileSpec = if
		   Ram == true -> [];
		   true -> Fname
	       end,
    {ok, Fd} = open(FileSpec, open_args(read_write, Ram)),
    case catch initiate_file(Fd, Tab, Fname, Type, Kp, Est, Ram, CacheSz, Auto)
	of
	{error, Reason} when Ram == true ->
	    file:close(Fd),
	    throw({error, Reason});
	{error, Reason} ->
	    file:close(Fd),
	    file:delete(Fname),
	    throw({error, Reason});
	Reply ->
	    Reply
    end.

%% -> {ok, head()} | throw(Error)
initiate_file(Fd, Tab, Fname, Type, Kp, Est, Ram, CacheSz, Auto) ->
    Factor = est_no_segments(Est),
    position(Fd, Fname, bof),
    truncate(Fd, Fname),
    Freelist = 0,
    Cookie = ?MAGIC,
    Version = ?FILE_FORMAT_VERSION,
    ClosedProperly = ?NOT_PROPERLY_CLOSED, % immediately overwritten
    N = 0,
    M = Next = ?SEGSZ * Factor,
    NoObjects = 0,
    pwrite(Fd, Fname, 0, 
	   <<Freelist:32,
	   Cookie:32,
	   ClosedProperly:32,
	   (type_to_code(Type)):32,
	   Version:32,
	   M:32,
	   Next:32,
	   Kp:32,
	   NoObjects:32,
	   N:32>>),

    Ftab = init_alloc(),

    %% now we need to initialize the segment pointer array,
    position(Fd, Fname, ?HEADSZ),
    zero(Fd, Fname, ?SEGARRSZ),
    %% We also need to initialize the first segement
    zero(Fd, Fname, ?SEGSZ),
    %% and we must set the first slot of the segment pointer array to
    %% point to the first segment

    Pos = ?SEGADDR(0),
    SegP = (?HEADSZ + (4 * ?SEGARRSZ)),
    pwrite(Fd, Fname, Pos, <<SegP:32>>),
    segp_cache(Pos, SegP),

    H0 = #head{freelists=Ftab, fptr = Fd},
    {H1, Ws} = init_more_segments(H0, 1, Factor, undefined, []),
    pwrite(Fd, Fname, Ws),

    %% Return a new nice head structure
    Head = #head{
      m  = M,
      next = Next,
      fptr = Fd,
      no_objects = NoObjects,
      n = N,
      type = Type,
      freelists = H1#head.freelists,
      auto_save = Auto,
      hash_bif = phash,
      keypos = Kp,
      est_no_objects = Est,
      
      ram_file = Ram, 
      filename = Fname, 
      name = Tab,
      cache = new_cache(CacheSz)
     },
    case catch do_perform_save(Head) of
	{NewHead, ok} ->
	    {ok, NewHead};
	{_, Error} ->
	    throw(Error)
    end.

est_no_segments(Est) when 1 + (Est div ?SEGSZ) > ?SEGARRSZ ->
    ?SEGARRSZ;
est_no_segments(Est) ->
    1 + (Est div ?SEGSZ).

init_more_segments(Head, SegNo, Factor, undefined, Ws) when SegNo < Factor ->
    init_more_segments(Head, SegNo, Factor, seg_zero(), Ws);
init_more_segments(Head, SegNo, Factor, SegZero, Ws) when SegNo < Factor ->
    {NewHead, W} = allocate_segment(Head, SegZero, SegNo),
    init_more_segments(NewHead, SegNo+1, Factor, SegZero, W++Ws);
init_more_segments(Head, _SegNo, _Factor, _SegZero, Ws) ->
    {Head, Ws}.

%% -> {ok, head(), ExtraInfo} | {error, Reason} (Reason lacking file name)
%% ExtraInfo = {convert_freelist, Version} | true | need_compacting 
check_file_header(FH, Fd) ->
    Test = 
	if
	    FH#fileheader.cookie /= ?MAGIC ->
		{error, not_a_dets_file};
	    FH#fileheader.type == badtype ->
		{error, invalid_type_code};
	    FH#fileheader.version /= ?FILE_FORMAT_VERSION -> 
		case lists:member(FH#fileheader.version,?CAN_BUMP_BY_REPAIR) of
		    true ->
			{error, version_bump};
		    false ->
			{error, bad_version}
		end;
	    FH#fileheader.trailer /= FH#fileheader.eof ->
		{error, not_closed};
	    FH#fileheader.closed_properly == ?CLOSED_PROPERLY ->
		case lists:member(FH#fileheader.version,
				  ?CAN_CONVERT_FREELIST) of
		    true ->
			{ok, {convert_freelist, FH#fileheader.version}, hash};
		    false ->
			{error, not_closed} % should not happen
		end;
	    FH#fileheader.closed_properly == ?CLOSED_PROPERLY2 ->
		{ok, true, hash};
	    FH#fileheader.closed_properly == 
	          ?CLOSED_PROPERLY2_NEED_COMPACTING  ->
		{ok, need_compacting, hash};
	    FH#fileheader.closed_properly == ?CLOSED_PROPERLY_NEW_HASH ->
		{ok, true, phash};
	    FH#fileheader.closed_properly == 
	         ?CLOSED_PROPERLY_NEW_HASH_NEED_COMPACTING  ->
		{ok, need_compacting, phash};
	    FH#fileheader.closed_properly == ?NOT_PROPERLY_CLOSED ->
		{error, not_closed};
	    FH#fileheader.closed_properly > 
	         ?CLOSED_PROPERLY_NEW_HASH_NEED_COMPACTING ->
		{error, not_closed};
	    true ->
		exit(abnormal)
	end,
    case Test of
	{ok, ExtraInfo, HashAlg} ->
	    H = #head{
	      m = FH#fileheader.m,
	      next = FH#fileheader.next,
	      fptr = Fd,
	      no_objects= FH#fileheader.no_objects,
	      n = FH#fileheader.n,
	      type = FH#fileheader.type,
	      update_mode = saved,
	      auto_save = infinity,             % not saved on file
	      fixed = false,			% not saved on file
	      freelists_p = FH#fileheader.freelist,
	      hash_bif = HashAlg,
	      keypos = FH#fileheader.keypos},
	    {ok, H, ExtraInfo};
	Error ->
	    Error
    end.

%% -> {ok, Fd, fileheader()} | throw(Error)
read_file_header(FileName, Access, RamFile) ->
    BF = if
	     RamFile == true ->
		 case file:read_file(FileName) of
		     {ok, B} -> B;
		     Err -> file_error(FileName, Err)
		 end;
	     true ->
		 FileName
	 end,
    {ok, Fd} = open(BF, open_args(Access, RamFile)),
    {ok, Bin} = pread_close(Fd, FileName, 0, ?HEADSZ),
    [Freelist, Cookie, CP, Type2, Version, M, Next, Kp, NoObjects, N] = 
	bin2ints(Bin),
    {ok, EOF} = position_close(Fd, FileName, eof),
    {ok, <<FileSize:32>>} = pread_close(Fd, FileName, EOF-4, 4),
    FH = #fileheader{freelist = Freelist,
		     cookie = Cookie,
		     closed_properly = CP,
		     type = code_to_type(Type2),
		     version = Version,
		     m = M,
		     next = Next,
		     keypos = Kp,
		     no_objects = NoObjects,
		     trailer = FileSize,
		     eof = EOF,
		     n = N},
    {ok, Fd, FH}.

open_args(Access, RamFile) ->
    A1 = case Access of
	     read -> [];
	     read_write -> [write]
	 end,
    A2 = case RamFile of
	     true -> [ram];
	     false -> [raw]
	   end,
    A1 ++ A2 ++ [binary, read].

type_to_code(set) -> ?SET;
type_to_code(bag) -> ?BAG;
type_to_code(duplicate_bag) -> ?DUPLICATE_BAG.

code_to_type(?SET) -> set;
code_to_type(?BAG) -> bag;
code_to_type(?DUPLICATE_BAG) -> duplicate_bag;
code_to_type(_Type) -> badtype.

cache_segps(Fd, FileName, M) ->
    NSegs = ((M - 1)  div ?SEGSZ) + 1,
    {ok, Bin} = pread_close(Fd, FileName, ?HEADSZ, 4 * NSegs),
    Fun = fun(S, P) -> segp_cache(P, S), P+4 end,
    lists:foldl(Fun, ?HEADSZ, bin2ints(Bin)).

bin2ints(<<>>) ->
    [];
bin2ints(<<Int:32, B/binary>>) ->
    [Int | bin2ints(B)].

%% -> {Head, Result}, Result = ok | Error | {thrown, Error}
finit(Head, _InitFun) when Head#head.access == read ->
    {Head, {error, {access_mode, Head#head.filename}}};
finit(Head, _InitFun) when Head#head.fixed == true ->
    {Head, {error, {fixed_table, Head#head.name}}};
finit(Head, InitFun) ->
    #head{fptr = Fd, type = Type, keypos = KeyPos, auto_save = Auto,
          cache = Cache, filename = Fname, access = Acc, ram_file = Ram,
          name = Tab} = Head,
    Est = Head#head.m,
    CacheSz = {Cache#cache.delay, Cache#cache.size},
    case catch fsck_try(Fd, Tab, Type, KeyPos, Fname, InitFun, Est) of
        ok ->
            case catch fopen(Tab, Fname, Type, KeyPos, false, Est, Ram,
                             CacheSz, Auto, Acc) of
                {ok, NewHead} ->
                    {NewHead, ok};
                Else ->
                    %% Head#head.fptr has been closed...
                    {Head, Else}
            end;
        Else ->
            %% Head#head.fptr has been closed...
            {Head, Else}
    end.

%% Given a filename, fsck it
fsck(Fname) ->
    catch begin
      {ok, Fd, FH} = read_file_header(Fname, read, false),
      ?DEBUGF("FileHeader: ~p~n", [FH]),	    
      fsck(Fd, make_ref(), Fname, FH, ?BUMP)
      end.

%% -> ok | Error
%% Fd is closed.
fsck(Fd, Tab, Fname, FH, Bump) ->
    #fileheader{type = Type, keypos = KeyPos, no_objects = Est} = FH,
    %% We first try and sort on slots using Est. If the number
    %% of objects turns out to be significantly different from
    %% Est, we try again with the correct number of objects.
    case catch fsck_try(Fd, Tab, Type, KeyPos, Fname, Bump, Est) of
        {try_again, NoObjs} ->
            case catch fsck_try(Fd, Tab, Type, KeyPos, Fname, Bump, NoObjs) of
                {try_again, _} ->
                    file:close(Fd),
                    {error, {cannot_repair, Fname}};
                Else ->
                    Else
            end;
        Else ->
            Else
    end.

-define(FSCK_SEGMENT, 10000).

-define(DCT(D, CT), [D | CT]).

%% -> ok | {try_again, integer()} (if Init is Bump) | throw(Error)
%% Fd is closed unless {try_again, _} is returned.
%% Initiating a table using a fun and repairing (or converting) a
%% file are completely different things, but nevertheless the same
%% method is used in both cases...
fsck_try(Fd, Tab, Type, KeyPos, Fname, Init, Est) ->
    Tmp = lists:concat([Fname, ".TMP"]),
    file:delete(Tmp),
    case catch fopen(Tab, Tmp, Type,  KeyPos, false, Est, false, 
		     ?DEFAULT_CACHE, infinity, read_write) of
	{error, Reason} -> 
	    file:close(Fd),
	    throw({error, Reason});
	{ok, Head} ->
	    Cntrs = ets:new(dets_repair, []),
	    IsBulkInit = case Init of
			     _ when function(Init) -> true;
			     _ -> false
			 end,
	    Input = case IsBulkInit of
			true -> 
			    Ref = make_ref(),
                            bulk_input(Head, KeyPos, Init, Cntrs, Ref);
			false ->
			    Bump = Init,
			    State0 = do_fsck(?BASE, Fd, []),
			    fsck_input(Head, State0, Fd, Bump, Cntrs)
		    end,
            Output = fsck_output(Head, Est, Cntrs, IsBulkInit),
	    TmpDir = filename:dirname(Fname),
	    Reply = 
		(catch file_sorter:sort(Input, Output, 
					[{format, 
					  fun(X) -> split_binary(X, 4) end},
					 {tmpdir, TmpDir}])),
	    L = ets:tab2list(Cntrs),
	    ets:delete(Cntrs),
	    case Reply of 
		{ok, NoDups, H0} ->
		    file:close(Fd),
		    H1 = H0#head{no_objects = H0#head.no_objects - NoDups},
		    SizeData = lists:reverse(lists:sort(L)),
		    {H2, R1} = 
			fsck_copy(SizeData, H1, IsBulkInit, NoDups),
		    R2 = fclose(H2#head{update_mode = dirty}),
		    R = if 
			    R1 == ok, R2 == ok ->
				rename(Tmp, Fname);
			    R1 == ok -> R2;
			    true -> R1
			end,
		    if 
			R == ok -> ok;
			true ->
			    file:delete(Tmp),
			    throw(R)
		    end;
		{try_again, _} = Return ->
		    close_files(L, Head),
		    file:delete(Tmp),
		    Return;
		Else ->
		    file:close(Fd),
		    close_files(L, Head),
		    file:delete(Tmp),
		    throw(Else)
	    end
    end.

fsck_copy([SzData | L], Head, Init, NoDups) ->
    {LogSz, Pos, FileName, Fd, NoObjects} = SzData,
    close_tmp(Fd),
    Out = Head#head.fptr,
    {ok, Pos} = file:position(Out, Pos),
    CR = file:copy({FileName, [raw,binary]}, Out),
    file:delete(FileName),
    Size = ?POW(LogSz-1),
    ExpectedSize = Size * NoObjects,
    case CR of 
	{ok, Copied} when Copied == ExpectedSize;
			  NoObjects == 0 -> % the segments
	    fsck_copy(L, Head, Init, NoDups);
	{ok, Copied} when Init == true ->
	    NoZeros = ExpectedSize - Copied,
	    Dups = NoZeros div Size,
	    Addr = Pos+Copied,
	    NewHead = free_n_objects(Head, Addr, Size-1, NoDups),
	    NewNoDups = NoDups - Dups,
	    fsck_copy(L, NewHead, Init, NewNoDups);
	{ok, _} -> % should never happen
	    close_files(L, Head),
	    Reason = if Init == true -> initialization_failed; 
			true -> repair_failed end,
            NR = {error, {Reason, Head#head.filename}},
	    {Head, NR};
	FError ->
	    close_files(L, Head),
	    NR = {error, {file_error, FileName, FError}},
	    {Head, NR}
    end;
fsck_copy([], Head, _Init, NoDups) when NoDups =/= 0 ->
    Error = {error, {initialization_failed, Head#head.filename}},
    {Head, Error};
fsck_copy([], Head, _Init, _NoDups) ->
    {Head, ok}.

free_n_objects(Head, _Addr, _Size, 0) ->
    Head;
free_n_objects(Head, Addr, Size, N) ->
    NewHead = free(Head, Addr, Size),
    NewAddr = Addr + Size + 1,
    free_n_objects(NewHead, NewAddr, Size, N-1).

close_files(L, Head) ->
    Fun = fun({_Size, _Pos, FileName, Fd, _No}) ->
		  close_tmp(Fd),
		  file:delete(FileName);
	     (_) ->
		  ok
	  end,
    lists:foreach(Fun, L),
    file:close(Head#head.fptr).

close_tmp(Fd) ->
    file:close(Fd).

bulk_input(Head, Kp, InitFun, Cntrs, Ref) ->
    fun(close) ->
	    ok;
       (read) ->
	    case catch {Ref, InitFun(read)} of
		{Ref, end_of_input} ->
		    end_of_input;
		{Ref, {L0, NewInitFun}} when list(L0), function(NewInitFun) ->
		    Fun = fun(T) -> 
				  Key = element(Kp, T),
				  BT = term_to_binary(T),
				  Sz = size(BT),
				  LogSz = sz2pos(Sz+?OHDSZ),
				  make_object(Head, Key, LogSz, BT) end,
		    L = lists:map(Fun, L0),
		    R = count_input(Cntrs, L, []),
		    {R, bulk_input(Head, Kp, NewInitFun, Cntrs, Ref)};
		{Ref, Value} ->
		    {error, {init_fun, Value}};
		Error ->
		    throw({thrown, Error})
	    end
    end.

%% Fd is not closed.
fsck_input(Head, State, Fd, Bump, Cntrs) ->
    fun(close) ->
	    ok;
       (read) ->
	    case State of
		done ->
		    end_of_input;
		{done, L} ->
		    R = count_input(Cntrs, L, []),
		    {R, fsck_input(Head, done, Fd, Bump, Cntrs)};
		{cont, L, Bin, Pos} ->
		    R = count_input(Cntrs, L, []),
		    NewState = do_fsck(Bin, Pos, Fd, 
				       Head#head.keypos, Head, Bump, []),
		    {R, fsck_input(Head, NewState, Fd, Bump, Cntrs)}
	    end
    end.

%% The ets table Cntrs is used for counting objects per size.
count_input(Cntrs, [[LogSz | B] | Ts], L) ->
    case catch ets:update_counter(Cntrs, LogSz, 1) of
	N when integer(N) -> ok;
	_Badarg -> true = ets:insert(Cntrs, {LogSz, 1})
    end,
    count_input(Cntrs, Ts, [B | L]);
count_input(_Cntrs, [], L) ->
    L.

fsck_output(Head, Est, Cntrs, IsBulkInit) ->
    fun(close) ->
	    {ok, 0, Head};
       ([]) ->
	    fsck_output(Head, Est, Cntrs, IsBulkInit);
       (L) ->
	    %% Descending sizes.
	    Count = lists:sort(ets:tab2list(Cntrs)),
	    RCount = lists:reverse(Count),
	    NoObjects = lists:foldl(fun({_Sz,No}, A) -> A + No end, 0, Count),
	    if
		%% Using number of objects for bags and duplicate bags
		%% is not ideal; number of (unique) keys should be
		%% used instead. The effect is that there will be more
		%% segments than "necessary".
		abs(NoObjects div ?SEGSZ - Est div ?SEGSZ) > 1,
		(NoObjects < ?MAXOBJS),
		IsBulkInit == false ->
		    {try_again, NoObjects};
		true ->
                    Head1 = Head#head{no_objects = NoObjects},
		    {Head2, CT} = allocate_all_objects(Head1, RCount, Cntrs),
		    [E | Es] = bin2term(L, []),
                    ZT = make_zero_table(512),
		    {NE, Acc, DCT1} = 
			output_slots(E, Es, [E], ZT, Head2, ?DCT(0, CT)),
		    NDCT = write_all_sizes(DCT1, Cntrs),
		    Max = ets:info(Cntrs, size),
		    fsck_output2(NE, Acc, Head2, ZT, Cntrs, NDCT, Max, Max)
	    end
    end.

-ifdef(vector).
-define(VNEW(N, E), vector:new(N, E)).
-define(VSET(I, V, E), vector:set(I, V, E)).
-define(VGET(I, V), vector:get(I, V)).
-else.
-define(VNEW(N, E), erlang:make_tuple(N, E)).
-define(VSET(I, V, E), setelement(I, V, E)).
-define(VGET(I, V), element(I, V)).
-endif.

%% By allocating bigger objects before smaller ones, holes in the
%% buddy system memory map are avoided. Unfortunately, the segments
%% are always allocated first, so if there are objects bigger than a
%% segment, there is a hole to handle. (Haven't considered placing the
%% segments among other objects of the same size.)
allocate_all_objects(Head, Count, Cntrs) ->
    SegSize = actual_seg_size(),
    {Head1, HSz, HN, HA} = alloc_hole(Count, Head, SegSize),
    {Max, _} = hd(Count),
    CT = ?VNEW(Max+1, not_used),
    {Head2, NCT} = allocate_all(Head1, Count, Cntrs, CT),
    Head3 = free_hole(Head2, HSz, HN, HA),
    {Head3, NCT}.

alloc_hole([{LSize,_} | _], Head, SegSz) when ?POW(LSize-1) > SegSz ->
    {_, SegAddr} = alloc(Head, SegSz-1),
    Size = ?POW(LSize-1)-1,
    {_, Addr} = alloc(Head, Size),
    N = (Addr - SegAddr) div SegSz,
    Head1 = alloc_many(Head, SegSz, N, SegAddr),
    {Head1, SegSz-1, N, SegAddr};
alloc_hole(_Count, Head, _SegSz) ->
    {Head, 0, 0, 0}.

free_hole(Head, _Size, 0, _Addr) ->
    Head;
free_hole(Head, Size, N, Addr) ->
    Head1 = free(Head, Addr, Size),
    free_hole(Head1, Size, N-1, Addr+Size+1).

%% One (temporary) file for each buddy size, write all objects of that
%% size to the file.
%% The contents of the ets-table Cntrs is replaced here...
allocate_all(Head, [{LSize,NoObjects} | Count], Cntrs, CT) ->
    Size = ?POW(LSize-1)-1,
    {_Head, Addr} = alloc(Head, Size),
    NewHead = alloc_many(Head, Size+1, NoObjects, Addr),
    {FileName, Fd} = temp_file(Head, LSize),
    true = ets:insert(Cntrs, {LSize, Addr, FileName, Fd, NoObjects}),
    NCT = ?VSET(LSize, CT, [Addr | []]),
    allocate_all(NewHead, Count, Cntrs, NCT);
allocate_all(Head, [], Cntrs, CT) ->
    %% And one file for the segments...
    {FileName, Fd} = temp_file(Head, ?FSCK_SEGMENT),
    Addr = ?SEGADDR(?SEGARRSZ),
    true = ets:insert(Cntrs, {?FSCK_SEGMENT, Addr, FileName, Fd, 0}),
    NCT = ?VSET(size(CT), CT, [Addr | []]),
    {Head, NCT}.

temp_file(Head, N) ->
    TmpName = lists:concat([Head#head.filename, '.', N]),
    {ok, Fd} = open(TmpName, [raw, binary, write]),
    {TmpName, Fd}.

fsck_output2(E, Acc, Head, ZT, Cntrs, DCT, 0, MaxNoChunks) ->
    NDCT = write_all_sizes(DCT, Cntrs),
    fsck_output2(E, Acc, Head, ZT, Cntrs, NDCT, MaxNoChunks, MaxNoChunks);
fsck_output2(E, Acc, Head, ZT, Cntrs, DCT, ChunkI, MaxNoChunks) ->
    fun(close) ->
	    NDCT = output_slot(Acc, ZT, Head, DCT),
	    write_all_sizes(NDCT, Cntrs),
	    ?DCT(NoDups, _) = NDCT,
	    {ok, NoDups, Head};
       (L) ->
	    Es = bin2term(L, []),
	    {NE, NAcc, NDCT} = 
		output_slots(E, Es, Acc, ZT, Head, DCT),
	    fsck_output2(NE, NAcc, Head, ZT, Cntrs, NDCT, ChunkI-1,MaxNoChunks)
    end.

bin2term([<<Slot:32, LogSize:8, BinTerm/binary>> | BTs], L) ->
    bin2term(BTs, [{Slot, LogSize, BinTerm} | L]);
bin2term([], L) ->
    lists:reverse(L).

write_all_sizes(?DCT(D, CT), Cntrs) ->
    ?DCT(D, write_sizes(1, size(CT), CT, Cntrs)).

write_sizes(Sz, Sz, CT, Cntrs) ->
    write_size(Sz, ?FSCK_SEGMENT, CT, Cntrs);
write_sizes(Sz, MaxSz, CT, Cntrs) ->
    NCT = write_size(Sz, Sz, CT, Cntrs),
    write_sizes(Sz+1, MaxSz, NCT, Cntrs).

write_size(Sz, I, CT, Cntrs) ->
    case ?VGET(Sz, CT) of
	not_used ->
	    CT;
	[Addr | L] ->
	    Fd = ets:lookup_element(Cntrs, I, 4),
	    case file:write(Fd, lists:reverse(L)) of
		ok ->
		    ?VSET(Sz, CT, [Addr | []]);
		Error ->
		    FileName = ets:lookup_element(Cntrs, I, 3),
		    file_error(FileName, Error)
	    end
    end.

output_slots(E, [E1 | Es], Acc, ZT, Head, DCT) 
                       when element(1, E) == element(1, E1) ->
    output_slots(E1, Es, [E1 | Acc], ZT, Head, DCT);
output_slots(_E, [E | L], Acc, ZT, Head, DCT) ->
    NDCT = output_slot(Acc, ZT, Head, DCT),
    output_slots(E, L, [E], ZT, Head, NDCT);
output_slots(E, [], Acc, _ZT, _Head, DCT) ->
    {E, Acc, DCT}.

output_slot([E], ZT, _Head, ?DCT(D, CT)) ->
    ?DCT(D, output_slot([{foo, E}], 0, ZT, foo, CT));
output_slot(Es0, ZT, Head, ?DCT(D, CT)) ->
    Kp = Head#head.keypos,
    Fun = fun({_Slot, _LSize, BinTerm} = E) -> 
		  Key = element(Kp, binary_to_term(BinTerm)),
		  {Key, E}
	  end,
    Es = lists:map(Fun, Es0),
    NEs = case Head#head.type of
	      set ->
		  [{Key0,_} = E | L0] = lists:sort(Es),
		  choose_one(lists:sort(L0), Key0, [E]);
	      bag -> 
		  lists:usort(Es);
	      duplicate_bag -> 
		  lists:sort(Es)
	  end,
    Dups = D + length(Es) - length(NEs),
    ?DCT(Dups, output_slot(NEs, 0, ZT, foo, CT)).

choose_one([{Key,_} | Es], Key, L) ->
    choose_one(Es, Key, L);
choose_one([{Key,_} = E | Es], _Key, L) ->
    choose_one(Es, Key, [E | L]);
choose_one([], _Key, L) ->
    L.

output_slot([E | Es], Next, ZT, _Slot, CT) ->
    {_Key, {Slot, LSize, BinTerm}} = E,
    Size = size(BinTerm),
    Size2 = ?POW(LSize-1),
    Z = get_zeros(Size2-Size-?OHDSZ, ZT),
    BinObject = [<<Next:32, Size:32, ?ACTIVE:32>>, BinTerm | Z],
    [Addr | L] = ?VGET(LSize, CT),
    NCT = ?VSET(LSize, CT, [Addr+Size2 | [BinObject | L]]),
    output_slot(Es, Addr, ZT, Slot, NCT);
output_slot([], Next, ZT, Slot, CT) ->
    I = size(CT),
    [Addr | L] = ?VGET(I, CT),
    {Pos, _} = slot_position(Slot),
    NoZeros = Pos - Addr,
    Z = get_zeros(NoZeros, ZT),
    BinObject = [Z | <<Next:32>>],
    Size = NoZeros+4,
    ?VSET(I, CT, [Addr+Size | [BinObject | L]]).

%% M is a power of 2.
make_zero_table(M) ->
    Z = lists:duplicate(M, 0),
    T = make_zero_tuple(Z, []),
    L = make_zero_powers(1 bsl 31, M, Z, []),
    {M, T, L}.

make_zero_tuple([], L) ->
    list_to_tuple(L);
make_zero_tuple(Z, L) ->
    make_zero_tuple(tl(Z), [Z | L]).

make_zero_powers(M, M, Z, L) ->
    lists:reverse([Z | L]);
make_zero_powers(I, M, Z, L) ->
    make_zero_powers(I div 2, M, [Z | Z], [Z | L]).

%% Inlined.
get_zeros(0, _) ->
    [];
get_zeros(N, {M, S, _}) when N < M ->
    element(N, S);
get_zeros(N, {M, _S, L}) when N rem M =:= 0 ->
    get_zeros2(N div M, L);
get_zeros(N, {M, S, L}) ->
    [element(N rem M, S) | get_zeros2(N div M, L)].

get_zeros2(0, _) ->
    [];
get_zeros2(N, [_ | L]) when N band 1 =:= 0 ->
    get_zeros2(N div 2, L);
get_zeros2(N, [Z | L]) ->
    [Z | get_zeros2(N div 2, L)].

%% A modified version of scan/3; repairing is a sake of its own.
do_fsck(Pos, F, L) ->
    case file:position(F, Pos) of
	{ok, _} ->
	    read_more_bytes(<<>>, 0, Pos, F, L);
	_Error ->
	    {done, L}
    end.

do_fsck(Bin = <<_N:32, Sz:32, Status:32, Tail/binary>>, 
	Pos, F, Kp, Head, Bump, L) ->
    if 
	Status == ?ACTIVE ->
	    case Tail of
		<<BinTerm:Sz/binary, Tail2/binary>> ->
		    case catch element(Kp, binary_to_term(BinTerm)) of
			{'EXIT', _} ->
			    skip_bytes(Bin, Bump, Pos, F, Kp, Head, Bump, L);
			Key when Bump == ?BUMP ->
			    LogSz = sz2pos(Sz+?OHDSZ),
			    Obj = make_object(Head, Key, LogSz, BinTerm),
			    NL = [Obj | L],
			    Skip = ?POW(LogSz-1) - Sz - ?OHDSZ,
			    skip_bytes(Tail2, Skip, Pos, F, Kp, Head, Bump,NL);
			Key when Bump == 1 ->
			    LogSz = sz2pos(Sz+?OHDSZ),
			    Obj = make_object(Head, Key, LogSz, BinTerm),
			    NL = [Obj | L],
			    do_fsck(Tail2, Pos, F, Kp, Head, Bump, NL)
		    end;
		_ ->
		    read_more_bytes(Bin, Sz, Pos, F, L)
	    end;
	true -> 
	    skip_bytes(Bin, Bump, Pos, F, Kp, Head, Bump, L)
    end;
do_fsck(Bin, Pos, F, _Kp, _Head, _Bump, L) ->
    read_more_bytes(Bin, 0, Pos, F, L).
    
%% Inlined.
make_object(Head, Key, LogSz, BT) ->
    Slot = db_hash(Key, Head),
    [LogSz | <<Slot:32, LogSz:8, BT/binary>>].

%% Inlined.
skip_bytes(Bin, Skip, Pos, F, Kp, Head, Bump, L) ->
    case Bin of
	<<_:Skip/binary, Tail/binary>> ->
	    do_fsck(Tail, Pos, F, Kp, Head, Bump, L);
	_ ->
	    NewPos = Pos + (Skip - size(Bin)),
	    do_fsck(NewPos, F, L)
    end.

read_more_bytes(B, Min, Pos, F, L) ->
    Max = if 
	      Min < ?CHUNK_SIZE -> ?CHUNK_SIZE; 
	      true -> Min 
	  end,
    case read_n(F, Max) of
	eof ->
	    {done, L};
	Bin ->
	    NewPos = Pos + size(Bin),
	    {cont, L, list_to_binary([B, Bin]), NewPos}
    end.

fclose(Head) ->
    {Head1, Res} = perform_save(Head),
    case Head1#head.ram_file of
	true -> 
	    ignore;
	false -> 
	    delete_cache(Head1#head.cache),
	    file:close(Head1#head.fptr)
    end,
    Res.

%% -> {NewHead, Res}
perform_save(Head) when Head#head.update_mode == dirty ->
    catch do_perform_save(Head);
perform_save(Head) when Head#head.update_mode == saved ->
    ?DEBUGF("~p: Already saved.~n", [Head#head.name]),
    %% Already synced
    {Head, ok};
perform_save(Head) ->
    ?DEBUGF("~p: Table is in error status, no save performed.~n", 
	    [Head#head.name]),
    {Head, Head#head.update_mode}. % Or 'ok'?

%% -> {NewHead, ok} | throw({Head, Error})
do_perform_save(H0) ->
    {H, []} = write_cache(H0),

    FL = get_freelists(H),
    B = term_to_binary(FL),
    Size = size(B),
    ?DEBUGF("size of freelist = ~p~n", [Size]),
    ?DEBUGF("head.m = ~p~n", [H#head.m]),
    ?DEBUGF("head.no_objects = ~p~n", [H#head.no_objects]),

    {ok, Pos} = position(H, eof),
    W1 = {?FREELIST_POS, <<Pos:32>>},
    W2 = {Pos, [<<0:32, Size:32, ?FREE:32>>, B]},
    
    W3 = {?D_POS, <<(H#head.m):32, 
	            (H#head.next):32, 
	            (H#head.keypos):32,
	            (H#head.no_objects):32,
		    (H#head.n):32>>},
    {ClosedProperly, ClosedProperlyNeedCompacitng} = 
	case H#head.hash_bif of
	    hash ->
		{?CLOSED_PROPERLY2, ?CLOSED_PROPERLY2_NEED_COMPACTING};
	    phash ->
		{?CLOSED_PROPERLY_NEW_HASH, 
		 ?CLOSED_PROPERLY_NEW_HASH_NEED_COMPACTING}
	end,
    W4 = 
	if 
	    Size > 1000, Size > H#head.no_objects ->
		{?CLOSED_PROPERLY_POS, 
		 <<ClosedProperlyNeedCompacitng:32>>};
	    true ->
		{?CLOSED_PROPERLY_POS, <<ClosedProperly:32>>}
	end,
    W5 = {?FILE_FORMAT_VERSION_POS, <<?FILE_FORMAT_VERSION:32>>},
    {H2, ok} = pwrite(H, [W1,W2,W3,W4,W5]),
    {ok, Pos2} = position(H2, eof),
    ?DEBUGF("Writing file size ~p, eof at ~p~n", [Pos2+4, Pos2]),
    {H3, ok} = pwrite(H2, [{Pos2, <<(Pos2 + 4):32>>}]),
    ok = ensure_written(H3),
    {H3#head{update_mode = saved, freelists_p = Pos}, ok}.

ensure_written(H) when H#head.ram_file == true ->
    {ok, EOF} = position(H, eof),
    {ok, Bin} = pread(H, 0, EOF, 0),
    case file:write_file(H#head.filename, Bin) of
	ok ->
	    ok;
	{error, Reason} ->
	    FError = {error, {file_error, H#head.filename, Reason}},
	    throw({H, FError})
    end;
ensure_written(H) when H#head.ram_file == false ->
    file:sync(H#head.fptr).
        
%% -> {Head, Result}
fselect(Head, MP, Spec, N) ->
    KeyPos = Head#head.keypos,
    case find_all_keys(Spec, KeyPos, []) of
	[] -> 
	    %% Complete match
	    MFun = fun(Term) ->
			   case ets:match_spec_run([Term], MP) of
			       [] ->
				   false;
			       [Else] ->
				   {true, Else}
			   end
		   end,
	    catch chunk_begin(Head, MFun, N);
	List ->
	    Keys = lists:usort(List),
	    {NewHead, Reply} = flookup_keys(Head, Keys),
	    case Reply of
		Objs when list(Objs) ->
		    MatchingObjs = ets:match_spec_run(Objs, MP),
		    {NewHead, {done, MatchingObjs}};
		Error ->
		    {NewHead, {done, Error}}
	    end
    end.

find_all_keys([], _, Ks) ->
    Ks;
find_all_keys([{H,_,_} | T], KeyPos, Ks) when tuple(H) ->
    case size(H) of
	Enough when Enough >= KeyPos ->
	    Key = element(KeyPos, H),
	    case contains_variable(Key) of
		true -> 
		    [];
		false ->
		    find_all_keys(T, KeyPos, [Key | Ks])
	    end;
	_ ->
	    find_all_keys(T, KeyPos, Ks)
    end;
find_all_keys(_, _, _) ->
    [].

%% {Head, Cont} | throw({Head, Error})
chunk_begin(Head, MFun, N) ->
    Fun = 
	fun(O, Ack) ->
		case MFun(O) of
		    {true, ObjOrBs} ->
			[ObjOrBs | Ack];
		    false ->
			Ack
		end
	end,
    {NewHead, []} = write_cache(Head),
    C0 = init_scan(NewHead, N),
    {NewHead, {cont, C0#dets_cont{arg = Fun}}}.

fmatch(Head, C) ->
    case scan(Head, binary, C) of
	{scan_error, Reason} ->
	    corrupt_reason(Head, Reason);
	{Ts, NC} ->
	    {Head, {cont, {Ts, NC}}}
    end.

%% -> {NewHead, Res}
fmatch_delete_begin(Head, _MP, ?PATTERN_TO_OBJECT_MATCH_SPEC('_'))  ->
    %% Handle the case where the file is emptied efficiently.
    {NewHead, Result} = fdelete_all_objects(Head),
    {NewHead, {done, Result}};
fmatch_delete_begin(Head, MP, Spec)  ->
    KeyPos = Head#head.keypos,
    MFun = fun(Term) ->
		   [] =/= ets:match_spec_run([Term], MP)
	   end,
    case find_all_keys(Spec, KeyPos, []) of
	[] -> 
	    {NewHead, []} = write_cache(Head),
	    C0 = init_scan(NewHead, default),
	    {NewHead, {cont, C0#dets_cont{arg = MFun}}};
	List ->
	    Keys = lists:usort(List),
            do_fmatch_delete_fixed(Head, Keys, MFun)
    end.

%% A note: It is possible that the binary in C contains the first part
%% of _one_ term. If that term is deleted here, the binary will not be
%% affected, which means that next call to scan/1 will find that term
%% active (unless it was free already...). The (remaining) terms of
%% the slot where the term resided will then be traversed once more.
fmatch_delete(Head, C) ->
    case scan(Head, term, C) of
	{scan_error, Reason} ->
	    {NewHead, Error} = corrupt_reason(Head, Reason),
	    {NewHead, {done, Error}};
	{[], _} ->
	    {Head, {done, ok}};
	{RTs, NC} ->
	    MFun = C#dets_cont.arg,
	    Terms = lists:filter(MFun, RTs),
	    do_fmatch_delete(Head, Terms, NC)
    end.

do_fmatch_delete_fixed(Head, Keys, MFun) ->
    case flookup_keys(Head, Keys) of
	{NewHead, ReadTerms} when list(ReadTerms) ->
	    Terms = lists:filter(MFun, ReadTerms),
	    do_fmatch_delete(NewHead, Terms, done);
	{NewHead, Error} ->
	    {NewHead, {done, Error}}
    end.

do_fmatch_delete(Head, Terms, What) ->
    case do_delete(Head, Terms, delete_object) of
	{NewHead, Res} when What == done; Res =/= ok ->
	    {NewHead, {done, Res}};
	{NewHead, ok} ->
	    {NewHead, {cont, What}}
    end.

fdelete_key(Head, Key) ->
    do_delete(Head, [Key], delete_key).

%% -> {NewHead, Reply}, Reply = ok | badarg | Error.
fdelete_object(Head, Object) ->
    Kp = Head#head.keypos,
    case check_objects([Object], Kp) of
	true ->
	    do_delete(Head, [Object], delete_object);
	false ->
	    {Head, badarg}
    end.

do_delete(Head, Things, What) ->
    case catch update_cache(Head, Things, What) of
	{NewHead, []} ->
	    {NewHead, ok};
	HeadError ->
	    HeadError
    end.

%% -> {NewHead, Result}
fdelete_all_objects(Head) ->
    #head{fptr = Fd, name = Tab, filename = Fname, type = Type, keypos = Kp, 
	  ram_file = Ram, auto_save = Auto, est_no_objects = Est,
	  cache = Cache} = Head, 
    CacheSz = {Cache#cache.delay, Cache#cache.size},
    case catch initiate_file(Fd, Tab, Fname, Type, Kp, Est, Ram, CacheSz, Auto)
	of
	{ok, NewHead} ->
	    {NewHead, ok};
	{error, Reason} ->
	    corrupt_reason(Head, Reason)
    end.

finfo(H) -> 
    case catch write_cache(H) of
	{H2, []} ->
	    Info = [{type, H2#head.type}, 
		    {keypos, H2#head.keypos}, 
		    {size, H2#head.no_objects},
		    {file_size, file_size(H2#head.fptr)},
		    {filename, H2#head.filename}],
	    {H2, Info};
	HeadError ->
	    HeadError
    end.

finfo(H, access) -> {H, H#head.access};
finfo(H, auto_save) -> {H, H#head.auto_save};
finfo(H, delayed_write) -> % undocumented
    C = H#head.cache,
    {H, {C#cache.delay, C#cache.size}};
finfo(H, filename) -> {H, H#head.filename};
finfo(H, file_size) ->
    case catch write_cache(H) of
	{H2, []} ->
	    {H2, file_size(H#head.fptr)};
	HeadError ->
	    HeadError
    end;
finfo(H, fixed) -> {H, H#head.fixed}; % true if fixtable/2 has been called
finfo(H, hash) -> {H, H#head.hash_bif};
finfo(H, keypos) -> {H, H#head.keypos};
finfo(H, memory) -> finfo(H, file_size);
finfo(H, pid) -> {H, self()};
finfo(H, ram_file) -> {H, H#head.ram_file};
finfo(H, size) -> 
    case catch write_cache(H) of
	{H2, []} ->
	    {H2, H2#head.no_objects};
	HeadError ->
	    HeadError
    end;
finfo(H, type) -> {H, H#head.type};
finfo(H, _) -> {H, undefined}.

file_size(F) -> 
    {ok, Pos} = file:position(F, eof),
    Pos.

fslot(H, Slot) ->
    catch begin
      {NH, []} = write_cache(H),
      Objs = slot_objs(H, Slot),
      {NH, Objs}
    end.

ffirst(H) ->
    catch ffirst1(H).

ffirst1(H) ->
    {NH, []} = write_cache(H),
    ffirst(NH, 0).

ffirst(H, Slot) ->
    case slot_objs(H, Slot) of
	'$end_of_table' -> {H, '$end_of_table'};
	[] -> ffirst(H, Slot+1);
	[X|_] -> {H, element(H#head.keypos, X)}
    end.

fnext(Head, Key) ->
    Slot = db_hash(Key, Head),
    catch fnext(Head, Key, Slot).

fnext(H, Key, Slot) ->
    {NH, []} = write_cache(H),
    case slot_objs(NH, Slot) of
	'$end_of_table' -> {NH, '$end_of_table'};
	L -> fnext_search(NH, Key, Slot, L)
    end.

fnext_search(H, K, Slot, L) ->
    Kp = H#head.keypos,
    case beyond_key(K, Kp, L) of
	[] -> fnext_slot(H, K, Slot+1);
	L2 -> {H, element(H#head.keypos, hd(L2))}
    end.

%% We've got to continue to search for the next key in the next slot
fnext_slot(H, K, Slot) ->
    case slot_objs(H, Slot) of
	'$end_of_table' -> {H, '$end_of_table'};
	[] -> fnext_slot(H, K, Slot+1);
	L -> {H, element(H#head.keypos, hd(L))}
    end.

beyond_key(_K, _Kp, []) -> [];
beyond_key(K, Kp, [H|T]) when element(Kp, H) /= K ->
    beyond_key(K, Kp, T);
beyond_key(K, Kp, [H|T]) when element(Kp, H) == K ->
    beyond_key2(K, Kp, T).

beyond_key2(_K, _Kp, []) -> [];
beyond_key2(K, Kp, [H|T]) when element(Kp, H) == K ->
    beyond_key2(K, Kp, T);
beyond_key2(_K, _Kp, L) ->
    L.

%% -> [term()] | throw({Head, Error})
slot_objs(H, Slot) when Slot >= H#head.next ->
    '$end_of_table';
slot_objs(H, Slot) ->
    {_Pos, Chain} = chain(H, Slot),
    collect_chain(H, Chain).

collect_chain(_H, 0) -> [];
collect_chain(H, Pos) ->
    {Next, _Sz, Term} = prterm(H, Pos, ?ReadAhead),
    [Term | collect_chain(H, Next)].

%% -> {NewHead, [LookedUpObject]} | {NewHead, Error}
flookup_keys(Head, Keys) ->
    catch do_flookup_keys(Head, Keys).

do_flookup_keys(Head, Keys) ->
    case cache_lookup(Head#head.type, Keys, Head) of
	false ->
	    update_cache(Head, Keys, lookup);
	{Found, []} ->
	    {Head, Found};
	{Found, Ks} ->
	    {Head2, R} = direct_lookup(Head, Ks),
	    {Head2, Found ++ R}
    end.

fmember(Head, Key) ->
    catch do_fmember(Head, Key).

do_fmember(Head, Key) ->
    Keys = [Key],
    {NewHead, LookedUpObjects} = 
	case cache_lookup(Head#head.type, Keys, Head) of
	    false ->
		update_cache(Head, Keys, lookup);
	    {Found, []} ->
		{Head, Found};
	    {[], Ks} ->
		direct_lookup(Head, Ks)
	end,
    {NewHead, LookedUpObjects =/= []}.

%% -> {NewHead, Reply}, Reply = ok | badarg | Error.
finsert(Head, Objects) ->
    case check_objects(Objects, Head#head.keypos) of
	true -> 
	    case catch update_cache(Head, Objects, insert) of
		{NewHead, []} ->
		    {NewHead, ok};
		Reply ->
		    Reply
	    end;
	false -> 
	    {Head, badarg}
    end.

check_objects([T | Ts], Kp) when tuple(T), size(T) >= Kp ->
    check_objects(Ts, Kp);
check_objects(L, _Kp) ->
    L == [].

h(I, HF) -> erlang:HF(I, ?BIG) - 1.  %% stupid BIF has 1 counts.

db_hash(Key, Head) ->
    H = h(Key, Head#head.hash_bif),
    Hash = H rem Head#head.m,
    if
	Hash < Head#head.n ->
	    H rem (2 * Head#head.m);
	true ->
	    Hash
    end.

do_fixtable(Head, false) when Head#head.fixed == true ->
    check_growth(Head),
    Head#head{fixed = false, freelists = get_freelists(Head)};
do_fixtable(Head, true) when Head#head.fixed == false ->
    Ftab = get_freelists(Head),
    Head#head{fixed = true, freelists = {Ftab, Ftab}};
do_fixtable(Head, _Value) ->
    Head.

check_growth(Head) ->
    if 
	Head#head.no_objects > Head#head.next -> % no_keys would be better
	    self() ! {self(), may_grow}; % Catch up.
	true ->
	    ok
    end.

%% -> {NewHead, ok} | {NewHead, Error}
may_grow(Head, _N, _How) when Head#head.fixed == true ->
    {Head, ok};
may_grow(Head, _N, _How) when Head#head.next >= ?MAXOBJS ->
    {Head, ok};
may_grow(Head, N, How) ->
    Extra = Head#head.no_objects + N - Head#head.next,
    case catch may_grow1(Head, Extra, How) of
	{error, Reason} -> % alloc may throw error
	    {Head, {error, Reason}};
	Reply ->
	    Reply
    end.

may_grow1(Head, Extra, many_times) when Extra > ?SEGSZ ->
    Reply = grow(Head, 1, undefined),
    case Reply of
	{_, ok} -> self() ! {self(), may_grow};
	_ -> ok
    end,
    Reply;
may_grow1(Head, Extra, _How) ->    
    grow(Head, Extra, undefined).

%% -> {Head, ok} | throw({Head, Error})
grow(Head, Extra, _SegZero) when Extra =< 0 ->
    {Head, ok};
grow(Head, Extra, undefined) ->
    grow(Head, Extra, seg_zero());
grow(Head, Extra, SegZero) ->
    #head{n = N, next = Next, m = M} = Head,
    SegNum = Next div ?SEGSZ,    
    {Head0, Ws1} = allocate_segment(Head, SegZero, SegNum),
    {Head1, ok} = pwrite(Head0, Ws1),
    %% If re_hash fails, segp_cache has been called, but it does not matter.
    {ok, Ws2} = re_hash(Head1, N),
    {Head2, ok} = pwrite(Head1, Ws2),
    NewHead =
	if 
	    N + ?SEGSZ == M ->
		Head2#head{n = 0, next = Next + ?SEGSZ, m = 2 * M};
	    true ->
		Head2#head{n = N + ?SEGSZ, next = Next + ?SEGSZ}
	end,
    grow(NewHead, Extra - ?SEGSZ, SegZero).

allocate_segment(Head, SegZero, SegNo) ->
    {NewHead, Segment} = alloc(Head, 4 * ?SEGSZ), % may throw error
    InitSegment = {Segment, SegZero},
    Pos = ?SEGADDR(SegNo),
    segp_cache(Pos, Segment),
    SegPointer = {Pos, <<Segment:32>>},
    {NewHead, [InitSegment, SegPointer]}.

seg_zero() ->
    make_zeros(4 * ?SEGSZ).

%% Re-hashing a segment, starting with SlotStart.
%%
%% On the average, half of the objects of the chain are put into a new
%% chain. If the slot of the old chain is i, then the slot of the new
%% chain is i+m.
%% Note that the insertion of objects into the new chain is simplified
%% by the fact that the chains are not sorted on key, which means that
%% each moved object can be inserted first in the new chain.
%% (It is also a fact that the objects with the same key are not sorted.)
%%
%% -> {ok, Writes} | throw({Head, Error})
re_hash(Head, SlotStart) ->
    {SlotPos, _4} = slot_position(SlotStart),
    {ok, Bin} = pread(Head, SlotPos, 4*?SEGSZ, 0),
    {Read, Cs} = split_bin(SlotPos, Bin, [], []),
    re_hash_read(Head, [], Read, Cs).

split_bin(Pos, <<P:32, B/binary>>, R, Cs) ->
    if
	P == 0 ->
	    split_bin(Pos+4, B, R, Cs);
	true ->
	    split_bin(Pos+4, B, [{P,?ReadAhead} | R], [[Pos] | Cs])
    end;
split_bin(_Pos, <<>>, R, Cs) ->
    {R, Cs}.

re_hash_read(Head, Cs, R, RCs) ->
    {ok, Bins} = pread(R, Head),
    re_hash_read(Head, R, RCs, Bins, Cs, [], []).

re_hash_read(Head, [{Pos, Size} | Ps], [C | Cs], 
	     [<<Next:32, Sz:32, _Status:32, Bin0/binary>> | Bins], 
	     DoneCs, R, RCs) ->
    case size(Bin0) of
	BinSz when BinSz >= Sz ->
	    case catch binary_to_term(Bin0) of
		{'EXIT', _} ->
		    throw(corrupt_reason(Head, bad_object));
		Term ->
		    Key = element(Head#head.keypos, Term),
		    New = h(Key, Head#head.hash_bif) rem (2 * Head#head.m),
		    NC = case New >= Head#head.m of
			     true -> [{Pos,New} | C];
			     false -> [Pos | C]
			 end,
		    if
			Next == 0 ->
			    NDoneCs = [NC | DoneCs], 
			    re_hash_read(Head, Ps, Cs, Bins, NDoneCs, R, RCs);
			true ->
			    NR = [{Next,?ReadAhead} | R],
			    NRCs = [NC | RCs],
			    re_hash_read(Head, Ps, Cs, Bins, DoneCs, NR, NRCs)
		    end
	    end;
	BinSz when Size == BinSz+?OHDSZ ->
	    NR = [{Pos, Sz+?OHDSZ} | R],
	    re_hash_read(Head, Ps, Cs, Bins, DoneCs, NR, [C | RCs]);
	_BinSz ->
	    throw({Head, {error, {premature_eof, Head#head.filename}}})
    end;
re_hash_read(Head, [], [], [], Cs, [], []) ->
    re_hash_traverse_chains(Cs, Head, [], [], []);
re_hash_read(Head, [], [], [], Cs, R, RCs) ->
    re_hash_read(Head, Cs, R, RCs).

re_hash_traverse_chains([C | Cs], Head, Rs, Ns, Ws) ->
    case re_hash_find_new(C, Rs, start, start) of
	false ->
	    re_hash_traverse_chains(Cs, Head, Rs, Ns, Ws);
	{NRs, FirstNew, LastNew} -> 
	    LastInNew = case C of
			    [{_,_} | _] -> true;
			    _ -> false
			end,
	    N = {FirstNew, LastNew, LastInNew},
	    NWs = re_hash_link(C, start, start, start, Ws),
	    re_hash_traverse_chains(Cs, Head, NRs, [N | Ns], NWs)
    end;
re_hash_traverse_chains([], Head, Rs, Ns, Ws) ->
    {ok, Bins} = pread(Rs, Head),
    {ok, insert_new(Rs, Bins, Ns, Ws)}.

re_hash_find_new([{Pos,NewSlot} | C], R, start, start) ->
    {SPos, _4} = slot_position(NewSlot),
    re_hash_find_new(C, [{SPos,4} | R], Pos, Pos);
re_hash_find_new([{Pos,_SPos} | C], R, _FirstNew, LastNew) ->
    re_hash_find_new(C, R, Pos, LastNew);
re_hash_find_new([_Pos | C], R, FirstNew, LastNew) ->
    re_hash_find_new(C, R, FirstNew, LastNew);
re_hash_find_new([], _R, start, start) ->
    false;
re_hash_find_new([], R, FirstNew, LastNew) ->
    {R, FirstNew, LastNew}.

re_hash_link([{Pos,_SPos} | C], LastOld, start, _LastInNew, Ws) ->
    re_hash_link(C, LastOld, Pos, true, Ws);
re_hash_link([{Pos,_SPos} | C], LastOld, LastNew, false, Ws) ->
    re_hash_link(C, LastOld, Pos, true, [{Pos,<<LastNew:32>>} | Ws]);
re_hash_link([{Pos,_SPos} | C], LastOld, _LastNew, LastInNew, Ws) ->
    re_hash_link(C, LastOld, Pos, LastInNew, Ws);
re_hash_link([Pos | C], start, LastNew, true, Ws) ->
    re_hash_link(C, Pos, LastNew, false, [{Pos,<<0:32>>} | Ws]);
re_hash_link([Pos | C], LastOld, LastNew, true, Ws) ->
    re_hash_link(C, Pos, LastNew, false, [{Pos,<<LastOld:32>>} | Ws]);
re_hash_link([Pos | C], _LastOld, LastNew, LastInNew, Ws) ->
    re_hash_link(C, Pos, LastNew, LastInNew, Ws);
re_hash_link([], _LastOld, _LastNew, _LastInNew, Ws) ->
    Ws.

insert_new([{NewSlotPos,_4} | Rs], [<<P:32>> = PB | Bins], [N | Ns], Ws) ->
    {FirstNew, LastNew, LastInNew} = N,
    Ws1 = case P of
	      0 when LastInNew == true ->
		  Ws;
	      0 ->
		  [{LastNew, <<0:32>>} | Ws];
	      _ ->
		  [{LastNew, PB} | Ws]
	  end,
    NWs = [{NewSlotPos, <<FirstNew:32>>} | Ws1],
    insert_new(Rs, Bins, Ns, NWs);
insert_new([], [], [], Ws) ->
    Ws.

zero(Fd, Fname, I) ->
    fwrite(Fd, Fname, make_zeros(4*I)).

make_zeros(0) -> [];
make_zeros(N) when N rem 2 == 0 ->
    P = make_zeros(N div 2),
    [P|P];
make_zeros(N) ->
    P = make_zeros(N div 2),
    [0,P|P].

%% Read term from file at position Pos
prterm(Head, Pos, ReadAhead) ->
    Res = pread(Head, Pos, ?OHDSZ, ReadAhead),
    ?DEBUGF("file:pread(~p, ~p, ?) -> ~p~n", [Head#head.filename, Pos, Res]),
    {ok, <<Next:32, Sz:32, _Status:32, Bin0/binary>>} = Res,
    ?DEBUGF("{Next, Sz} = ~p~n", [{Next, Sz}]),
    Bin = case size(Bin0) of
	      Actual when Actual >= Sz ->
		  Bin0;
	      _ ->
		  {ok, Bin1} = pread(Head, Pos +  ?OHDSZ, Sz, 0),
		  Bin1
	  end,
    Term = binary_to_term(Bin),
    {Next, Sz, Term}.

%% Implement the ets match algorithm in erlang itself
%% return false | {true, Bindings}
%% Max 10 variables :-(

do_match(Obj, Pat) ->
    case do_match(Obj, Pat, {no,no,no,no,no,no,no,no,no,no}) of
	false -> false;
	{bs, Bs} -> {true, Bs}
    end.

add_binding(Pos, Bs, Obj) ->
    case catch setelement(Pos, Bs, Obj) of
	{'EXIT', _} ->
	    Bs2 = grow_tuple(Bs, size(Bs)),
	    add_binding(Pos, Bs2, Obj);
	Other -> 
	    Other
    end.

grow_tuple(Tup, Sz) ->
    L = tuple_to_list(Tup),
    L2 = lists:duplicate(Sz + 10, no),
    list_to_tuple(L ++ L2).

binding(Pos, Tup) ->
    case catch element(Pos, Tup) of
	{'EXIT', _} ->
	    Bs2 = grow_tuple(Tup, size(Tup)),
	    {no, Bs2};
	Other -> Other
    end.

do_match(X, X, Bs) -> 
    {bs, Bs};
do_match([H1 | T1], [H2 | T2], Bs) -> 
    case do_match(H1, H2, Bs) of
	{bs, Bs2} -> do_match(T1, T2, Bs2);
	false -> false
    end;
do_match(Tup1, Tup2, Bs) when tuple(Tup1),tuple(Tup2), 
                              size(Tup1) == size(Tup2) ->
    e_match(Tup1, Tup2, size(Tup1), Bs);
do_match(_Obj, '_', Bs) -> {bs, Bs}; % optimization
do_match(Obj, Pat, Bs) -> 
    case is_var(Pat) of
	{true, Pos} when integer(Pos) ->
	    case binding(Pos + 1, Bs) of
		no ->
		    {bs, add_binding(Pos + 1, Bs, Obj)};
		{no, Bs2} ->
		    {bs, add_binding(Pos + 1, Bs2, Obj)};
		Obj ->
		    {bs, Bs};
		_ ->
		    false
	    end;
	{true, wild} ->
	    {bs, Bs};
	false ->
	    false
    end.

e_match(_, _, 0, Bs) -> {bs, Bs};
e_match(T1, T2, Pos, Bs) ->
    case do_match(element(Pos, T1), element(Pos, T2), Bs) of
	{bs, Bs2} -> e_match(T1, T2, Pos-1, Bs2);
	false -> false
    end.

is_var(X) when atom(X) ->
    case atom_to_list(X) of
	[$$, Dig] when $0 =< Dig, Dig =< $9 -> {true, Dig - $0};
	[$_] -> {true, wild};
	[$$, Dig | Tail] -> accumulate_digs(lists:reverse([Dig |Tail]), 0, 1);
	_ -> false
    end;
is_var(_) -> false.

accumulate_digs([], Ack, _) -> {true, Ack};
accumulate_digs([Dig|T], Ack, Pow) when $0 =< Dig, Dig =< $9 ->
    accumulate_digs(T, Ack + (Dig - $0) * Pow, Pow * 10);
accumulate_digs(_,_,_) -> false.

contains_variable('_') ->
    true;
contains_variable(A) when atom(A) ->
    case atom_to_list(A) of
	[$$ | T] ->
	    case (catch list_to_integer(T)) of
		{'EXIT', _} ->
		    false;
		_ ->
		    true
	    end;
	_ ->
	    false
    end;
contains_variable(T) when tuple(T) ->
    contains_variable(tuple_to_list(T));
contains_variable([]) ->
    false;
contains_variable([H|T]) ->
    case contains_variable(H) of
	true ->
	    true;
	false ->
	    contains_variable(T)
    end;
contains_variable(_) ->
    false.

%%%%%%%%%%  server code %%%%%%%%%%%

stop() ->
    dets_server:stop().

do_open_file(Fname, Verbose) ->
    process_flag(trap_exit, true),
    case catch fopen(Fname) of
	{error, Reason} ->
	    Error = err({error, Reason}),
	    ?SERVER_NAME ! {self(), Error},
	    exit(normal);
	{ok, Head} ->
	    ?SERVER_NAME ! {self(), {ok, Head#head.name}},
	    maybe_put(verbose, Verbose),
	    open_file_loop(Head, 0);
	{'EXIT', _Reason} = Error ->
	    ?SERVER_NAME ! {self(), Error},
	    exit(normal)
    end.

do_open_file(Tab, Fname, Type, Kp, Rep, Verb, Est, Ram, CacheSz, Auto, Acc) ->
    ?PROFILE(ep:do()),
    process_flag(trap_exit, true),
    case catch fopen(Tab, Fname, Type, Kp, Rep, Est, Ram, CacheSz, Auto,Acc) of
	{error, {tooshort, _}} ->
	    file:delete(Fname),
	    do_open_file(Tab, Fname, Type, Kp, Rep, Verb, Est, Ram, CacheSz,
			 Auto, Acc);
	{error, Reason} ->
	    Error = err({error, Reason}),
	    ?SERVER_NAME ! {self(), Error},
	    exit(normal);
	{ok, Head} ->
	    ?SERVER_NAME ! {self(), {ok, Tab}},
	    maybe_put(verbose, Verb),
	    open_file_loop(Head, 0);
	{'EXIT', _Reason} = Error ->
	    ?SERVER_NAME ! {self(), Error},
	    exit(normal)
    end.

maybe_put(_, undefined) ->
    ignore;
maybe_put(K, V) ->
    put(K, V).

open_file_loop(Head, N) when record(Head, head), integer(N) ->
    receive
	{From, Op} ->
	    case catch apply_op(Op, From, Head, N) of
		ok -> 
		    open_file_loop(Head, N);
		{'EXIT', normal} ->
		    exit(normal);
		{'EXIT', Reason} = Exit ->
		    error_logger:format
		      ("** dets: table ~p caught EXIT with reason ~p~n",
		       [Head#head.name, Reason]),
		    From ! {self(), Exit},
		    open_file_loop(Head, N);
		{N2, H2} ->
		    open_file_loop(H2, N2);
		H2 ->
		    open_file_loop(H2, N)
	    end
    end.

apply_op(Op, From, Head, N) ->
    case Op of
        {initialize, InitFun} ->
            {H2, Res} = finit(Head, InitFun),
            From ! {self(), Res},
            H2;
	{lookup_keys, Keys} ->
	    {H2, Res} = flookup_keys(Head, Keys),
	    From ! {self(), Res},
	    H2;
	{member, Key} ->
	    {H2, Res} = fmember(Head, Key),
	    From ! {self(), Res},
	    H2;
	{insert, Objs} when Head#head.update_mode =/= saved ->
	    {H2, Res} = finsert(Head, Objs),
	    From ! {self(), Res},
	    {N + 1, H2};
	{add_user, Tab, Fname, Type, Keypos, _, Ram, _CacheSz, _Auto, Access}->
	    Res = if
		      Tab == Head#head.name,
		      Head#head.keypos == Keypos,
		      Head#head.type == Type,
		      Head#head.ram_file == Ram,
		      Head#head.access == Access,
		      Fname == Head#head.filename ->
			  {ok, Tab};
		      true ->
			  err({error, incompatible_arguments})
		  end,
	    From ! {self(), Res},
	    ok;
	{delete_key, Key} when Head#head.update_mode =/= saved ->
	    {H2, Res} = fdelete_key(Head, Key),
	    From ! {self(), Res},
	    {N + 1, H2};
	{delete_object, Object} when Head#head.update_mode =/= saved ->
	    {H2, Res} = fdelete_object(Head, Object),
	    From ! {self(), Res},
	    {N + 1, H2};
	delete_all_objects ->
	    {H2, Res} = fdelete_all_objects(Head),
	    From ! {self(), Res},
	    {0, H2};
	close  ->
	    From ! {self(), {closed, fclose(Head)}},
	    ?PROFILE(ep:done()),
	    exit(normal);
	first ->
	    {H2, Res} = ffirst(Head),
	    From ! {self(), Res},
	    H2;
	{next, Key} ->
	    {H2, Res} = fnext(Head, Key),
	    From ! {self(), Res},
	    H2;
	{match, MP, Spec, Limit} ->
	    {H2, Res} = fselect(Head, MP, Spec, Limit),
	    From ! {self(), Res},
	    H2;
	{match, State} ->
	    {H2, Res} = fmatch(Head, State),
	    From ! {self(), Res},
	    H2;
	{init_match_delete, MP, Spec} when Head#head.update_mode =/= saved ->
	    {H2, Res} = fmatch_delete_begin(Head, MP, Spec),
	    From ! {self(), Res},
	    {N + 1, H2};
	{match_delete, State} when Head#head.update_mode =/= saved ->
	    {H2, Res} = fmatch_delete(Head, State),
	    From ! {self(), Res},
	    {N + 1, H2};
	{slot, Slot} ->
	    {H2, Res} = fslot(Head, Slot),
	    From ! {self(), Res},
	    H2;
	{update_counter, Key, Incr} when Head#head.update_mode =/= saved ->
	    {NewHead, Res} = do_update_counter(Head, Key, Incr),
	    From ! {self(), Res},
	    {N + 1, NewHead};
	info ->
	    {H2, Res} = finfo(Head),
	    From ! {self(), Res},
	    H2;
	{info, Tag} ->
	    {H2, Res} = finfo(Head, Tag),
	    From ! {self(), Res},
	    H2;
	write_cache ->
	    {H2, Res0} = (catch write_cache(Head)),
	    Res = if Res0 == [] -> ok; true -> Res0 end,
	    From ! {self(), Res},
	    H2;
	{delayed_write, WrTime} ->
	    delayed_write(Head, WrTime);
	sync ->
	    {NewHead, Res} = perform_save(Head),
	    From ! {self(), Res},
	    {0, NewHead};
	{set_verbose, What} ->
	    set_verbose(What), 
	    From ! {self(), ok},
	    ok;
        wait_fixtable ->
            From ! {self(), ok},
            wait_fixtable(From, Head);
	{fixtable, Fixed} ->
	    NewHead = do_fixtable(Head, Fixed),
	    From ! {self(), ok},
	    NewHead;
	may_grow when Head#head.update_mode =/= saved ->
	    %% Won't grow more if the table is full.
	    {H2, _Res} = may_grow(Head, 0, many_times),
	    {N + 1, H2};
	{corrupt, Reason} ->
	    {H2, Error} = corrupt_reason(Head, Reason),
	    From ! {self(), Error},
	    H2;
	{where, Object} ->
	    {H2, Res} = where_is_object(Head, Object),
	    From ! {self(), Res},
	    H2;
	auto_save ->
	    case Head#head.update_mode of
		saved ->
		    Head;
		dirty when N == 0 ->
		    %% The updates seems to have declined
		    vformat("** dets: Auto save of ~p\n", [Head#head.name]), 
		    %% Ignore errors.
		    {NewHead, _Res} = perform_save(Head),
		    {0, NewHead};
		dirty -> 
		    %% Reset counter and try later
		    start_auto_save_timer(Head#head.auto_save),
		    {0, Head};
		{error, _Reason} ->
		    Head
	    end;

	WriteOp when Head#head.access == read_write,
		     Head#head.update_mode == saved ->
	    case catch mark_dirty(Head) of
		ok ->
		    start_auto_save_timer(Head#head.auto_save),
		    H2 = Head#head{update_mode = dirty},
		    apply_op(WriteOp, From, H2, 0);
		{NewHead, Error} ->
		    From ! {self(), Error},
		    NewHead
	    end;

	WriteOp when tuple(WriteOp), Head#head.access == read ->
	    Reason = {access_mode, Head#head.filename},
	    From ! {self(), err({error, Reason})},
	    ok
    end.

%% The (very long) time that fixtable_server waits for a reply from
%% dets:fixtable/2 may not be enough. Synchronization should help.
wait_fixtable(From, Head) ->
    receive 
        {From, {fixtable, Fixed}} ->
            NewHead = do_fixtable(Head, Fixed),                    
            From ! {self(), ok},
            wait_fixtable(From, NewHead);
        {From, fixed_ok} ->
            From ! {self(), ok},
            Head
    end.

mark_dirty(Head) ->
    mark_not_closed(Head),
    position(Head, Head#head.freelists_p),
    truncate(Head).

mark_not_closed(Head) ->
    pwrite(Head, [{?CLOSED_PROPERLY_POS, <<?NOT_PROPERLY_CLOSED:32>>}]),
    file:sync(Head#head.fptr).

set_verbose(true) ->
    put(verbose, yes);
set_verbose(_) ->
    erase(verbose).

do_update_counter(Head, _Key, _Incr) when Head#head.type =/= set ->
    {Head, badarg};
do_update_counter(Head, Key, Incr) ->
    case flookup_keys(Head, [Key]) of
	{H1, [O]} ->
	    Kp = H1#head.keypos,
	    case catch try_update_tuple(O, Kp, Incr) of
		{'EXIT', _} ->
		    {H1, badarg};
		{New, Term} ->
		    case finsert(H1, [Term]) of
			{H2, ok} -> 
			    {H2, New};
			Reply -> 
			    Reply
		    end
	    end;
	{H1, []} ->
	    {H1, badarg};
	HeadError ->
	    HeadError
    end.
    
try_update_tuple(O, _Kp, {Pos, Incr}) ->
    try_update_tuple2(O, Pos, Incr);
try_update_tuple(O, Kp, Incr) ->
    try_update_tuple2(O, Kp+1, Incr).

try_update_tuple2(O, Pos, Incr) ->
    New = element(Pos, O) + Incr,
    {New, setelement(Pos, O, New)}.

start_auto_save_timer(infinity) ->
    ok;
start_auto_save_timer(Millis) ->
    erlang:send_after(Millis, self(), {self(), auto_save}).

where_is_object(Head, Object) ->
    Keypos = Head#head.keypos,
    case check_objects([Object], Keypos) of
	true ->
	    case catch write_cache(Head) of
		{NewHead, []} ->
		    Key = element(Keypos, Object),
		    Slot = db_hash(Key, NewHead),
		    {NewHead, find_object(NewHead, Object, Slot)};
		HeadError ->
		    HeadError
	    end;
	false ->
	    {Head, badarg}
    end.

find_object(H, _Obj, Slot) when Slot >= H#head.next ->
    false;
find_object(H, Obj, Slot) ->
    {_Pos, Chain} = chain(H, Slot),
    case catch find_obj(H, Obj, Chain) of
	{ok, Pos} ->
	    {ok, Pos};
	_Else ->
	    false
    end.

find_obj(H, Obj, Pos) when Pos > 0 ->
    {Next, _Sz, Term} = prterm(H, Pos, ?ReadAhead),
    if 
	Term == Obj ->
	    {ok, Pos};
	true ->
	    find_obj(H, Obj, Next)
    end.

%%%%%  client functions %%%%

%% Assuming that a file already exists, open it with the
%% parameters as already specified in the file itself.
%% Return a ref leading to the file.
open_file(File) ->
    einval(dets_server:open_file(File), [File]).

open_file(Tab, Args) ->     
    case catch defaults(Tab, Args) of
	[{file, File}, 
	 {type, Type}, 
	 {keypos, KeyPos}, 
	 {repair, Rep}, 
	 {estimated_no_objects, Est}, 
	 {ram_file, Ram},
	 {delayed_write, CacheSz},
	 {auto_save, Auto},
	 {access, Acc}] ->
            einval(dets_server:open_file(Tab, File, Type, KeyPos, Rep, 
                                         Est, Ram, CacheSz, Auto, Acc),
		   [Tab, Args]);
	_ ->
	    erlang:fault(badarg, [Tab, Args])
    end.

init_table(Tab, InitFun) when function(InitFun) ->
    case treq(Tab, {initialize, InitFun}) of
        {thrown, Thrown} -> throw(Thrown);
        Else -> Else
    end;
init_table(Tab, InitFun) ->
    erlang:fault(badarg, [Tab, InitFun]).

from_ets(DTab, ETab) ->
    ets:safe_fixtable(ETab, true),
    Spec = ?PATTERN_TO_OBJECT_MATCH_SPEC('_'),
    LC = ets:select(ETab, Spec, 100),
    InitFun = from_ets_fun(LC, ETab),
    Reply = treq(DTab, {initialize, InitFun}),
    ets:safe_fixtable(ETab, false),
    case Reply of 
        {thrown, Thrown} -> throw(Thrown);
        Else -> Else
    end.

from_ets_fun(LC, ETab) ->
    fun(close) ->
            ok;
       (read) when LC == '$end_of_table' ->
            end_of_input;
       (read) ->
            {L, C} = LC,
            {L, from_ets_fun(ets:select(C), ETab)}
    end.

close(Tab) ->  
    dets_server:close(Tab).

lookup(Tab, Key) ->
    badarg(treq(Tab, {lookup_keys, [Key]}), [Tab, Key]).

lookup_keys(Tab, Keys) ->
    case check_list(Keys) of
	true -> badarg(treq(Tab, {lookup_keys, Keys}), [Tab, Keys]);
	false -> erlang:fault(badarg, [Tab, Keys])
    end.

member(Tab, Key) ->
    badarg(treq(Tab, {member, Key}), [Tab, Key]).    

insert(Tab, Objs) when list(Objs) ->
    badarg(treq(Tab, {insert, Objs}), [Tab, Objs]);
insert(Tab, Obj) ->
    badarg(treq(Tab, {insert, [Obj]}), [Tab, Obj]).

sync(Tab) ->
    badarg(treq(Tab, sync), [Tab]).

match_object(Tab, Pat) ->
    badarg(safe_match(Tab, Pat, object), [Tab, Pat]).

match_object(Tab, Pat, N) ->
    badarg(init_chunk_match(Tab, Pat, object, N), [Tab, Pat, N]).
    
match_object(State) when State#dets_cont.what == object ->
    chunk_match(State);
match_object(Term) ->
    erlang:fault(badarg, [Term]).

select(Tab, Pat) ->
    badarg(safe_match(Tab, Pat, select), [Tab, Pat]).

select(Tab, Pat, N) ->
    badarg(init_chunk_match(Tab, Pat, select, N), [Tab, Pat, N]).
    
select(State) when State#dets_cont.what == select ->
    chunk_match(State);
select(Term) ->
    erlang:fault(badarg, [Term]).

match(Tab, Pat) ->
    badarg(safe_match(Tab, Pat, bindings), [Tab, Pat]).

match(Tab, Pat, N) ->
    badarg(init_chunk_match(Tab, Pat, bindings, N), [Tab, Pat, N]).
    
match(State) when State#dets_cont.what == bindings ->
    chunk_match(State);
match(Term) ->
    erlang:fault(badarg, [Term]).

select_delete(Tab, Pat) ->
    badarg(match_delete(Tab, Pat, select), [Tab, Pat]).

match_delete(Tab, Pat) ->
    badarg(match_delete(Tab, Pat, object), [Tab, Pat]).

match_delete(Tab, Pat, What) ->
    safe_fixtable(Tab, true),    
    case compile_match_spec(What, Pat) of
	{Spec, MP} ->
	    Proc = dets_server:get_pid(Tab),
	    R = req(Proc, {init_match_delete, MP, Spec}),
	    do_match_delete(Tab, Proc, R);
	badarg ->
	    badarg
    end.

do_match_delete(Tab, _Proc, {done, R}) ->
    safe_fixtable(Tab, false),
    R;
do_match_delete(Tab, Proc, {cont, State}) ->
    do_match_delete(Tab, Proc, req(Proc, {match_delete, State})).

delete(Tab, Key) ->
    badarg(treq(Tab, {delete_key, Key}), [Tab, Key]).

to_ets(DTab, ETab) ->
    case ets:info(ETab, protection) of
	undefined ->
	    erlang:fault(badarg, [DTab, ETab]);
        _ ->
	    Fun = fun(X, T) -> true = ets:insert(T, X), T end,
	    foldl(Fun, ETab, DTab)
    end.

traverse(Tab, Fun) ->
    Ref = make_ref(),
    TFun = 
	fun(O, Ack) ->
		case Fun(O) of
		    continue  ->
			Ack;
		    {continue, Val} ->
			[Val | Ack];
		    {done, Value} ->
			throw({Ref, [Value | Ack]});
		    Other ->
			throw({Ref, Other})
		end
	end,
    do_traverse(TFun, [], Tab, Ref).

foldr(Fun, Acc, Tab) ->
    foldl(Fun, Acc, Tab).

foldl(Fun, Acc, Tab) ->
    Ref = make_ref(),
    do_traverse(Fun, Acc, Tab, Ref).

do_traverse(Fun, Acc, Tab, Ref) ->
    safe_fixtable(Tab, true),    
    Proc = dets_server:get_pid(Tab),
    R = (catch do_trav(Proc, Acc, Fun, Ref)),
    safe_fixtable(Tab, false),
    case R of
	{Ref, Result} ->
	    Result;
	{'EXIT', Reason} ->
	    erlang:fault(Reason);
	Thrown ->
	    throw(Thrown)
    end.

do_trav(Proc, Acc, Fun, Ref) ->
    {Spec, MP} = compile_match_spec(object, '_'),
    case req(Proc, {match, MP, Spec, default}) of
	{done, {error, Reason}} ->
	    {Ref, {error, Reason}};
	{cont, State} ->
	    do_trav(State, Proc, Acc, Fun, Ref)
    end.
    
do_trav(#dets_cont{pos = eof}, _Proc, Acc, _Fun, Ref) ->
    {Ref, Acc};
do_trav(State, Proc, Acc, Fun, Ref) ->
    case req(Proc, {match, State}) of
	{cont, {Bins, NewState}} ->
	    do_trav_bins(NewState, Proc, Acc, Fun, Ref, Bins);
	Error ->
	    {Ref, Error}
    end.

do_trav_bins(State, Proc, Acc, Fun, Ref, []) ->
    do_trav(State, Proc, Acc, Fun, Ref);
do_trav_bins(State, Proc, Acc, Fun, Ref, [Bin | Bins]) ->
    %% Unpack one binary at a time, using the client's heap.
    case catch binary_to_term(Bin) of 
	{'EXIT', _} ->
	    {Ref, req(Proc, {corrupt, bad_object})};
	Term ->
	    NewAcc = Fun(Term, Acc),
	    do_trav_bins(State, Proc, NewAcc, Fun, Ref, Bins)
    end.

delete_object(Tab, O) ->
    badarg(treq(Tab, {delete_object, O}), [Tab, O]).

delete_all_objects(Tab) ->
    badarg(treq(Tab, delete_all_objects), [Tab]).

first(Tab) ->
    badarg(treq(Tab, first), [Tab]).

next(Tab, Key) ->
    badarg(treq(Tab, {next, Key}), [Tab, Key]).

update_counter(Tab, Key, C) ->
    badarg(treq(Tab, {update_counter, Key, C}), [Tab, Key, C]).

slot(Tab, Slot) ->
    badarg(treq(Tab, {slot, Slot}), [Tab, Slot]).

info(Tab) ->
    case catch dets_server:get_pid(Tab) of
	{'EXIT', _Reason} ->
	    undefined;
	Pid ->
	    req(Pid, info)
    end.

info(Tab, owner) ->
    case catch dets_server:get_pid(Tab) of
	Pid when pid(Pid) ->
	    Pid;
	_ ->
	    undefined
    end;
info(Tab, safe_fixed) ->
    case catch dets_server:get_pid(Tab) of
	{'EXIT', _Reason} ->
	    undefined;
	_Pid ->
	    fixtable_server:info(?SERVER_NAME, Tab)
    end;
info(Tab, Tag) ->
    case catch dets_server:get_pid(Tab) of
	{'EXIT', _Reason} ->
	    undefined;
	Pid ->
	    req(Pid, {info, Tag})
    end.

fixtable(Tab, Bool) when Bool == true; Bool == false ->
    badarg(treq(Tab, {fixtable, Bool}), [Tab, Bool]);
fixtable(Tab, Term) ->
    erlang:fault(badarg, [Tab, Term]).

safe_fixtable(Tab, Bool) when Bool == true; Bool == false ->
    case catch dets_server:get_pid(Tab) of
	Pid when pid(Pid) ->
            case req(Pid, wait_fixtable) of
                ok ->
		    fixtable_server:safe_fixtable(?MODULE, Tab, Bool),
                    req(Pid, fixed_ok);
                Error -> 
                    Error
            end;
	_ ->
	    erlang:fault(badarg, [Tab, Bool])
    end;
safe_fixtable(Tab, Term) ->
    erlang:fault(badarg, [Tab, Term]).

all() ->
    dets_server:all().

pid2name(Pid) ->
    dets_server:pid2name(Pid).

verbose() ->           
    verbose(true).

verbose(What) ->
    ok = dets_server:verbose(What),
    All = dets_server:all(),
    Fun = fun(Tab) -> treq(Tab, {set_verbose, What}) end,
    lists:foreach(Fun, All),
    All.

safe_match(Tab, Pat, What) ->
    safe_fixtable(Tab, true),
    R = do_safe_match(init_chunk_match(Tab, Pat, What, default), []),
    safe_fixtable(Tab, false),
    R.
    
do_safe_match({error, Error}, _L) ->
    {error, Error};
do_safe_match({L, C}, LL) ->
    do_safe_match(chunk_match(C), L++LL);
do_safe_match('$end_of_table', L) ->
    L;
do_safe_match(badarg, _L) ->
    badarg.

%% What = object | bindings | select
init_chunk_match(Tab, Pat, What, N) when integer(N), N >= 0; N == default ->
    case compile_match_spec(What, Pat) of
	{Spec, MP} ->
	    case req(dets_server:get_pid(Tab), {match, MP, Spec, N}) of
		{done, {error, Reason}} ->
		    {error, Reason};
		{done, L} ->
		    {L, #dets_cont{what = What, pos = eof}};
		{cont, State} ->
		    chunk_match(State#dets_cont{what = What, tab = Tab})
	    end;
	badarg ->
	    badarg
    end;
init_chunk_match(_Tab, _Pat, _What, _) ->
    badarg.

chunk_match(#dets_cont{pos = eof}) ->
    '$end_of_table';
chunk_match(State) ->
    Proc = dets_server:get_pid(State#dets_cont.tab),
    case req(Proc, {match, State}) of
	{cont, {Bins, NewState}} ->
	    Fun = NewState#dets_cont.arg,
	    foldl_bins(Bins, Fun, NewState, Proc, []);
	Error ->
	    Error
    end.

foldl_bins([], _Fun, State, _Proc, Terms) ->
    {Terms, State};
foldl_bins([Bin | Bins], Fun, State, Proc, Terms) ->
    case catch binary_to_term(Bin) of 
	{'EXIT', _} ->
	    req(Proc, {corrupt, bad_object});
	Term ->
	    NewTerms = Fun(Term, Terms),
	    foldl_bins(Bins, Fun, State, Proc, NewTerms)
    end.

check_list(L) ->
    case catch length(L) of
	{'EXIT', _} -> false;
	_ -> true
    end.

%% -> {Spec, binary()} | badarg
compile_match_spec(select, Spec) ->
    case catch ets:match_spec_compile(Spec) of
	X when binary(X) ->
	    {Spec, X};
	_ ->
	    badarg
    end;
compile_match_spec(object, Pat) ->
    compile_match_spec(select, ?PATTERN_TO_OBJECT_MATCH_SPEC(Pat));
compile_match_spec(bindings, Pat) ->
    compile_match_spec(select, ?PATTERN_TO_BINDINGS_MATCH_SPEC(Pat)).

treq(Tab, R) ->
    case catch dets_server:get_pid(Tab) of
	Pid when pid(Pid) ->
	    req(Pid, R);
	_ ->
	    badarg
    end.

req(Proc, R) -> 
    Proc ! {self(), R},
    receive 
	{Proc, Reply} -> 
	    Reply;
	{'EXIT', Proc, Reason} ->
	    exit(Reason)
    end.

%% Inlined.
einval({error, {file_error, _, einval}}, A) ->
    erlang:fault(badarg, A);
einval(Reply, _A) ->
    Reply.

%% Inlined.
badarg(badarg, A) ->
    erlang:fault(badarg, A);
badarg(Reply, _A) ->
    Reply.

%% Process the args list as provided to open_file/2
defaults(Tab, Args) ->
    Defaults = [{file, to_list(Tab)},
		{type, set},
		{keypos, 1},
		{repair, true}, 
		{estimated_no_objects, ?DEFAULT_MIN_NO_SLOTS},
		{ram_file, false},
		{delayed_write, ?DEFAULT_CACHE},
		{auto_save, timer:minutes(?DEFAULT_AUTOSAVE)},
		{access, read_write}],
    Fun = fun repl/2,
    lists:foldl(Fun, Defaults, Args).

to_list(T) when atom(T) -> atom_to_list(T);
to_list(T) -> T.

repl({file, File}, Defs) ->
    lists:keyreplace(file, 1, Defs, {file, to_list(File)});
repl({type, T}, Defs) ->
    mem(T, [set, bag, duplicate_bag]),
    lists:keyreplace(type, 1, Defs, {type, T});
repl({keypos, P}, Defs) when integer(P), P > 0 ->
    lists:keyreplace(keypos, 1, Defs, {keypos, P});
repl({repair, T}, Defs) ->
    mem(T, [true, false, force]),
    lists:keyreplace(repair, 1, Defs, {repair, T});
repl({ram_file, Bool}, Defs) ->
    mem(Bool, [true, false]),
    lists:keyreplace(ram_file, 1, Defs, {ram_file, Bool});
repl({delayed_write, default}, Defs) ->
    lists:keyreplace(delayed_write, 1, Defs, {delayed_write, ?DEFAULT_CACHE});
repl({delayed_write, {Delay,Size} = C}, Defs) when integer(Delay), Delay >= 0,
						   integer(Size), Size >= 0 ->
    lists:keyreplace(delayed_write, 1, Defs, {delayed_write, C});
repl({auto_save, Int}, Defs) when integer(Int), Int >= 0 ->
    lists:keyreplace(auto_save, 1, Defs, {auto_save, Int});
repl({auto_save, infinity}, Defs) ->
    lists:keyreplace(auto_save, 1, Defs, {auto_save, infinity});
repl({estimated_no_objects, I}, Defs)  ->
    positive_int_or_default(I),
    lists:keyreplace(estimated_no_objects, 1, Defs,{estimated_no_objects,I});
repl({access, A}, Defs) ->
    mem(A, [read, read_write]),
    lists:keyreplace(access, 1, Defs, {access, A});
repl({_, _}, _) ->
    exit(badarg).

positive_int_or_default(default) -> ?DEFAULT_MIN_NO_SLOTS;
positive_int_or_default(I) when I >= ?DEFAULT_MIN_NO_SLOTS -> I;
positive_int_or_default(I) when I > 0 -> ?DEFAULT_MIN_NO_SLOTS.

mem(X, L) ->
    case lists:member(X, L) of
	true -> true;
	false -> exit(badarg)
    end.

%%%%%%%%%%%%%%%%%  UTILITY functions %%%%%%%%%%%%%%%%

%%% The write cache is an ets table of {Key, [Item]} where Item is one
%%% of: delete_key, lookup, {delete_object,object()}, or {insert,
%%% object()}. Sequences of delete and insert operations are inserted
%%% in the cache without doing any file operations. When the cache is
%%% considered full, a lookup operation is requested, or after some
%%% delay (see below), the contents of the cache are written on the
%%% file, and the cache emptied.
%%%
%%% Data is not allowed to linger more than 'delay' milliseconds in
%%% the write cache. A delayed_write message is received when some
%%% datum has become too old. If 'wrtime' is equal to 'undefined',
%%% then the cache is empty and no such delayed_write message has been
%%% scheduled. Otherwise there is a delayed_write message scheduled,
%%% and the value of 'wrtime' is the time when the cache was last
%%% written, or when it was first updated after the cache was last
%%% written.

%% -> {Head, [LookedUpObject]} | throw({Head, Error})
%% Lookup of keys bypassing the cache.
direct_lookup(Head, Ks) ->
    F = fun(K) -> {K, {keep,lookup,[]}} end,
    WL = lists:map(F, Ks),
    eval_work_list(Head, WL).

%% -> {[Found], [KeyToLookUp]} | false
cache_lookup(Type, Ks, Head) ->
    C = Head#head.cache,
    T = C#cache.tbl,    
    cache_lookup(Type, Ks, T, [], []).

cache_lookup(Type, [K | Ks], T, Acc, Lu) ->
    case ets:lookup(T, K) of
	[] -> 
	    cache_lookup(Type, Ks, T, Acc, [K | Lu]);
	L ->
	    [WL | _I] = wl(L, Type),
	    case WL of
		{delete, skip, Objs} ->
		    NAcc = cache_found(Objs, Acc),
		    cache_lookup(Type, Ks, T, NAcc, Lu);
		_ ->
		    false
	    end
    end;
cache_lookup(_Head, [], _T, Acc, Lu) ->
    {Acc, Lu}.

cache_found([{Obj,-1}], Acc) ->
    [Obj | Acc];
cache_found([{_Obj,0} | Objs], Acc) ->
    cache_found(Objs, Acc);
cache_found([{Obj,N} | Objs], Acc) when N < 0 ->
    cache_found(Objs, lists:duplicate(-N, Obj) ++ Acc);
cache_found([], Acc) ->
    Acc.

%% When writing the cache, a 'work list' is first created:
%%   WorkList = [{Key, {Delete,Lookup,[Inserted]}}]
%%   Delete = keep | delete
%%   Lookup = skip | lookup
%%   Inserted = {object(), No}
%%   No = integer()
%% If No =< 0 then there will be -No instances of object() on the file
%% when the cache has been written. If No > 0 then No instances of
%% object() will be added to the file.
%% If Delete has the value 'delete', then all objects with the key Key
%% have been deleted. (This could be viewed as a shorthand for {Object,0}
%% for each object Object on the file not mentioned in some Inserted.)
%% If Lookup has the value 'lookup', all objects with the key Key will
%% be returned.
%%
%% -> {NewHead, [LookedUpObject]} | throw({NewHead, Error})
write_cache(Head) ->
    C = Head#head.cache,
    T = C#cache.tbl,
    case ets:info(T, size) of
	0 ->
	    {Head, []};
	_Sz ->
	    Type = Head#head.type,
	    %% NoInsertedKeys is an upper limit on the number of new keys.
	    {NoInsertedKeys, WL} = make_wl(ets:first(T), T, Type, 0, []),
	    Head1 = Head#head{cache = reset_cache(C)},
	    Return = {_NewHead, Reply}  = 
		case may_grow(Head1, NoInsertedKeys, once) of
		    {Head2, ok} ->
			eval_work_list(Head2, WL);
		    HeadError ->
			throw(HeadError)
		end,
	    case Reply of
		_ when list(Reply) ->
		    Return;
		_ ->
		    throw(Return)
	    end
    end.

%% FIXME. This is probably not optimal...
reset_cache(C) ->
    ets:match_delete(C#cache.tbl, '_'),
    case C#cache.wrtime of 
	undefined ->
	    C;
	_WrTime ->
	    C#cache{wrtime = now()}
    end.

new_cache({Delay, Size}) ->
    #cache{tbl = ets:new(dets_cache, [duplicate_bag]), 
	   wrtime = undefined,
	   delay = Delay,
	   size = Size}.

delete_cache(C) ->
    %% If ets:delete/1 were used, dead-lock can occur: the
    %% fixtable_server waits for fixtable/2 to return, but ets:delete/1
    %% waits for the fixtable server to reply.
    ets:db_delete(C#cache.tbl).

%% -> {Head, [LookedUpObject]} | throw({Head, Error})
update_cache(Head, KeysOrObjects, What) ->
    Cache = Head#head.cache,
    T = Cache#cache.tbl,
    F = if 
	    What == delete_key; What == lookup ->
		fun(K) -> {K, What} end;
	    What == delete_object; What == insert ->
		Kp = Head#head.keypos,
		fun(O) -> {element(Kp, O), {What, O}} end
	end,
    IsEmpty = ets:info(T, size) == 0,
    ok = cache_binary2(KeysOrObjects, F, T),
    case ets:info(T, memory) of
	Sz when Sz >= Cache#cache.size; What == lookup ->
	    %% The cache is considered full, or lookup.
	    write_cache(Head);
	_Sz when Cache#cache.wrtime == undefined ->
	    %% Empty cache. Schedule a delayed write.
	    Now = now(), Me = self(),
	    erlang:send_after(Cache#cache.delay, Me, 
			      {Me, {delayed_write, Now}}),
	    {Head#head{cache = Cache#cache{wrtime = Now}}, []};
	_Sz when true == IsEmpty ->
	    %% Empty cache that has been written after the currently
	    %% scheduled delayed write.
	    {Head#head{cache = Cache#cache{wrtime = now()}}, []};
	_Sz ->
	    %% Cache is not empty, delayed write has been scheduled.
	    {Head, []}
    end.

cache_binary2([E | Es], F, Tbl) ->
    true = ets:insert(Tbl, F(E)),
    cache_binary2(Es, F, Tbl);
cache_binary2([], _F, _Tbl) ->
    ok.

%% Called after some delay.
%% -> NewHead
delayed_write(Head, WrTime) ->
    Cache = Head#head.cache,
    case Cache#cache.wrtime of
	LastWrTime when LastWrTime == WrTime ->
	    %% The cache has not been emptied during the last delay.
	    case catch write_cache(Head) of
		{Head2, []} ->
		    Cache2 = Head2#head.cache,
		    NewCache = Cache2#cache{wrtime = undefined},
		    Head2#head{cache = NewCache};
		{NewHead, _Error} -> % Head.update_mode has been updated
		    NewHead
	    end;
	LastWrTime ->
	    %% The cache was emptied during the delay.
	    %% Has anything been written since then?
	    case ets:info(Cache#cache.tbl, size) of
		0 ->
		    %% No, further delayed write not needed.
		    NewCache = Cache#cache{wrtime = undefined},
		    Head#head{cache = NewCache};
		_ ->
		    %% Yes, schedule a new delayed write.
		    {MS1,S1,M1} = WrTime,
		    {MS2,S2,M2} = LastWrTime,
		    WrT = M1+1000000*(S1+1000000*MS1),
		    LastWrT = M2+1000000*(S2+1000000*MS2),
		    When = round((LastWrT - WrT)/1000), Me = self(),
		    erlang:send_after(When, Me, 
				      {Me, {delayed_write, LastWrTime}}),
		    Head
	    end
    end.

make_wl('$end_of_table', _T, _Type, N, WL) ->
    {N, WL};
make_wl(Key, T, Type, N, WL) ->
    L = ets:lookup(T, Key),
    [Cs | I] = wl(L, Type),
    make_wl(ets:next(T, Key), T, Type, N+I, [{Key,Cs} | WL]).
    
wl(L, Type) ->
    wl(L, Type, keep, skip, 0, []).

wl([{_Key, delete_key} | Cs], Type, _Del, Lookup, _I, _Objs) ->
    wl(Cs, Type, delete, Lookup, 0, []);
wl([{_Key, {delete_object, Object}} | Cs], Type, Del, Lookup, I, Objs) ->
    NObjs = lists:keydelete(Object, 1, Objs),
    wl(Cs, Type, Del, Lookup, I, [{Object,0} | NObjs]);
wl([{_Key, {insert, Object}} | Cs], Type, _Del, Lookup, _I, _Objs) 
                    when Type == set ->
    wl(Cs, Type, delete, Lookup, 1, [{Object,-1}]);
wl([{_Key, {insert, Object}} | Cs], Type, Del, Lookup, _I, Objs) ->
    NObjs = 
	case lists:keysearch(Object, 1, Objs) of
	    {value, {_, 0}} ->
		lists:keyreplace(Object, 1, Objs, {Object,-1});
	    {value, {_, _C}} when Type == bag -> % C == 1; C == -1
		Objs;
	    {value, {_, C}} when C < 0 -> % when Type == duplicate_bag
		lists:keyreplace(Object, 1, Objs, {Object,C-1});
	    {value, {_, C}} -> % when C > 0, Type == duplicate_bag
		lists:keyreplace(Object, 1, Objs, {Object,C+1});
	    false when Del == delete ->
		[{Object, -1} | Objs];
	    false ->
		[{Object, 1} | Objs]
	end,
    wl(Cs, Type, Del, Lookup, 1, NObjs);
wl([{_Key, lookup} | Cs], Type, Del, _Lookup, I, Objs) ->
    wl(Cs, Type, Del, lookup, I, Objs);
wl([], _Type, Del, Lookup, I, Objs) ->
    [{Del, Lookup, Objs} | I].

%% -> {Head, [LookedUpObject]} | throw({Head, Error})
eval_work_list(Head, WorkLists) ->
    SWLs = tag_with_slot(WorkLists, Head, []),
    R1 = sofs:relation(SWLs, 2),
    P1 = sofs:to_external(sofs:partition_family(1, R1)),
    {PerSlot, SlotPositions} = remove_slot_tag(P1, [], []),
    {ok, Bins} = pread(SlotPositions, Head),
    first_object(PerSlot, SlotPositions, Bins, Head, [], [], []).

tag_with_slot([{K,_} = WL | WLs], Head, L) ->
    tag_with_slot(WLs, Head, [{db_hash(K, Head), WL} | L]);
tag_with_slot([], _Head, L) ->
    L.

remove_slot_tag([{_S,SWLs} | SSWLs], Ls, SP) ->
    remove_slot_tag(SWLs, SSWLs, [], Ls, SP);
remove_slot_tag([], Ls, SP) ->
    {Ls, SP}.

remove_slot_tag([{S,WL}], SSWLs, L, Ls, SP) ->
    remove_slot_tag(SSWLs, [{S,[WL|L]} | Ls], [slot_position(S) | SP]);
remove_slot_tag([{_S,WL} | SLWs], SSWLs, L, Ls, SP) ->
    remove_slot_tag(SLWs, SSWLs, [WL | L], Ls, SP).

%% The initial chain pointers and the first object in each chain are
%% read "in parallel", that is, with one call to file:pread/2 (two
%% calls altogether). The following chain objects are read one by
%% one. This is a compromise: if the chains are long and threads are
%% active, it would be faster to keep a state for each chain and read
%% the objects of the chains in parallel, but the overhead would be
%% quite substantial.

first_object([SP | SPs], [{P1,_4} | Ss], [<<P2:32>> | Bs], Head,
	      ObjsToRead, ToRead, Ls) when P2 == 0 ->
    {_Slot,WorkLists} = SP,
    L0 = [{old,P1}],
    L = eval_slot(Head, ?ReadAhead, P2, WorkLists, L0),
    first_object(SPs, Ss, Bs, Head, ObjsToRead, ToRead, [L | Ls]);
first_object([SP | SPs], [{P1,_4} | Ss], [<<P2:32>> | Bs], Head, 
	      ObjsToRead, ToRead, Ls) ->
    {_Slot,WorkLists} = SP,
    E = {P1,P2,WorkLists},
    first_object(SPs, Ss, Bs, Head, 
		 [E | ObjsToRead], [{P2, ?ReadAhead} | ToRead], Ls);
first_object([], [], [], Head, ObjsToRead, ToRead, Ls) ->
    {ok, Bins} = pread(ToRead, Head),
    case catch eval_first(Bins, ObjsToRead, Head, Ls) of
	{ok, NLs} -> 
	    case create_writes(NLs, Head, [], [], 0) of
		{Head1, [], LUs, 0} ->
		    {Head1, LUs};
		{Head1, Ws, LUs, No} ->
		    {Head2, Ws2} = update_no_objects(Head1, Ws, No),
		    {NewHead, ok} = pwrite(Head2, Ws2),
		    {NewHead, LUs}
	    end;
	_ -> 
	    throw(corrupt_reason(Head, bad_object))
    end.

eval_first([<<Next:32, Sz:32, _Status:32, Bin/binary>> | Bins], 
	   [SP | SPs], Head, Ls) ->
    {P1, P2, WLs} = SP,
    L0 = [{old,P1}],
    case size(Bin) of
	BinSz when BinSz >= Sz ->
	    Term = binary_to_term(Bin),
	    Key = element(Head#head.keypos, Term),
	    L = find_key(Head, P2, Next, Sz, Term, Key, WLs, L0),
	    eval_first(Bins, SPs, Head, [L | Ls]);
	_BinSz ->
	    L = eval_slot(Head, Sz+?OHDSZ, P2, WLs, L0),
	    eval_first(Bins, SPs, Head, [L | Ls])
    end;
eval_first([], [], _Head, Ls) ->
    {ok, Ls}.

eval_slot(_Head, _TrySize, Pos, [], L) when Pos == 0 ->
    L;
eval_slot(Head, _TrySize, Pos, [WL | WLs], L) when Pos == 0 ->
    {_Key, {_Delete, LookUp, Objects}} = WL,
    NL = end_of_key(Objects, LookUp, L),
    eval_slot(Head, ?ReadAhead, Pos, WLs, NL);
eval_slot(Head, TrySize, Pos, WLs, L) ->
    {NextPos, Size, Term} = prterm(Head, Pos, TrySize),
    Key = element(Head#head.keypos, Term),
    find_key(Head, Pos, NextPos, Size, Term, Key, WLs, L).

find_key(Head, Pos, NextPos, Size, Term, Key, WLs, L) ->
    case lists:keysearch(Key, 1, WLs) of
	{value, {_, {Delete, LookUp, Objects}} = WL} ->
	    NWLs = lists:delete(WL, WLs),
	    {NewObjects, NL} = eval_object(Size, Term, Delete, LookUp, 
					   Objects, Head, Pos, L),
	    eval_key(Key, Delete, LookUp, NewObjects, Head, NextPos,NWLs,NL);
	false ->
	    L0 = [{old,Pos} | L],
	    eval_slot(Head, ?ReadAhead, NextPos, WLs, L0)
    end.

eval_key(_Key, _Delete, _Lookup, _Objects, Head, Pos, WLs, L) 
                            when Head#head.type == set  ->
    eval_slot(Head, ?ReadAhead, Pos, WLs, L);
eval_key(_Key, _Delete, LookUp, Objects, Head, Pos, WLs, L) when Pos == 0 ->
    NL = end_of_key(Objects, LookUp, L),
    eval_slot(Head, ?ReadAhead, Pos, WLs, NL);
eval_key(Key, Delete, LookUp, Objects, Head, Pos, WLs, L) ->
    {NextPos, Size, Term} = prterm(Head, Pos, ?ReadAhead),
    case element(Head#head.keypos, Term) of
	Key ->
	    {NewObjects, NL} = eval_object(Size, Term, Delete, LookUp,Objects,
					   Head, Pos, L),
	    eval_key(Key, Delete, LookUp, NewObjects, Head, NextPos, WLs, NL);
	Key2 ->
	    L1 = end_of_key(Objects, LookUp, L),
	    find_key(Head, Pos, NextPos, Size, Term, Key2, WLs, L1)
    end.

%% All objects in Objects have the key Key.
eval_object(Size, Term, Delete, LookUp, Objects, Head, Pos, L) ->
    Type = Head#head.type,
    case lists:keysearch(Term, 1, Objects) of
	{value, {_Object, N}} when N == 0 ->
	    L1 = [{delete,Pos,Size} | L],
	    {Objects, L1};
	{value, {_Object, N}} when N < 0, Type == set ->
	    L1 = [{old,Pos} | L],
	    wl_lookup(LookUp, Objects, Term, L1);
	{value, {Object, _N}} when Type == bag -> % when N == 1; N == -1
	    L1 = [{old,Pos} | L],
	    Objects1 = lists:keydelete(Object, 1, Objects),
	    wl_lookup(LookUp, Objects1, Term, L1);
	{value, {Object, N}} when N < 0, Type == duplicate_bag ->
	    L1 = [{old,Pos} | L],
	    Objects1 = lists:keyreplace(Object, 1, Objects, {Object,N+1}),
	    wl_lookup(LookUp, Objects1, Term, L1);
	{value, {_Object, N}} when N > 0, Type == duplicate_bag ->
	    L1 = [{old,Pos} | L],
	    wl_lookup(LookUp, Objects, Term, L1);
	false when Type == set, Delete == delete, [] =/= Objects  ->
	    [{Term2,-1}] = Objects,
	    Bin2 = term_to_binary(Term2),
	    NSize = size(Bin2),
	    Overwrite = 
		if
		    NSize == Size ->
			true;
		    true ->
			SizePos = sz2pos(Size+?OHDSZ),
			NSizePos = sz2pos(NSize+?OHDSZ),
			SizePos == NSizePos
		end,
	    E = if 
		    Overwrite == true ->
			{overwrite,Bin2,Pos};
		    true ->
			{replace,Bin2,Pos,Size}
		end,
	    wl_lookup(LookUp, Objects, Term, [E | L]);
	false when Delete == delete ->
	    L1 = [{delete,Pos,Size} | L],
	    {Objects, L1};
	false ->
	    L1 = [{old,Pos} | L],
	    wl_lookup(LookUp, Objects, Term, L1)
    end.

%% Inlined.
wl_lookup(lookup, Objects, Term, L) ->
    NL = [{lookup,Term} | L],
    {Objects, NL};
wl_lookup(skip, Objects, _Term, L) ->
    {Objects, L}.

end_of_key([{Object,N0} | Objs], LookUp, L) when N0 =/= 0 ->
    N = abs(N0),
    L1 = [{insert,N,term_to_binary(Object)} | L],
    NL = if 
	     LookUp == lookup ->
		 lists:duplicate(N, {lookup,Object}) ++ L1;
	     true ->
		 L1
	 end,
    end_of_key(Objs, LookUp, NL);
end_of_key([_ | Objects], LookUp, L) ->
    end_of_key(Objects, LookUp, L);
end_of_key([], _LookUp, L) ->
    L.

create_writes([L | Ls], H, Ws, LUs, No) ->
    {NH, NWs, NLUs, NNo} = create_writes(L, H, Ws, LUs, No, 0, true),
    create_writes(Ls, NH, NWs, NLUs, NNo);
create_writes([], H, Ws, LUs, No) ->
    {H, lists:reverse(Ws), LUs, No}.

create_writes([{old,Pos} | L], H, Ws, LUs, No, _Next, true) ->
    create_writes(L, H, Ws, LUs, No, Pos, true);
create_writes([{old,Pos} | L], H, Ws, LUs, No, Next, false) ->
    W = {Pos, <<Next:32>>},
    create_writes(L, H, [W | Ws], LUs, No, Pos, true);
create_writes([{insert,N,Bin} | L], H, Ws, LUs, No, Next, _NextIsOld) ->
    {NH, NWs, Pos} = create_inserts(N, H, Ws, Next, size(Bin), Bin),
    create_writes(L, NH, NWs, LUs, No+N, Pos, false);
create_writes([{overwrite,Bin,Pos} | L], H, Ws, LUs, No, Next, _) ->
    Size = size(Bin),
    W = {Pos, [<<Next:32, Size:32, ?ACTIVE:32>>, Bin]},
    create_writes(L, H, [W | Ws], LUs, No, Pos, true);
create_writes([{replace,Bin,Pos,OSize} | L], H, Ws, LUs, No, Next, _) ->
    Size = size(Bin),
    H1 = free(H, Pos, OSize+?OHDSZ),
    {NH, NewPos} = alloc(H1, ?OHDSZ + Size),    
    W1 = {NewPos, [<<Next:32, Size:32, ?ACTIVE:32>>, Bin]},
    NWs = if 
	      Pos == NewPos -> 
		  [W1 | Ws];
	      true -> 
		  W2 = {Pos+?STATUS_POS, <<?FREE:32>>},
		  [W1,W2 | Ws]
	  end,
    create_writes(L, NH, NWs, LUs, No, NewPos, false);
create_writes([{delete,Pos,Size} | L], H, Ws, LUs, No, Next, _) ->
    NH = free(H, Pos, Size+?OHDSZ),
    NWs = [{Pos+?STATUS_POS,<<?FREE:32>>} | Ws],
    create_writes(L, NH, NWs, LUs, No-1, Next, false);
create_writes([{lookup,Term}| L], H, Ws, LUs, No, Pos, NextIsOld) ->
    create_writes(L, H, Ws, [Term | LUs], No, Pos, NextIsOld);
create_writes([], H, Ws, LUs, No, _Next, _NextIsOld) ->
    {H, Ws, LUs, No}.

create_inserts(0, H, Ws, Next, _Size, _Bin) ->
    {H, Ws, Next};
create_inserts(N, H, Ws, Next, Size, Bin) ->
    {NH, Pos} = alloc(H, ?OHDSZ + Size),    
    W = {Pos, [<<Next:32, Size:32, ?ACTIVE:32>>, Bin]},
    create_inserts(N-1, NH, [W | Ws], Pos, Size, Bin).

slot_position(S) ->
    Pos = ?SEGADDR(S div ?SEGSZ),
    Segment = get_segp(Pos),
    FinalPos = Segment + (4 * (S rem ?SEGSZ)),
    {FinalPos, 4}.

%% Inlined.
get_segp(Pos) ->
    get(Pos).

rename(From, To) ->
    case file:rename(From, To) of
        ok ->
            ok;
        {error, Reason} ->
            {error, {file_error, {From, To}, Reason}}
    end.

%% -> {ok, Bins} | throw({NewHead, Error})
pread(Positions, Head) ->
    R = case file:pread(Head#head.fptr, Positions) of
	    {ok, Bins} ->
		%% file:pread/2 can return 'eof' as "data".
		case lists:member(eof, Bins) of
		    true ->
			{error, {premature_eof, Head#head.filename}};
		    false ->
			{ok, Bins}
		end;
	    {error, Reason} when enomem == Reason; einval == Reason ->
		{error, {bad_object_header, Head#head.filename}};
	    {error, Reason} ->
		{file_error, Head#head.filename, Reason}
	end,
    case R of
	{ok, _Bins} ->
	    R;
	Error ->
	    throw(corrupt(Head, Error))
    end.

%% -> {ok, binary()} | throw({NewHead, Error})
pread(Head, Pos, Min, Extra) ->
    R = case file:pread(Head#head.fptr, Pos, Min+Extra) of
	    {error, Reason} when enomem == Reason; einval == Reason ->
		{error, {bad_object_header, Head#head.filename}};
	    {error, Reason} ->
		{file_error, Head#head.filename, Reason};
	    {ok, Bin} when size(Bin) < Min ->
		{error, {premature_eof, Head#head.filename}};
	    OK -> OK
	end,
    case R of
	{ok, _Bin} ->
	    R;
	Error ->
	    throw(corrupt(Head, Error))
    end.
	    
%% -> {Head, ok} | throw({Head, Error})
pwrite(Head, Bins) ->
    case file:pwrite(Head#head.fptr, Bins) of
	ok ->
	    {Head, ok};
	{error, Reason} ->
	    Error = {error, {file_error, Head#head.filename, Reason}},
	    throw(corrupt(Head, Error))
    end.

%% -> {ok, Pos} | throw({Head, Error})
position(Head, Pos) ->
    case file:position(Head#head.fptr, Pos) of
	{error, Reason} -> 
	    FError = {error, {file_error, Head#head.filename, Reason}},
	    throw(corrupt(Head, FError));
	OK -> OK
    end.
	    
%% -> ok | throw({Head, Error})
truncate(Head) ->
    case file:truncate(Head#head.fptr) of
	ok    -> ok;
	{error, Reason} -> 
	    FError = {error, {file_error, Head#head.filename, Reason}},
	    throw(corrupt(Head, FError))
    end.
	    
open(FileSpec, Args) ->
    case file:open(FileSpec, Args) of
	{ok, Fd} ->
	    {ok, Fd};
	Error ->
	    file_error(FileSpec, Error)
    end.

fwrite(Fd, FileName, B) ->
    case file:write(Fd, B) of
	ok    -> ok;
	Error -> file_error_close(Fd, FileName, Error)
    end.

position(Fd, FileName, Pos) ->
    case file:position(Fd, Pos) of
	{error, Error} -> file_error(FileName, {error, Error});
	OK -> OK
    end.
	    
position_close(Fd, FileName, Pos) ->
    case file:position(Fd, Pos) of
	{error, Error} -> file_error_close(Fd, FileName, {error, Error});
	OK -> OK
    end.
	    
truncate(Fd, FileName) ->
    case file:truncate(Fd) of
	ok    -> ok;
	Error -> file_error(FileName, Error)
    end.

pwrite(Fd, FileName, Position, B) ->
    case file:pwrite(Fd, Position, B) of
	ok -> ok;
	Error -> file_error(FileName, {error, Error})
    end.

pwrite(Fd, FileName, Bins) ->
    case file:pwrite(Fd, Bins) of
	ok ->
	    ok;
	{error, {_NoWrites, Reason}} ->
	    file_error(FileName, {error, Reason})
    end.

pread_close(Fd, FileName, Pos, Size) ->
    case file:pread(Fd, Pos, Size) of
	{error, Error} ->
	    file_error_close(Fd, FileName, {error, Error});
	{ok, Bin} when size(Bin) < Size ->
	    file:close(Fd),
	    throw({error, {tooshort, FileName}});
	OK -> OK
    end.
	    
file_error(FileName, {error, Reason}) ->
    throw({error, {file_error, FileName, Reason}}).

file_error_close(Fd, FileName, {error, Reason}) ->
    file:close(Fd),
    throw({error, {file_error, FileName, Reason}}).
	    

%%% Scan the file from start to end by reading chunks.

%% -> dets_cont()
init_scan(Head, NoObjs) ->
    L = all_allocated(get_freelists(Head)),
    Pos = case L of
	      <<>> -> 
		  eof;
	      {From, _To, _L} ->
		  From
	  end,
    #dets_cont{no_objs = NoObjs, pos = Pos, bin = <<>>, alloc = L}.

%% -> {[RTerm], dets_cont()} | {scan_error, Reason}
%% RTerm = {Pos, Next, Size, Status, Term}
scan(_Head, _Repr, C) when size(C#dets_cont.alloc) == 0 ->
    {[], C};
scan(Head, Repr, C) -> % when record(C, dets_cont)
    #dets_cont{no_objs = No, pos = Pos, alloc = L0, bin = Bin} = C,
    F = Head#head.fptr,
    {From, To, L} = L0,
    R = case No of
	    default ->
		0;
	    _ when integer(No) ->
		-No-1
	end,
    scan(Bin, F, From, To, Pos, L, [], R, {Repr, C}).

scan(Bin, _F, From, To, Pos, L0, Ts, -1, {_Repr,C}) ->
    L = {From, To, L0},
    {Ts, C#dets_cont{pos = Pos, bin = Bin, alloc = L}};
scan(B = <<_N:32, Sz:32, St:32, T/binary>>, F, From, To, Pos, L, Ts, R, C) ->
    if 
	St == ?ACTIVE;
	St == ?FREE -> % deleted after scanning started
	    case T of
		<<BinTerm:Sz/binary, T2/binary>> ->
		    case catch to_object(BinTerm, C) of
			{'EXIT', _} ->
			    {scan_error, bad_object};
			Object ->
			    NTs = [Object | Ts],
			    OSz = Sz + ?OHDSZ,
			    Skip = ?POW(sz2pos(OSz)-1) - OSz,
                            F2 = From + OSz,
			    NR = if 
				     R < 0 ->
					 R + 1;
				     true ->
					 R + OSz + Skip
				 end,
			    scan_skip(T2, F, F2, To, Pos, Skip, L, NTs, NR, C)
		    end;
		_ ->
		    scan_read(B, F, From, To, Pos, Sz, L, Ts, R, C)
	    end;
	true -> % a segment
	    scan_skip(B, F, From, To, Pos, actual_seg_size(), L, Ts, R, C)
    end;
scan(B, F, From, To, Pos, L, Ts, R, C) ->
    scan_read(B, F, From, To, Pos, 0, L, Ts, R, C).
    
scan_skip(Bin, F, From, To, Pos, Skip, L, Ts, R, C) when From + Skip == To ->
    scan_next_allocated(Bin, F, From, Pos, L, Ts, R, C);
scan_skip(Bin, F, From, To, Pos, Skip, L, Ts, R, C) ->
    SkipPos = From + Skip,
    case Bin of
	<<_:Skip/binary, Tail/binary>> ->
	    scan(Tail, F, SkipPos, To, Pos, L, Ts, R, C);
	_ ->
	    scan_read(<<>>, F, SkipPos, To, SkipPos, 0, L, Ts, R, C)
    end.

scan_next_allocated(_Bin, _F, _From, _Pos, <<>> = EB, Ts, _R, {_Repr, C}) ->
    {Ts, C#dets_cont{pos = eof, bin = EB, alloc = EB}};
scan_next_allocated(Bin, F, From0, Pos, <<From:32, To:32, L/binary>>, 
		    Ts, R, C) ->
    Skip = From - From0,
    scan_skip(Bin, F, From0, To, Pos, Skip, L, Ts, R, C).

scan_read(Bin, _F, From, To, Pos, _Min, L0, Ts, 
	  R, {_Repr, C}) when R >= ?CHUNK_SIZE ->
    %% We may have read (much) more than CHUNK_SIZE, if there are holes.
    %% If Bin is non-empty, it contains the first part of _one_ term.
    L = {From, To, L0},
    {Ts, C#dets_cont{pos = Pos, bin = Bin, alloc = L}};
scan_read(B, F, From, To, Pos, Min, L, Ts, R, C) ->
    Max = if 
	      Min < ?CHUNK_SIZE -> ?CHUNK_SIZE; 
	      true -> Min 
	  end,
    case pread_n(F, Pos, Max) of
	eof ->
	    {scan_error, premature_eof};
	Bin ->
	    NewPos = Pos + size(Bin),
	    NewBin = if 
			 size(B) == 0 -> Bin;
			 true -> list_to_binary([B, Bin])
		     end,
	    scan(NewBin, F, From, To, NewPos, L, Ts, R, C)
    end.

%% Inlined.
to_object(BinTerm, {binary, _C}) ->
    BinTerm;
to_object(BinTerm, _C) -> 
    binary_to_term(BinTerm).

%% Update no_objects on the file too, if the number of segments that
%% fsck2/5 use for estimate has changed.
update_no_objects(Head, Ws, 0) -> {Head, Ws};
update_no_objects(Head, Ws, Delta) ->
    No = Head#head.no_objects,
    NewNo = No + Delta,
    NWs = 
	if 
	    NewNo > ?MAXOBJS ->
		Ws;
	    No div ?SEGSZ == NewNo div ?SEGSZ ->
		Ws;
	    true ->
		[{?NO_OBJECTS_POS, <<NewNo:32>>} | Ws]
	end,
    {Head#head{no_objects = NewNo}, NWs}.

%% Given, a slot, return the {Pos, Chain} in the file where the
%% objects hashed to this slot reside. Pos is the position in the
%% file where the chain pointer is written and Chain is the position
%% in the file where the first object resides.
chain(Head, Slot) ->
    Pos = ?SEGADDR(Slot div ?SEGSZ),
    Segment = get_segp(Pos),
    FinalPos = Segment + (4 * (Slot rem ?SEGSZ)),
    {ok, <<Chain:32>>} = pread(Head, FinalPos, 4, 0),
    {FinalPos, Chain}.

%% Twice the size of a segment due to the bug in sz2pos/1.
actual_seg_size() ->
    ?POW(sz2pos(?SEGSZ*4)-1).

segp_cache(Pos, Segment) ->
    put(Pos, Segment).

pread_obj_head(Fd, FileName, Pos) ->  %% check for possible eof here
    {ok, <<Next:32, Size:32, Status:32>>} =
	pread_close(Fd, FileName, Pos, ?OHDSZ),
    {Next, Size, Status}.

read_n(F, Max) ->
    case file:read(F, Max) of
	{ok, Bin} ->
	    Bin;
	_ ->
	    eof
    end.

pread_n(F, Position, Max) ->
    case file:pread(F, Position, Max) of
	{ok, Bin} ->
	    Bin;
	_ ->
	    eof
    end.

err(Error) ->
    case get(verbose) of
	yes -> 
	    error_logger:format("** dets: failed with ~w~n", [Error]),
	    Error;
	undefined  ->
	    Error
    end.

%%%%%%%%%%%%%%%%%  DEBUG functions %%%%%%%%%%%%%%%%

file_info(F) ->
    case catch read_file_header(F, read, false) of
	{ok, Fd, FH} ->
	    #fileheader{closed_properly = CP, keypos = Kp,
			m = M, next = Next, n = N, version = Version,
			type = Type, no_objects = NoObjects} 
		= FH,
	    R = if
		    CP == 0 ->
			{error, not_closed};
		    FH#fileheader.cookie /= ?MAGIC ->
			{error, not_a_dets_file};
		    FH#fileheader.version /= ?FILE_FORMAT_VERSION ->
			{error, bad_version};
		    true ->
			{ok, [{closed_properly,CP},{keypos,Kp},{m, M},
			      {n,N},{next,Next},{no_objects,NoObjects},
			      {type,Type},{version,Version}]}
		end,
	    file:close(Fd),
	    R;
	Other ->
	    Other
    end.

get_head_field(File, Field) ->
    {ok, _} = file:position(File, Field),
    read_4(File).

%% debug fun to inspect position Pos in an open file
inspect_chain(H, Pos) ->
    F = H#head.fptr,
    {ok, Pos} = file:position(F, Pos),
    case read_4(F) of
	0 -> 
	    0;
	I ->
	    {ok,  <<Sz:32, Status:32>>} = file:read(F, 8),
	    case Status of 
		?FREE -> 
		    {ok, Bin} = file:read(F, Sz),
		    {free, {next, I}, {obj, catch term_to_binary(Bin)}};
		?ACTIVE ->
		    {ok, Bin} = file:read(F, Sz),
		    {active, {next, I}, {obj, catch term_to_binary(Bin)}};
		_ ->
		    not_an_object
	    end
    end.

%% Dump the contents of a DAT file to the tty
%% internal debug function which ignores the closed properly thingie
%% and just tries anyway

view(Fn) ->
    case file:open(Fn, [raw, binary, read]) of
	{ok, F} ->
	    {ok, Fd, FH} = read_file_header(Fn, read, false),
	    {ok, H0, ExtraInfo} = check_file_header(FH, Fd),
	    Ftab = init_freelist(H0, ExtraInfo),
	    H = H0#head{freelists=Ftab},
	    v_free_list(H),
	    v_segments(H, 0),
	    file:close(F);
	X -> 
	    X
    end.

v_free_list(H) ->
    io:format("FREE LIST ...... \n",[]),
    io:format("~p~n", [all_free(get_freelists(H))]),
    io:format("END OF FREE LIST \n",[]).

v_segments(_H, ?SEGARRSZ) ->
    done;
v_segments(H, SegNo) ->
    {ok, _} = file:position(H#head.fptr, ?SEGADDR(SegNo)),
    Seg = read_4(H#head.fptr),
    if
	Seg == 0 ->
	    done;
	true ->
	    io:format("SEGMENT ~w ", [SegNo]),
	    io:format("At position ~w~n", [Seg]),
	    v_segment(H, SegNo, Seg, 0),
	    v_segments(H, SegNo+1)
    end.

v_segment(_H, _, _SegPos, ?SEGSZ) ->
    done;
v_segment(H, SegNo, SegPos, SegSlot) ->
    Slot = SegSlot + (SegNo * ?SEGSZ),
    {ok, _} = file:position(H#head.fptr, SegPos + (4 * SegSlot)),
    Chain = read_4(H#head.fptr),
    if 
	Chain == 0 ->  %% don't print empty chains
	    true;
	true ->
	    io:format("   <~p>~p: [",[SegPos + (4 * SegSlot), Slot]),
	    print_chain(H, Chain)
    end,
    v_segment(H, SegNo, SegPos, SegSlot+1).

print_chain(_H, 0) ->
    io:format("] \n", []);
print_chain(H, Pos) ->
    {ok, _} = file:position(H#head.fptr, Pos),
    case rterm(H#head.fptr) of
	{ok, 0, _Sz, Term} ->
	    io:format("<~p>~p] \n",[Pos, Term]);
	{ok, Next, _Sz, Term} ->
	    io:format("<~p>~p, ", [Pos, Term]),
	    print_chain(H, Next);
	Other ->
	    io:format("~nERROR ~p~n", [Other])
    end.

%% Can't be used at the bucket level!!!!
%% Only when we go down a chain
rterm(F) ->
    case catch rterm2(F) of
	{'EXIT', Reason} -> %% truncated DAT file 
	    vformat("** dets: Corrupt or truncated dets file~n", []), 
	    {error, Reason};
	Other -> 
	    Other
    end.

rterm2(F) ->
    {ok, <<Next:32, Sz:32, _:32>>} = file:read(F, ?OHDSZ),
    {ok, Bin} = file:read(F, Sz),
    Term = binary_to_term(Bin),
    {ok, Next, Sz, Term}.

%% Where in the (open) table is Object located? 
%% The address of the first matching object is returned.
%% -> {ok, Address} | false
where(Tab, Object) ->
    badarg(treq(Tab, {where, Object}), [Tab, Object]).

%% -> {NewHead, Error}
corrupt_reason(Head, Reason) ->
    Error = {error, {Reason, Head#head.filename}},    
    corrupt(Head, Error).

corrupt(Head, Error) ->
    case get(verbose) of
	yes -> 
	    error_logger:format("** dets: Corrupt table ~p: ~p\n", 
				[Head#head.name, Error]);
	_ -> ok
    end,
    case Head#head.update_mode of
	{error, _} ->
	    {Head, Error};
	_ ->
	    {Head#head{update_mode = Error}, Error}
    end.

vformat(F, As) ->
    case get(verbose) of
	yes -> error_logger:format(F, As);
	_ -> ok
    end.

read_4(File) ->
    {ok, <<Four:32>>} = file:read(File, 4),
    Four.


%%%%%%%%%%%%%%% allocation routines %%%%%%%%%%%%%%
%%% Algorithm : We use a buddy system on each file. This is nicely described
%%%             In i.e. the last chapter of the first-grade text book 
%%%             Data structures and algorithms by Aho, Hopcroft and
%%%             Ullman. I think buddy systems were invented by Knuth, a long
%%%             time ago.

init_freelist(Head, {convert_freelist,_Version}) ->
    %% This function converts the saved freelist of the form
    %% [{Slot1,Addr1},{Addr1,Addr2},...,{AddrN,0},{Slot2,Addr},...]
    %% i.e each slot is a linked list which ends with a 0.
    %% This is stored in a bplus_tree per Slot.
    %% Each Slot is a position in a tuple.

    #head{fptr = Fd, filename = Fname} = Head,
    Ftab = init_alloc(), % create tuple with empty bplus_trees
    Pos = Head#head.freelists_p,
    {0, Size, _Status} = pread_obj_head(Fd, Fname, Pos),
    {ok, B} = pread_close(Fd, Fname, Pos+?OHDSZ, Size),
    FreeList = lists:reverse(binary_to_term(B)),
    init_slots_from_old_file(FreeList, Ftab);
init_freelist(Head, _) ->
    Pos = Head#head.freelists_p,
    #head{fptr = Fd, filename = Fname} = Head,
    {0, Size, _Status} = pread_obj_head(Fd, Fname, Pos),
    {ok, B} = pread_close(Fd, Fname, Pos+?OHDSZ, Size),
    binary_to_term(B). % if we stored the bplus_tree as is

init_slots_from_old_file([{Slot,Addr} | T], Ftab) ->
    init_slot(Slot,[{Slot,Addr} | T], Ftab);
init_slots_from_old_file([], Ftab) ->
    Ftab.

init_slot(_Slot,[], Ftab) ->
    Ftab; % should never happen
init_slot(_Slot,[{_Addr,0}|T], Ftab) ->
    init_slots_from_old_file(T, Ftab);
init_slot(Slot,[{_Slot1,Addr}|T], Ftab) ->
    Stree = element(Slot, Ftab),
    %%    io:format("init_slot ~p:~p~n",[Slot, Addr]),
    init_slot(Slot,T,setelement(Slot, Ftab, bplus_insert(Stree, Addr))).

%%% The free lists are kept in RAM, and written to the end of the file
%%% from time to time. It is possible that a considerable amount of
%%% memory is used for a fragmented file.
%%%
%%% To make things (slightly) worse (from a memory usage point of
%%% view), each traversal of the file starts with making a "map" of
%%% the allocated areas; only the allocated areas will be
%%% traversed. Creating a map involves inspecting and sorting the free
%%% lists. Since the map is passed on between client and server, it
%%% has to be a binary (to avoid copying a possibly huge term).
%%%
%%% An active map should always be protected by fixing the table. This
%%% prevents insertion of objects into the mapped area (where some
%%% objects may have been deleted). The means for implementing this
%%% protection is a copy of the free lists (using even more memory, if
%%% objects are inserted). The position to write an inserted object is
%%% found by looking at the free lists from the time when the table
%%% was fixed; areas within the mapped area that have been freed are
%%% hidden from the allocator.

%% -> free_table()
%% A free table is a tuple of ?MAXBUD elements, element i handling
%% buddies of size 2**(i-1).
init_alloc() ->
    Empty = bplus_empty_tree(),
    %% initiate a tuple with ?MAXBUD "Empty" elements
    Ftab = erlang:make_tuple(?MAXBUD, Empty),
    setelement(?MAXBUD, Ftab, bplus_insert(Empty, ?BASE)). 

%% Only used when repairing.
alloc_many(Head, _Sz, 0, _A0) ->
    Head;
alloc_many(Head, Sz, N, A0) ->
    Ftab = Head#head.freelists,
    Head#head{freelists = alloc_many1(Ftab, 1, Sz * N, A0, Head)}.

%% -> NewFtab | throw(Error)
alloc_many1(Ftab, Pos, Size, A0, H) ->
    {FPos, Addr} = find_first_free(Ftab, Pos, Pos, H),
    true = Addr >= A0, % assertion
    if 
	?POW(FPos - 1) >= Size ->
	    alloc_many2(Ftab, sz2pos(Size), Size, A0, H);
	true ->
	    NewFtab = reserve_buddy(Ftab, FPos, FPos, Addr),
	    NSize = Size - ?POW(FPos-1),
	    alloc_many1(NewFtab, FPos, NSize, Addr, H)
    end.

alloc_many2(Ftab, _Pos, 0, _A0, _H) ->
    Ftab;
alloc_many2(Ftab, Pos, Size, A0, H) when Size band ?POW(Pos-1) > 0 ->
    {FPos, Addr} = find_first_free(Ftab, Pos, Pos, H),
    true = Addr >= A0, % assertion
    NewFtab = reserve_buddy(Ftab, FPos, Pos, Addr),
    NSize = Size - ?POW(Pos - 1),
    alloc_many2(NewFtab, Pos-1, NSize, Addr, H);
alloc_many2(Ftab, Pos, Size, A0, H) ->
    alloc_many2(Ftab, Pos-1, Size, A0, H).

%% -> {NewHead, Addr} | throw(Error)
alloc(Head, Sz) when Head#head.fixed == true -> % when Sz > 0
    ?DEBUGM("alloc of size ~p (fixed)", [Sz]),
    Pos = sz2pos(Sz),
    {Frozen, Ftab} = Head#head.freelists,
    {FPos, Addr} = find_first_free(Frozen, Pos, Pos, Head),
    NewFrozen = reserve_buddy(Frozen, FPos, Pos, Addr),
    Ftab1 = undo_free(Ftab, FPos, Addr),
    NewFtab = move_down(Ftab1, FPos, Pos, Addr),
    NewFreelists = {NewFrozen, NewFtab},
    {Head#head{freelists = NewFreelists}, Addr};
alloc(Head, Sz) when Head#head.fixed == false -> % when Sz > 0
    ?DEBUGM("alloc of size ~p", [Sz]),
    Pos = sz2pos(Sz),
    Ftab = Head#head.freelists,
    {FPos, Addr} = find_first_free(Ftab, Pos, Pos, Head),
    NewFtab = reserve_buddy(Ftab, FPos, Pos, Addr),
    {Head#head{freelists = NewFtab}, Addr}.

find_first_free(_Ftab, Pos, _Pos0, Head) when Pos > ?MAXBUD ->
    throw({error, {no_more_space_on_file, Head#head.filename}});
find_first_free(Ftab, Pos, Pos0, Head) ->
    PosTab = element(Pos, Ftab),
    case bplus_lookup_first(PosTab) of
	undefined -> 
	    find_first_free(Ftab, Pos+1, Pos0, Head);
	{ok, Addr} when Addr + ?POW(Pos0-1) > ?POW(?MAXBUD-1)-?MAXFREELISTS ->
	    %% We would occupy (some of) the area reserved for the free lists.
	    throw({error, {no_more_space_on_file, Head#head.filename}});
	{ok, Addr} ->
	    {Pos, Addr}
    end.

%% When the table is fixed, free/4 may have joined buddies so that the
%% requested block is now part of some larger block. We have to find
%% that block, and insert free buddies along the way.
undo_free(Ftab, Pos, Addr) ->
    PosTab = element(Pos, Ftab),
    case bplus_lookup(PosTab, Addr) of
	undefined ->
	    {BuddyAddr, MoveUpAddr} = my_buddy(Addr, ?POW(Pos-1)),
	    NewFtab = setelement(Pos, Ftab, bplus_insert(PosTab, BuddyAddr)),
	    undo_free(NewFtab, Pos+1, MoveUpAddr);
	{ok, Addr} ->
	    NewPosTab = bplus_delete(PosTab, Addr),
	    setelement(Pos, Ftab, NewPosTab)
    end.

reserve_buddy(Ftab, Pos, Pos0, Addr) ->
    PosTab = element(Pos, Ftab),
    NewPosTab = bplus_delete(PosTab, Addr),
    NewFtab = setelement(Pos, Ftab, NewPosTab),
    move_down(NewFtab, Pos, Pos0, Addr).

move_down(Ftab, Pos, Pos, _Addr) ->
    ?DEBUGM(" to address ~p, table ~p (~p bytes)~n", 
	    [_Addr, Pos, ?POW(Pos-1)]),
    Ftab;
move_down(Ftab, Pos, Pos0, Addr) ->
    Pos_1 = Pos - 1,
    Size = ?POW(Pos_1),
    HighBuddy = (Addr + (Size bsr 1)),
    NewPosTab_1 = bplus_insert(element(Pos_1, Ftab), HighBuddy),
    NewFtab = setelement(Pos_1, Ftab, NewPosTab_1), 
    move_down(NewFtab, Pos_1, Pos0, Addr).

free(Head, Addr, Sz) ->
    ?DEBUGM("free of size ~p at address ~p~n", [Sz, Addr]),
    Ftab = get_freelists(Head),
    Pos = sz2pos(Sz),
    set_freelists(Head, free_in_pos(Ftab, Addr, Pos)).

free_in_pos(Ftab, _Addr, Pos) when Pos > ?MAXBUD ->
    Ftab;
free_in_pos(Ftab, Addr, Pos) ->
    PosTab = element(Pos, Ftab),
    {BuddyAddr, MoveUpAddr} = my_buddy(Addr, ?POW(Pos-1)),
    case bplus_lookup(PosTab, BuddyAddr) of
	undefined -> % no buddy found
	    ?DEBUGM("  table ~p, no buddy~n", [Pos]),
	    setelement(Pos, Ftab, bplus_insert(PosTab, Addr));
	{ok, BuddyAddr} -> % buddy found
	    PosTab1 = bplus_delete(PosTab, Addr),
	    PosTab2 = bplus_delete(PosTab1, BuddyAddr),
	    ?DEBUGM("  table ~p, with buddy ~p~n", [Pos, BuddyAddr]),
	    NewFtab = setelement(Pos, Ftab, PosTab2),
	    free_in_pos(NewFtab, MoveUpAddr, Pos+1)
    end.

get_freelists(Head) when Head#head.fixed == false ->
    Head#head.freelists;
get_freelists(Head) when Head#head.fixed == true ->
    {_Frozen, Current} = Head#head.freelists,
    Current.

set_freelists(Head, Ftab) when Head#head.fixed == false ->
    Head#head{freelists = Ftab};
set_freelists(Head, Ftab) when Head#head.fixed == true ->
    {Frozen, _} = Head#head.freelists,
    Head#head{freelists = {Frozen,Ftab}}.

%% Bug: If Sz0 is equal to 2**k for some k, then 2**(k+1) bytes are
%% allocated (wasting 2**k bytes).
sz2pos(N) when N > 0 ->
    log(N, 16, 2).

log(N, _I, L) when N < 2 ->
    L;
log(N, I, L) when N >= 1 bsl I ->
    log(N bsr I, I bsr 1, L+I);
log(N, I, L) ->
    log(N, I bsr 1, L).

%% Calculate the buddy of Addr
my_buddy(Addr, Sz) ->
    case (Addr - ?BASE) band Sz of
	0 -> % even, buddy is higher addr
	    {Addr+Sz, Addr};
	_ -> % odd, buddy is lower addr
            T = Addr-Sz,
	    {T, T}
    end.

all_free(Tab) ->
    case all_free(all(Tab), ?BASE, ?BASE, []) of
	[{?BASE,?BASE} | L] -> L;
	L -> L
    end.
    
all_free([], X0, Y0, F) ->
    lists:reverse([{X0,Y0} | F]);
all_free([{X,Y} | L], X0, Y0, F) when Y0 == X ->
    all_free(L, X0, Y, F);
all_free([{X,Y} | L], X0, Y0, F) when Y0 < X ->
    all_free(L, X, Y, [{X0,Y0} | F]).

all_allocated(Tab) ->
    all_allocated(all(Tab), 0, ?BASE, []).

all_allocated([], _X0, _Y0, []) ->
    <<>>;
all_allocated([], _X0, _Y0, A0) ->
    [<<From:32, To:32>> | A] = lists:reverse(A0),
    {From, To, list_to_binary(lists:reverse(A))};
all_allocated([{X,Y} | L], X0, Y0, A) when Y0 == X ->
    all_allocated(L, X0, Y, A);
all_allocated([{X,Y} | L], _X0, Y0, A) when Y0 < X ->
    all_allocated(L, X, Y, [<<Y0:32,X:32>> | A]).

all(Tab) ->
    all(Tab, size(Tab), []).

all(_Tab, 0, L) ->
    %% This is not as bad as it looks. L contains less than 32 runs,
    %% so there will be a quite small number of merges.
    lists:sort(L);
all(Tab, I, L) ->
    LL = collect_tree(element(I, Tab), I, L),
    all(Tab, I-1, LL).

%%%-----------------------------------------------------------------
%%% These functions implements a B+ tree.
%%% The code is originally written by lelle@erlang.ericsson.se,
%%%-----------------------------------------------------------------

-define(max_size, 16).
-define(min_size, 8).
%%-----------------------------------------------------------------
%% Finds out the type of the node: 'l' or 'n'.
%%-----------------------------------------------------------------
-define(NODE_TYPE(Tree), element(1, Tree)).
%% Finds out if a node/leaf is full or not.
-define(FULL(Tree), (bplus_get_size(Tree) >= ?max_size)).
%% Finds out if a node/leaf is filled up over its limit.
-define(OVER_FULL(Tree), (bplus_get_size(Tree) > ?max_size)).
%% Finds out if a node/leaf has less items than allowed.
-define(UNDER_FILLED(Tree), (bplus_get_size(Tree) < ?min_size)).
%% Finds out if a node/leaf has as few items as minimum allowed.
-define(LOW_FILLED(Tree), (bplus_get_size(Tree) =< ?min_size)).
%%Returns a key in a leaf at position Pos.
-define(GET_LEAF_KEY(Leaf, Pos), element(Pos+1, Leaf)).
-define(GET_LEAF_KEY_VAL(Leaf, Pos), element(Pos+1, Leaf)).

%% Special for dets.
collect_tree(v, _TI, Acc) -> Acc;
collect_tree(T, TI, Acc) ->
    collect_tree2(T, TI-1, Acc).

collect_tree2(Tree, TI, Acc) ->
    S = bplus_get_size(Tree),
    case ?NODE_TYPE(Tree) of
	l ->
	    collect_leaf(Tree, S, TI, Acc);
	n ->
	    collect_node(Tree, S, TI, Acc)
    end.
    
collect_leaf(_Leaf, 0, _TI, Acc) ->
    Acc;
collect_leaf(Leaf, I, TI, Acc) ->
    Key = ?GET_LEAF_KEY(Leaf, I),
    V = {Key, Key+?POW(TI)},
    collect_leaf(Leaf, I-1, TI, [V | Acc]).

collect_node(_Node, 0, _TI, Acc) ->
    Acc;
collect_node(Node, I, TI, Acc) ->
    Acc1 = collect_tree2(bplus_get_tree(Node, I), TI, Acc),
    collect_node(Node, I-1, TI, Acc1).

%%-----------------------------------------------------------------
%% Func: empty_tree/0
%% Purpose: Creates a new empty tree.
%% Returns: tree()
%%-----------------------------------------------------------------
bplus_empty_tree() -> v.

%%-----------------------------------------------------------------
%% Func: lookup/2
%% Purpose: Looks for Key in the Tree.
%% Returns: {ok, {Key, Val}} | 'undefined'.
%%-----------------------------------------------------------------
bplus_lookup(v, _Key) -> undefined;
bplus_lookup(Tree, Key) ->
    case ?NODE_TYPE(Tree) of
	l ->
	    bplus_lookup_leaf(Key, Tree);
	n ->
	    {_, SubTree} = bplus_select_sub_tree(Tree, Key),
	    bplus_lookup(SubTree, Key)
    end.

%%-----------------------------------------------------------------
%% Searches through a leaf until the Key is ok or
%% when it is determined that it does not exist.
%%-----------------------------------------------------------------
bplus_lookup_leaf(Key, Leaf) -> 
    bplus_lookup_leaf_2(Key, Leaf, bplus_get_size(Leaf)).

bplus_lookup_leaf_2(_, _, 0) -> undefined;
bplus_lookup_leaf_2(Key, Leaf, N) ->
    case ?GET_LEAF_KEY(Leaf, N) of
	Key -> {ok, Key};
	_ ->
	    bplus_lookup_leaf_2(Key, Leaf, N-1)
    end.

%%-----------------------------------------------------------------
%% Func: lookup_first/1
%% Purpose: Finds the smallest key in the entire Tree.
%% Returns: {ok, {Key, Val}} | 'undefined'.
%%-----------------------------------------------------------------
bplus_lookup_first(v) -> undefined;
bplus_lookup_first(Tree) ->
    case ?NODE_TYPE(Tree) of
	l ->
	    % Then it is the leftmost key here.
	    {ok, ?GET_LEAF_KEY(Tree, 1)};         
	n ->
	    % Look in the leftmost subtree.
	    bplus_lookup_first(bplus_get_tree(Tree, 1))
    end.


%%-----------------------------------------------------------------
%% Func: insert/3
%% Purpose: Inserts a new {Key, Value} into the tree.
%% Returns: tree()
%%-----------------------------------------------------------------
bplus_insert(v, Key) -> bplus_mk_leaf([Key]);
bplus_insert(Tree, Key) ->
    NewTree = bplus_insert_in(Tree, Key),
    case ?OVER_FULL(NewTree) of
	false ->
	    NewTree;
	% If the node is over-full the tree will grow.
	true ->
	    {LTree, DKey, RTree} = 
		case ?NODE_TYPE(NewTree) of
		    l ->
			bplus_split_leaf(NewTree);
		    n ->
			bplus_split_node(NewTree)
		end,
	    bplus_mk_node([LTree, DKey, RTree])
    end.

%%-----------------------------------------------------------------
%% Func: delete/2
%% Purpose: Deletes a key from the tree (if present).
%% Returns: tree()
%%-----------------------------------------------------------------
bplus_delete(v, _Key) -> v;
bplus_delete(Tree, Key) ->
    NewTree = bplus_delete_in(Tree, Key),
    S = bplus_get_size(NewTree),
    case ?NODE_TYPE(NewTree) of
	l ->
	    if
		S == 0 ->
		    v;
		true ->
		    NewTree
	    end;
	n ->
	    if
		S == 1 ->
		    bplus_get_tree(NewTree, 1);
		true ->
		    NewTree
	    end
    end.


%%% -----------------------
%%% Help function to insert.
%%% -----------------------

bplus_insert_in(Tree, Key) ->
    case ?NODE_TYPE(Tree) of
	l ->
	    bplus_insert_in_leaf(Tree, Key);
	n ->
	    {Pos, SubTree} = bplus_select_sub_tree(Tree, Key),  
            % Pos = "the position of the subtree".
	    NewSubTree = bplus_insert_in(SubTree, Key),
	    case ?OVER_FULL(NewSubTree) of
		false ->
		    bplus_put_subtree(Tree, [NewSubTree, Pos]);
		true ->
		    case bplus_reorganize_tree_ins(Tree, NewSubTree, Pos) of
			{left, {LeftT, DKey, MiddleT}} ->
			    bplus_put_subtree(bplus_put_lkey(Tree, DKey, Pos),
					[LeftT, Pos-1, MiddleT, Pos]);
			{right, {MiddleT, DKey, RightT}} ->
			    bplus_put_subtree(bplus_put_rkey(Tree, DKey, Pos),
					[MiddleT, Pos, RightT, Pos+1]);
			{split, {LeftT, DKey, RightT}} ->
			    bplus_extend_tree(Tree, {LeftT, DKey, RightT}, Pos)
		    end
	    end
    end.

%%-----------------------------------------------------------------
%% Inserts a key in correct position in a leaf.
%%-----------------------------------------------------------------
bplus_insert_in_leaf(Leaf, Key) ->
    bplus_insert_in_leaf_2(Leaf, Key, bplus_get_size(Leaf), []).

bplus_insert_in_leaf_2(Leaf, Key, 0, Accum) ->
    bplus_insert_in_leaf_3(Leaf, 0, [Key|Accum]);
bplus_insert_in_leaf_2(Leaf, Key, N, Accum) ->
    K = ?GET_LEAF_KEY(Leaf, N),
    if
	Key < K ->
	    % Not here!
	    bplus_insert_in_leaf_2(Leaf, Key, N-1, [K|Accum]);
	K < Key ->
	    % Insert here.
	    bplus_insert_in_leaf_3(Leaf, N-1, [K, Key|Accum]);
	K == Key ->
	    % Replace (?).
	    bplus_insert_in_leaf_3(Leaf, N-1, [ Key|Accum])
    end.

bplus_insert_in_leaf_3(_Leaf, 0, LeafList) ->
    bplus_mk_leaf(LeafList);
bplus_insert_in_leaf_3(Leaf, N, LeafList) ->
    bplus_insert_in_leaf_3(Leaf, N-1, [?GET_LEAF_KEY(Leaf, N)|LeafList]).


%%% -------------------------
%%% Help functions for delete.
%%% -------------------------

bplus_delete_in(Tree, Key) ->
    case ?NODE_TYPE(Tree) of
	l ->
	    bplus_delete_in_leaf(Tree, Key);
	n ->
	    {Pos, SubTree} = bplus_select_sub_tree(Tree, Key),  
	    % Pos = "the position of the subtree".
	    NewSubTree = bplus_delete_in(SubTree, Key),
	    % Check if it has become to small now
	    case ?UNDER_FILLED(NewSubTree) of
		false ->
		    bplus_put_subtree(Tree, [NewSubTree, Pos]);
		true ->
		    case bplus_reorganize_tree_del(Tree, NewSubTree, Pos) of
			{left, {LeftT, DKey, MiddleT}} ->
			    bplus_put_subtree(bplus_put_lkey(Tree, DKey, Pos),
					[LeftT, Pos-1, MiddleT, Pos]);
			{right, {MiddleT, DKey, RightT}} ->
			    bplus_put_subtree(bplus_put_rkey(Tree, DKey, Pos),
					[MiddleT, Pos, RightT, Pos+1]);
			{join_left, JoinedTree} ->
			    bplus_joinleft_tree(Tree, JoinedTree, Pos);
			{join_right, JoinedTree} ->
			    bplus_joinright_tree(Tree, JoinedTree, Pos)
		    end
	    end
    end.

%%-----------------------------------------------------------------
%% Deletes a key from the leaf returning a new (smaller) leaf.
%%-----------------------------------------------------------------
bplus_delete_in_leaf(Leaf, Key) ->
    bplus_delete_in_leaf_2(Leaf, Key, bplus_get_size(Leaf), []).

bplus_delete_in_leaf_2(Leaf, _, 0, _) -> Leaf;
bplus_delete_in_leaf_2(Leaf, Key, N, Accum) ->
    K = ?GET_LEAF_KEY(Leaf, N),
    if
	Key == K ->
            % Remove this one!
	    bplus_delete_in_leaf_3(Leaf, N-1, Accum);
	true ->
	    bplus_delete_in_leaf_2(Leaf, Key, N-1, [K|Accum])
    end.

bplus_delete_in_leaf_3(_Leaf, 0, LeafList) ->
    bplus_mk_leaf(LeafList);
bplus_delete_in_leaf_3(Leaf, N, LeafList) ->
    bplus_delete_in_leaf_3(Leaf, N-1, [?GET_LEAF_KEY(Leaf, N)|LeafList]).



%%-----------------------------------------------------------------
%% Selects and returns which subtree the search should continue in.
%%-----------------------------------------------------------------
bplus_select_sub_tree(Tree, Key) ->
    bplus_select_sub_tree_2(Tree, Key, bplus_get_size(Tree)).

bplus_select_sub_tree_2(Tree, _Key, 1) -> {1, bplus_get_tree(Tree, 1)};
bplus_select_sub_tree_2(Tree, Key, N) ->
    K = bplus_get_lkey(Tree, N),
    if
	K > Key ->
	    bplus_select_sub_tree_2(Tree, Key, N-1);
	K =< Key ->
            % Here it is!
	    {N, bplus_get_tree(Tree, N)}
    end.

%%-----------------------------------------------------------------
%% Selects which brother that should take over some of our items.
%% Or if they are both full makes a split.
%%-----------------------------------------------------------------
bplus_reorganize_tree_ins(Tree, NewSubTree, 1) ->
    RTree = bplus_get_tree(Tree, 2),  % 2 = Pos+1 = 1+1.
    case ?FULL(RTree) of
	false ->
	    bplus_reorganize_tree_r(Tree, NewSubTree, 1, RTree);
	true ->
            % It is full, we must split this one!
	    bplus_reorganize_tree_s(NewSubTree)
    end;
bplus_reorganize_tree_ins(Tree, NewSubTree, Pos) ->
    Size = bplus_get_size(Tree),
    if
	Pos == Size ->
            % Pos is the rightmost postion!.
            % Our only chance is the left one.
	    LTree = bplus_get_tree(Tree, Pos-1),
 	    case ?FULL(LTree) of
		false ->
		    bplus_reorganize_tree_l(Tree, NewSubTree, Pos, LTree);
		true ->
		    % It is full, we must split this one!
		    bplus_reorganize_tree_s(NewSubTree)
	    end;
	true ->
            % Pos is somewhere inside the node.
	    LTree = bplus_get_tree(Tree, Pos-1),
	    RTree = bplus_get_tree(Tree, Pos+1),
	    SL = bplus_get_size(LTree),
	    SR = bplus_get_size(RTree),
	    if
		SL > SR ->
		    bplus_reorganize_tree_r(Tree, NewSubTree, Pos, RTree);
		SL < SR ->
		    bplus_reorganize_tree_l(Tree, NewSubTree, Pos, LTree);
		true ->
		    case ?FULL(LTree) of
			false ->
			    bplus_reorganize_tree_l(Tree, NewSubTree, Pos, LTree);
			true ->
			    bplus_reorganize_tree_s(NewSubTree)
		    end
	    end
    end.

%%-----------------------------------------------------------------
%% This function fills over items from brothers to maintain the minimum
%% number of items per node/leaf.
%%-----------------------------------------------------------------
bplus_reorganize_tree_del(Tree, NewSubTree, 1) ->
    % The case when Pos is at leftmost position.
    RTree = bplus_get_tree(Tree, 2),  % 2 = Pos+1 = 1+1.
    case ?LOW_FILLED(RTree) of
	false ->
	    bplus_reorganize_tree_r(Tree, NewSubTree, 1, RTree);
	true ->
            % It is to small, we must join them!
	    bplus_reorganize_tree_jr(Tree, NewSubTree, 1, RTree)
    end;
bplus_reorganize_tree_del(Tree, NewSubTree, Pos) ->
    Size = bplus_get_size(Tree),
    if
	Pos == Size ->
            % Pos is the rightmost postion!.
            % Our only chance is the left one.
	    LTree = bplus_get_tree(Tree, Pos-1),
	    case ?LOW_FILLED(LTree) of
		false ->
		    bplus_reorganize_tree_l(Tree, NewSubTree, Pos, LTree);
		true ->
                    % It is to small, we must join this one!
		    bplus_reorganize_tree_jl(Tree, NewSubTree, Pos, LTree)
	    end;
	true ->
            % Pos is somewhere inside the node.
	    LTree = bplus_get_tree(Tree, Pos-1),
	    RTree = bplus_get_tree(Tree, Pos+1),
	    SL = bplus_get_size(LTree),
	    SR = bplus_get_size(RTree),
	    if
		SL>SR ->
		    bplus_reorganize_tree_l(Tree, NewSubTree, Pos, LTree);
		SL < SR ->
		    bplus_reorganize_tree_r(Tree, NewSubTree, Pos, RTree);
		true ->
		    case ?LOW_FILLED(LTree) of
			false ->
			    bplus_reorganize_tree_l(Tree, NewSubTree, Pos, LTree);
			true ->
			    bplus_reorganize_tree_jl(Tree, NewSubTree, Pos, LTree)
		    end
	    end
    end.


bplus_reorganize_tree_l(Tree, NewSubTree, Pos, LTree) ->
    case ?NODE_TYPE(NewSubTree) of
	l ->
	    {left, bplus_split_leaf(
		     bplus_mk_leaf(
		       lists:append(bplus_leaf_to_list(LTree),
				    bplus_leaf_to_list(NewSubTree))))};
	n ->
	    {left, bplus_split_node(
		     bplus_mk_node(
		       lists:append([bplus_node_to_list(LTree),
				     [bplus_get_lkey(Tree, Pos)],
				     bplus_node_to_list(NewSubTree)])))}
    end.

bplus_reorganize_tree_r(Tree, NewSubTree, Pos, RTree) ->
    case ?NODE_TYPE(NewSubTree) of
	l ->
	    {right, 
	     bplus_split_leaf(
	       bplus_mk_leaf(
		 lists:append([bplus_leaf_to_list(NewSubTree),
			       bplus_leaf_to_list(RTree)])))};
	n ->
	    {right, 
	     bplus_split_node(
	       bplus_mk_node(
		 lists:append([bplus_node_to_list(NewSubTree),
			       [bplus_get_rkey(Tree, Pos)],
			       bplus_node_to_list(RTree)])))}
    end.

bplus_reorganize_tree_s(NewSubTree) ->
    case ?NODE_TYPE(NewSubTree) of
	l ->
	    {split, bplus_split_leaf(NewSubTree)};
	n ->
	    {split, bplus_split_node(NewSubTree)}
    end.

bplus_reorganize_tree_jl(Tree, NewSubTree, Pos, LTree) ->
    case ?NODE_TYPE(NewSubTree) of
	l ->
	    {join_left, 
	     bplus_mk_leaf(lists:append([bplus_leaf_to_list(LTree),
					 bplus_leaf_to_list(NewSubTree)]))};
	n ->
	    {join_left, 
	     bplus_mk_node(lists:append([bplus_node_to_list(LTree),
					 [bplus_get_lkey(Tree, Pos)],
					 bplus_node_to_list(NewSubTree)]))}
    end.

bplus_reorganize_tree_jr(Tree, NewSubTree, Pos, RTree) ->
    case ?NODE_TYPE(NewSubTree) of
	l ->
	    {join_right, 
	     bplus_mk_leaf(lists:append([bplus_leaf_to_list(NewSubTree),
					 bplus_leaf_to_list(RTree)]))};
	n ->
	    {join_right, 
	     bplus_mk_node(lists:append([bplus_node_to_list(NewSubTree),
					 [bplus_get_rkey(Tree, Pos)],
					 bplus_node_to_list(RTree)]))}
    end.


%%-----------------------------------------------------------------
%% Takes a leaf and divides it into two equal big leaves.
%% The result is returned in a tuple. The dividing key is also returned.
%%-----------------------------------------------------------------
bplus_split_leaf(Leaf) ->
    S = bplus_get_size(Leaf),
    bplus_split_leaf_2(Leaf, S, S div 2, []).

bplus_split_leaf_2(Leaf, Pos, 1, Accum) -> 
    K = ?GET_LEAF_KEY(Leaf, Pos),
    bplus_split_leaf_3(Leaf, Pos-1, [], K, [K|Accum]);
bplus_split_leaf_2(Leaf, Pos, N, Accum) ->
    bplus_split_leaf_2(Leaf, Pos-1, N-1, [?GET_LEAF_KEY(Leaf, Pos)|Accum]).

bplus_split_leaf_3(_, 0, LeftAcc, DKey, RightAcc) ->
    {bplus_mk_leaf(LeftAcc), DKey, bplus_mk_leaf(RightAcc)};
bplus_split_leaf_3(Leaf, Pos, LeftAcc, DKey, RightAcc) ->
    bplus_split_leaf_3(Leaf, Pos-1, [?GET_LEAF_KEY(Leaf, Pos)|LeftAcc],
		       DKey, RightAcc).

%%-----------------------------------------------------------------
%% Takes a node and divides it into two equal big nodes.
%% The result is returned in a tuple. The dividing key is also returned.
%%-----------------------------------------------------------------
bplus_split_node(Node) ->
    S = bplus_get_size(Node),
    bplus_split_node_2(Node, S, S div 2, []).

bplus_split_node_2(Node, Pos, 1, Accum) ->
    bplus_split_node_3(Node, Pos-1, [], bplus_get_lkey(Node, Pos),
		 [bplus_get_tree(Node, Pos)|Accum]);
bplus_split_node_2(Node, Pos, N, Accum) ->
    bplus_split_node_2(Node, Pos-1, N-1, [bplus_get_lkey(Node, Pos),
				    bplus_get_tree(Node, Pos)|Accum]).

bplus_split_node_3(Node, 1, LeftAcc, DKey, RightAcc) ->
    {bplus_mk_node([bplus_get_tree(Node, 1)|LeftAcc]), DKey, 
     bplus_mk_node(RightAcc)};
bplus_split_node_3(Node, Pos, LeftAcc, DKey, RightAcc) ->
    bplus_split_node_3(Node, Pos-1,
		       [bplus_get_lkey(Node, Pos), 
			bplus_get_tree(Node, Pos)|LeftAcc],
		       DKey, RightAcc).

%%-----------------------------------------------------------------
%% Inserts a joined tree insted of the old one at position Pos and
%% the one nearest left/right brother.
%%-----------------------------------------------------------------
bplus_joinleft_tree(Tree, JoinedTree, Pos) ->
    bplus_join_tree_2(Tree, JoinedTree, Pos, bplus_get_size(Tree), []).
bplus_joinright_tree(Tree, JoinedTree, Pos) ->
    bplus_join_tree_2(Tree, JoinedTree, Pos+1, bplus_get_size(Tree), []).

bplus_join_tree_2(Tree, JoinedTree, Pos, Pos, Accum) ->
    bplus_join_tree_3(Tree, Pos-2, [JoinedTree|Accum]);
bplus_join_tree_2(Tree, JoinedTree, Pos, N, Accum) ->
    bplus_join_tree_2(Tree, JoinedTree, Pos, N-1,
		[bplus_get_lkey(Tree, N), bplus_get_tree(Tree, N)|Accum]).

bplus_join_tree_3(_Tree, 0, Accum) -> bplus_mk_node(Accum);
bplus_join_tree_3(Tree, Pos, Accum) ->
    bplus_join_tree_3(Tree, Pos-1, [bplus_get_tree(Tree, Pos), 
				    bplus_get_rkey(Tree, Pos)|Accum]).

%%% ---------------------------------
%%% Primitive datastructure functions.
%%% ---------------------------------

%%-----------------------------------------------------------------
%% Constructs a node out of list format.
%%-----------------------------------------------------------------
bplus_mk_node(NodeList) -> list_to_tuple([ n |NodeList]).

%%-----------------------------------------------------------------
%% Converts the node into list format.
%%-----------------------------------------------------------------
bplus_node_to_list(Node) ->
    [_|NodeList] = tuple_to_list(Node),
    NodeList.

%%-----------------------------------------------------------------
%% Constructs a leaf out of list format.
%%-----------------------------------------------------------------
bplus_mk_leaf(KeyList) -> list_to_tuple([l|KeyList]).

%%-----------------------------------------------------------------
%% Converts a leaf into list format.
%%-----------------------------------------------------------------
bplus_leaf_to_list(Leaf) ->
    [_|LeafList] = tuple_to_list(Leaf),
    LeafList.

%%-----------------------------------------------------------------
%% Changes subtree "pointers" in a node.
%%-----------------------------------------------------------------
bplus_put_subtree(Tree, []) -> Tree;
bplus_put_subtree(Tree, [NewSubTree, Pos|Rest]) ->
    bplus_put_subtree(setelement(Pos*2, Tree, NewSubTree), Rest).

%%-----------------------------------------------------------------
%% Replaces the tree at position Pos with two new trees.
%%-----------------------------------------------------------------
bplus_extend_tree(Tree, Inserts, Pos) ->
    bplus_extend_tree_2(Tree, Inserts, Pos, bplus_get_size(Tree), []).

bplus_extend_tree_2(Tree, {T1, DKey, T2}, Pos, Pos, Accum) ->
    bplus_extend_tree_3(Tree, Pos-1, [T1, DKey, T2|Accum]);
bplus_extend_tree_2(Tree, Inserts, Pos, N, Accum) ->
    bplus_extend_tree_2(Tree, Inserts, Pos, N-1,
		  [bplus_get_lkey(Tree, N), bplus_get_tree(Tree, N)|Accum]).

bplus_extend_tree_3(_, 0, Accum) -> bplus_mk_node(Accum);
bplus_extend_tree_3(Tree, N, Accum) ->
    bplus_extend_tree_3(Tree, N-1, [bplus_get_tree(Tree, N), 
				    bplus_get_rkey(Tree, N)|Accum]).

%%-----------------------------------------------------------------
%% Changes the dividing key between two trees.
%%-----------------------------------------------------------------
bplus_put_lkey(Tree, DKey, Pos) -> setelement(Pos*2-1, Tree, DKey).
bplus_put_rkey(Tree, DKey, Pos) -> setelement(Pos*2+1, Tree, DKey).


%%-----------------------------------------------------------------
%% Calculates the number of items in a node/leaf.
%%-----------------------------------------------------------------
bplus_get_size(Tree) ->
    case ?NODE_TYPE(Tree) of
	l ->
	    size(Tree)-1;
	n ->
	    size(Tree) div 2
    end.

%%-----------------------------------------------------------------
%% Returns a tree at position Pos from an internal node.
%%-----------------------------------------------------------------
bplus_get_tree(Tree, Pos) -> element(Pos*2, Tree).

%%-----------------------------------------------------------------
%% Returns dividing keys, left of or right of a tree.
%%-----------------------------------------------------------------
bplus_get_lkey(Tree, Pos) -> element(Pos*2-1, Tree).
bplus_get_rkey(Tree, Pos) -> element(Pos*2+1, Tree).
