%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%----------------------------------------------------------------------
%%% File    : fprof.erl
%%% Author  : Raimo Niskanen <raimo@erix.ericsson.se>
%%% Purpose : Yet another profiling tool. Uses trace to file and 
%%%           offline analysis.
%%% Created : 23 May 2001 by Raimo Niskanen <raimo@erix.ericsson.se>
%%%----------------------------------------------------------------------

-module(fprof).
-author('raimo@erix.ericsson.se').

%% External exports
-export([start/0, start/1, start/2, start/3, start/4, start/5, stop/0]).
-export([profile/4, profile/5]).
-export([analyse/0, analyse/1, analyse/2]).

-import(lists, [flatten/1, flatten/2, reverse/1, reverse/2, sort/1, foldl/3]).

-define(DEFAULT_FILESPEC, "fprof.log").
-define(FNAME_WIDTH, 72).
-define(NR_WIDTH, 15).

-record(clocks, {
	  id,
	  cnt = 0,   % Number of calls
	  own = 0,   % Own time (wall clock)
	  acc = 0}). % Accumulated time : own + subfunctions (wall clock)

-record(proc, {
	  id,
	  parent = undefined,
	  spawned_as = undefined,
	  init_log = [],
	  init_cnt = 2}).

-record(funcstat, {
	  callers_sum,   % #clocks{id = {Pid, Caller, Func}}
	  called_sum,    % #clocks{id = {Pid, Caller, Func}}
	  callers = [],  % [#clocks{}, ...]
	  called = []}). % [#clocks{}, ...]

-record(opts, {
	  out = []}).



%% Debug exports
-export([analyse_this/0, analyse_this/1, analyse_this/2]).

-export([trace_call_1/5, burn/0]).

-define(debug, 9).

-ifdef(debug).
-define(dbg(Level, F, A), 
	if Level >= ?debug -> io:format(F, A), ok;
	   true -> ok
	end).
-else.
-define(dbg(Level, F, A), ok).
-endif.

%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------

start() ->
    start([]).

start(FileSpec) ->
    start(FileSpec, self()).

start(FileSpec, Procs) ->
    start(FileSpec, Procs, []).

start([], Procs, Flags) ->
    start(?DEFAULT_FILESPEC, Procs, Flags);
start(FileSpec, Procs, Flags) when list(Flags) ->
    dbg:stop(),
    TraceFlags = 
	case lists:member(trace, Flags) of
	    true ->
		[call, return_to, send, 'receive', procs,
		 set_on_spawn];
	    false ->
		[call, return_to, 
		 running, procs, garbage_collection, 
		 arity, timestamp, set_on_spawn]
	end,
    TracerFun = dbg:trace_port(file, FileSpec),
    case dbg:tracer(port, dbg:trace_port(file, FileSpec)) of
	{ok, _Pid} ->
%	    dbg:p(all, [clear]),
	    dbg:ctp({'_', '_', '_'}),
	    case my_p(Procs, TraceFlags) of
		ok ->
		    case my_tpl([{'_', '_', '_'}, on_load], 
				[{'_', [], []}]) of
			ok ->
			    ok;
			{error, _} = Error ->
			    dbg:stop(),
			    Error
		    end;
		{error, _} = Error ->
		    dbg:stop(),
		    Error
	    end;
	{error, _} = Error ->
	    Error
    end;
start(FileSpec, Procs, Flag) ->
    start(FileSpec, Procs, [Flag]).



start(FileSpec, M, F, A) ->
    start(FileSpec, M, F, A, []).

start(FileSpec, M, F, A, Flags) when atom(M), atom(F), list(A), list(Flags) ->
    Parent = self(),
    Ref = make_ref(),
    Pid = spawn(
      fun () ->
	      Parent ! {self(), Ref, ready},
	      receive {Parent, Ref, go} -> ok end,
	      Parent ! 
		  {self(), Ref, tracee, 
		   spawn(
		     fun () ->
			     Parent ! {Ref, apply(M, F, A)}
		     end)}
      end),
    receive {Pid, Ref, ready} -> ok end,
    start(FileSpec, Pid, Flags),
    Pid ! {Parent, Ref, go},
    receive 
	{Pid, Ref, tracee, Tracee} -> 
	    MRef = erlang:monitor(process, Tracee),
	    receive
		{'DOWN', MRef, _, _, Reason} = Msg ->
		    receive {Ref, Result} -> Result 
		    after 0 -> exit(Reason)
		    end;
		{Ref, Result} = Msg ->
		    Result
	    end
    end;
start(FileSpec, M, F, A, Flag) when atom(M), atom(F), list(A) ->
    start(FileSpec, M, F, A, [Flag]).
	      


stop() ->
    dbg:stop().



profile(FileSpec, M, F, Args) ->
    profile(FileSpec, M, F, Args, []).

profile(FileSpec, M, F, Args, Flags) 
  when atom(M), atom(F), list(Args), list(Flags) ->
    Result = start(FileSpec, M, F, Args, Flags),
    stop(),
    Result;
profile(FileSpec, M, F, Args, Flag) 
  when atom(M), atom(F), list(Args) ->
    profile(FileSpec, M, F, Args, [Flag]).



analyse() ->
    analyse([]).

analyse(FileSpec) ->
    analyse(FileSpec, []).

analyse([], Flags) when list(Flags) ->
    analyse(?DEFAULT_FILESPEC, Flags);
analyse(FileSpec, Flags) when list(Flags) ->
    case lists:member(trace, Flags) of
	true ->
	    dbg:trace_client(file, FileSpec);
	false ->
	    Self = self(),
	    Pid = dbg:trace_client(file, FileSpec, 
				   {fun handler/2, 
				    {init, Self, group_leader(), Flags}}),
	    Mref = erlang:monitor(process, Pid),
	    receive
		{Pid, ok} ->
		    ok;
		{'DOWN', Mref, _, _, Reason} ->
		    {error, Reason}
	    end
    end;
analyse(FileSpec, Flag) ->
    analyse(FileSpec, [Flag]).

analyse_this(stop) ->
    trace_off(),
    case whereis(fprof_tracer) of
	undefined ->
	    {error, not_running};
	Pid ->
	    io:format("Stopping tracer ~w~n", [Pid]),
	    MRef = erlang:monitor(process, Pid),
	    exit(Pid, normal),
	    receive
		{Pid, ok} ->
		    receive {'DOWN', MRef, _, _, _} -> ok end,
		    ok;
		{'DOWN', MRef, _, _, Reason} ->
		    receive {Pid, ok} -> 
			    ok 
		    after 0 -> 
			    {error, Reason} 
		    end
	    end
    end.

analyse_this() ->
    analyse_this(self(), []).

analyse_this(Proc, Flags) when list(Flags) ->
    trace_off(),
    Tracer = spawn_tracer(Flags),
    trace_on(Proc, Tracer, prof),
    ok.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

trace_off() ->
    erlang:trace(all, false,
		 [send, 'receive', procs, call, silent, return_to,
		  running, garbage_collection, timestamp, arity, 
		  set_on_spawn, set_on_first_spawn, 
		  set_on_link, set_on_first_link]),
    erlang:trace_pattern({'_', '_', '_'}, false, []),
    erlang:trace_pattern(on_load, false, []).

trace_on(Proc, Tracer, prof) ->    
    erlang:trace_pattern({'_', '_', '_'}, true, []),
    erlang:trace_pattern(on_load, true, []),
    my_trace(Proc, true, 
	     [call, return_to, 
	      running, procs, garbage_collection, 
	      arity, timestamp, set_on_spawn,
	      {tracer, Tracer}]);
trace_on(Proc, Tracer, trace) ->
    erlang:trace_pattern({'_', '_', '_'}, true, []),
    erlang:trace_pattern(on_load, true, []),
    my_trace(Proc, true, 
	     [call, return_to, send, 'receive', procs,
	      set_on_spawn, {tracer, Tracer}]).

my_trace([], _How, _FlagList) ->
    ok;
my_trace([PidSpec | Tail], How, FlagList) ->
    erlang:trace(PidSpec, How, FlagList),
    my_trace(Tail, How, FlagList);
my_trace(PidSpec, How, FlagList) ->
    erlang:trace(PidSpec, How, FlagList),
    ok.



my_p([Proc | Tail], Flags) ->
    dbg:p(Proc, Flags),
    my_p(Tail, Flags);
my_p([], _Flags) ->
    ok;
my_p(Proc, Flags) ->
    case dbg:p(Proc, Flags) of
	{ok, _} ->
	    ok;
	Error ->
	    Error
    end.

my_tpl([MFA | Tail], MatchSpec) ->
    dbg:tpl(MFA, MatchSpec),
    my_tpl(Tail, MatchSpec);
my_tpl([], _MatchSpec) ->
    ok;
my_tpl(MFA, MatchSpec) ->
    case dbg:tpl(MFA, MatchSpec) of
	{ok, _} ->
	    ok;
	Error ->
	    Error
    end.



spawn_tracer(Flags) ->
    Parent = self(),
    GroupLeader = group_leader(),
    Ref = make_ref(),
    Pid = 
	spawn(
	  fun() ->
		  process_flag(trap_exit, true),
		  register(fprof_tracer, self()),
		  Parent ! {self(), Ref, started},
		  handler_proxy({init, Parent, GroupLeader, Flags})
	  end),
    MRef = erlang:monitor(process, Pid),
    receive
	{Pid, Ref, started} ->
	    Pid;
	{'DOWN', MRef, _, _, Reason} ->
	    exit(Reason)
    end.
    
handler_proxy(State) ->
    receive
	Trace when element(1, Trace) == trace ->
	    handler_proxy(handler(Trace, State));
	Trace when element(1, Trace) == trace_ts ->
	    handler_proxy(handler(Trace, State));
	end_of_trace ->
	    handler(end_of_trace, State);
	{'EXIT', Pid, Reason} ->
	    put(parent, Pid),
	    handler(end_of_trace, State),
	    exit(Reason)
    end.



%% Entry point - create data tables and state data
handler(Trace, {init, Parent, GroupLeader, Flags}) ->
    ?dbg(1, "Trace: ~p ~p~n", [Parent, Flags]),
    put(parent, Parent),
    put(group_leader, GroupLeader),
    put(flags, Flags),
    put(table, ets:new(?MODULE, [set, private, {keypos, #clocks.id}])),
    io:format(get(group_leader), "Reading trace file...~n", []),
    handler(Trace, {0, undefined});
%% end of trace
handler(end_of_trace, {N, TS} = State) ->
    ?dbg(1, "End of trace~n", []),
    FirstTS = erase(first_ts),
    T = ts_sub(TS, FirstTS),
    io:format(get(group_leader), "~nFinished reading ~w trace entries~s.~n",
	      [N,
	       case T of
		   undefined ->
		       "";
		   _ ->
		       io_lib:format(" spanning ~.3f ms", [T*0.001])
	       end]),
    end_of_trace(N, FirstTS, TS);
%% call
handler({trace_ts, Pid, call, {M, F, Arity} = Func, TS} = Trace, State)
  when integer(Arity) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_call(Pid, {Func, TS}),
    handler_fsm(State, TS);
handler({trace_ts, Pid, call, {M, F, Args} = MFArgs, TS} = Trace, State)
  when list(Args) ->
    ?dbg(1, "    ~p~n", [Trace]),
    Func = mfarity(MFArgs),
    trace_call(Pid, {Func, TS}),
    handler_fsm(State, TS);
handler({trace_ts, Pid, return_to, undefined, TS} = Trace, State) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_return_to(Pid, {undefined, TS}),
    handler_fsm(State, TS);
handler({trace_ts, Pid, return_to, {M, F, Arity} = Func, TS} = Trace, State)
  when integer(Arity) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_return_to(Pid, {Func, TS}),
    handler_fsm(State, TS);
handler({trace_ts, Pid, return_to, {M, F, Args} = MFArgs, TS} = Trace, State)
  when list(Args) ->
    ?dbg(1, "    ~p~n", [Trace]),
    Func = mfarity(MFArgs),
    trace_return_to(Pid, {Func, TS}),
    handler_fsm(State, TS);
handler({trace_ts, Parent, spawn, Pid, MFArgs, TS} = Trace, State) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_spawn(Pid, {MFArgs, TS}, Parent),
    handler_fsm(State, TS);
handler({trace_ts, Parent, spawn, Pid, TS} = Trace, State) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_spawn(Pid, {undefined, TS}, Parent),
    handler_fsm(State, TS);
handler({trace_ts, Pid, exit, _Reason, TS} = Trace, State) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_exit(Pid, TS),
    handler_fsm(State, TS);
handler({trace_ts, Pid, out, {M, F, Arity} = Func, TS} = Trace, State)
  when integer(Arity) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_out(Pid, {Func, TS}),
    handler_fsm(State, TS);
handler({trace_ts, Pid, out, {M, F, Args} = MFArgs, TS} = Trace, State)
  when list(Args) ->
    ?dbg(1, "    ~p~n", [Trace]),
    Func = mfarity(MFArgs),
    trace_out(Pid, {Func, TS}),
    handler_fsm(State, TS);
handler({trace_ts, Pid, in, {M, F, Arity} = Func, TS} = Trace, State)
  when integer(Arity) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_in(Pid, {Func, TS}),
    handler_fsm(State, TS);
handler({trace_ts, Pid, in, {M, F, Args} = MFArgs, TS} = Trace, State)
  when list(Args) ->
    ?dbg(1, "    ~p~n", [Trace]),
    Func = mfarity(MFArgs),
    trace_in(Pid, {Func, TS}),
    handler_fsm(State, TS);
handler({trace_ts, Pid, gc_start, _Info = Func, TS} = Trace, State) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_gc_start(Pid, TS),
    handler_fsm(State, TS);
handler({trace_ts, Pid, gc_end, _Info = Func, TS} = Trace, State) ->
    ?dbg(1, "    ~p~n", [Trace]),
    trace_gc_end(Pid, TS),
    handler_fsm(State, TS);
handler(Trace, State) ->
    ?dbg(1, "#   ~p~n", [Trace]),
    handler_fsm(State, undefined).



handler_fsm({M, undefined}, NewTS) ->
    put(first_ts, NewTS),
    handler_fsm({M, NewTS}, NewTS);
handler_fsm({M, TS}, NewTS) ->
    N = M+1,
    if (N rem 1000) == 0 ->
	    GroupLeader = get(group_leader),
	    io:put_chars(GroupLeader, "."),
	    if (N rem 100000) == 0 ->
		    io:put_chars(GroupLeader, "\n");
	       true -> ok
	    end;
       true ->
	    ok
    end,
    case NewTS of
	undefined ->
	    {N, TS};
	_ ->
	    {N, NewTS}
    end.



trace_call(Pid, {Func, TS} = Call) ->
    Stack = get(Pid),
    ?dbg(0, "trace_call(~p, ~p) ~p~n", [Pid, Call, Stack]),
    case Stack of
	undefined ->
	    Table = get(table),
	    init_log(Table, Pid, Func),
	    trace_clock(Table, Pid, 1, [Call], #clocks.cnt),
	    put(Pid, [Call]);
	[] ->
	    Table = get(table),
	    init_log(Table, Pid, Func),
	    trace_clock(Table, Pid, 1, [Call], #clocks.cnt),
	    put(Pid, [Call]);
	[{suspend, _TS0} | _] ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]});
	[{garbage_collect, _TS0} | _] ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]});
	[{Func, _TS0} | _] ->
	    ok;
	[_Call0 | _] ->
	    Table = get(table),
	    init_log(Table, Pid, Func),
	    trace_clock(Table, Pid, TS, Stack, #clocks.own),
	    NewStack = trace_call_1(Stack, Pid, Call, Stack, []),
	    trace_clock(Table, Pid, 1, [Call | NewStack], #clocks.cnt),
	    put(Pid, [Call | NewStack]);
	_ ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]})
    end,
    ok.

%% Search down the call stack to find a call to Func, 
%% save the calls passed.
trace_call_1([], _Pid, _Call, Stack, _R) ->
    Stack;
trace_call_1([{Func, _TS0} = Call0 | Stack1], Pid, {Func, TS}, Stack, R) ->
    trace_call_2(Stack1, reverse(R), Pid, Call0, Stack, R, Stack1, 0);
trace_call_1([Call0 | Stack1], Pid, Call, Stack, R) ->
    trace_call_1(Stack1, Pid, Call, Stack, [Call0 | R]).

%% Found a call to Func. Go further down the call stack as long as 
%% the saved calls from the stack top matches the calls below
%% Func. If Func is found again at the end of the saved call stack,
%% we have a call cycle.
trace_call_2([{Func, _TS0A} | RestA], [], 
	     Pid, {Func, TS} = Call, _Stack, R, Stack1, N) ->
    NewStack = reverse(R, RestA),
    trace_call_3(Stack1, N, get(table), Pid, TS, reverse([Call | R], RestA));
trace_call_2([{Func0, _TS0A} = Call0A | RestA], [{Func0, _TS0B} | RestB], 
	     Pid, Call, Stack, R, Stack1, N) ->
    trace_call_2(RestA, RestB, Pid, Call, Stack, R, Stack1, N+1);
trace_call_2(_StackA, _StackB,
	     _Pid, _Call, Stack, _R, _Stack1, _N) ->
    Stack.

%% Collaps the cycle in the call stack.
trace_call_3(_Stack1, 0, _Table, _Pid, _TS, Stack) ->
    Stack;
trace_call_3(Stack1, N, Table, Pid, TS, Stack) ->
    trace_clock(Table, Pid, TS, Stack1, #clocks.acc),
    trace_call_3(tl(Stack1), N-1, Table, Pid, TS, Stack).



trace_return_to(Pid, {Func, TS} = Call) ->
    Stack = get(Pid),
    ?dbg(0, "trace_return_to(~p, ~p) ~p~n", [Pid, Call, Stack]),
    case Stack of
	undefined ->
	    NewStack = [Call],
	    trace_clock(get(table), Pid, 1, NewStack, #clocks.cnt),
	    put(Pid, NewStack);
	[] ->
	    case Func of
		undefined ->
		    put(Pid, []);
		_ ->
		    NewStack = [Call],
		    trace_clock(get(table), Pid, 1, NewStack, #clocks.cnt),
		    put(Pid, NewStack)
	    end;
	[{suspend, _TS0} | _] ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]});
	[{garbage_collect, _TS0} | _] ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]});
	[{Func, TS0} | _] ->
	    ok;
	[{Func0, TS0} | _] = Stack ->
	    Table = get(table),
	    trace_clock(Table, Pid, TS, Stack, #clocks.own),
	    case Func of
		undefined ->
		    put(Pid, trace_return_to_1(Table, Pid, Call, Stack));
		_ ->
		    put(Pid, 
			[Call | trace_return_to_1(Table, Pid, Call, Stack)])
	    end;
	_ ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Pid, Call]})
    end,
    ok.

trace_return_to_1(_Table, _Pid, _Call, []) ->
    [];
trace_return_to_1(Table, Pid, {Func, TS}, [{Func, _TS0} | Rest] = Stack) ->
    trace_clock(Table, Pid, TS, Stack, #clocks.acc),
    Rest;
trace_return_to_1(Table, Pid, {_Func, TS} = Call, [Call0 | Rest] = Stack) ->
    trace_clock(Table, Pid, TS, Stack, #clocks.acc),
    trace_return_to_1(Table, Pid, Call, Rest).



trace_spawn(Pid, {undefined, TS} = Call, Parent) ->
    case get(Pid) of
	undefined ->
	    Table = get(table),
	    NewStack = [{suspend, TS}],
	    trace_clock(Table, Pid, 1, NewStack, #clocks.cnt),
	    put(Pid, NewStack),
	    ets:insert(Table, #proc{id = Pid, parent = Parent});
	Stack ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]})
    end;
trace_spawn(Pid, {MFArgs, TS} = Call, Parent) ->
    Func = mfarity(MFArgs),
    case get(Pid) of
	undefined ->
	    Table = get(table),
	    NewStack = [{Func, TS}],
	    trace_clock(Table, Pid, 1, NewStack, #clocks.cnt),
	    NewerStack = [{suspend, TS} | NewStack],
	    trace_clock(Table, Pid, 1, NewerStack, #clocks.cnt),
	    put(Pid, NewerStack),
	    ets:insert(Table, #proc{id = Pid, parent = Parent,
				    spawned_as = MFArgs});
	Stack ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]})
    end.



trace_exit(Pid, TS) ->
    Stack = get(Pid),
    case Stack of
	undefined ->
	    ok;
	[] ->
	    ok;
	[{garbage_collect, TS0} = Call0 
	 | [{suspend, TS1} | [_Call2 | _] = Stack2] = Stack1] ->
	    Table = get(table),
	    trace_clock(Table, Pid, TS, Stack1, #clocks.acc),
	    FixStack = [Call0 | Stack2],
	    trace_clock(Table, Pid, TS, FixStack, #clocks.own),
	    trace_exit_1(Table, Pid, TS, FixStack);
	[_Call0 | _] ->
	    Table = get(table),
	    trace_clock(Table, Pid, TS, Stack, #clocks.own),
	    trace_exit_1(Table, Pid, TS, Stack)
    end,
    put(Pid, []).

trace_exit_1(Table, Pid, TS, []) ->
    ok;
trace_exit_1(Table, Pid, TS, [_Call0 | Rest] = Stack) ->
    trace_clock(Table, Pid, TS, Stack, #clocks.acc),
    trace_exit_1(Table, Pid, TS, Rest).



trace_out(Pid, {_Func, TS} = Call) ->    
    Stack = get(Pid),
    ?dbg(0, "trace_out(~p, ~p) ~p~n", [Pid, Call, Stack]),
    case Stack of
	undefined ->
	    NewStack = [{suspend, TS}],
	    trace_clock(get(table), Pid, 1, NewStack, #clocks.cnt),
	    put(Pid, NewStack);
	[] ->
	    NewStack = [{suspend, TS}],
	    trace_clock(get(table), Pid, 1, NewStack, #clocks.cnt),
	    put(Pid, NewStack);
	[{suspend, _TS0} | _] ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]});
% 	[{Func, TS0} | _] ->
% 	    Table = get(table),
% 	    trace_clock(Table, Pid, TS, Stack, #clocks.own),
% 	    NewStack = [{suspend, TS} | Stack],
% 	    trace_clock(Table, Pid, 1, NewStack, #clocks.cnt),
% 	    put(Pid, NewStack);
% 	[{garbage_collect, TS0} | _] ->
% 	    Table = get(table),
% 	    trace_clock(Table, Pid, TS, Stack, #clocks.own),
% 	    NewStack = [{suspend, TS} | Stack],
% 	    trace_clock(Table, Pid, 1, NewStack, #clocks.cnt),
% 	    put(Pid, NewStack);
	[{Func0, TS0} = Call0 | _] ->
	    Table = get(table),
	    trace_clock(Table, Pid, TS, Stack, #clocks.own),
	    NewStack = [{suspend, TS} | Stack],
	    trace_clock(Table, Pid, 1, NewStack, #clocks.cnt),
	    put(Pid, NewStack);
	_ ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]})
    end.
    


trace_in(Pid, {_Func, TS} = Call) ->	    
    Stack = get(Pid),
    ?dbg(0, "trace_in(~p, ~p) ~p~n", [Pid, Call, Stack]),
    case Stack of
	undefined ->
	    put(Pid, []);
	[] ->
	    put(Pid, []);
% 	[{suspend, TS0} | [{garbage_collect, TS1} | Stack2] = Stack1] ->
% 	    Table = get(table),
% 	    trace_clock(Table, Pid, TS, Stack, #clocks.acc),
% 	    trace_clock(Table, Pid, TS, Stack1, #clocks.acc),
% 	    put(Pid, [{garbage_collect, TS} | Stack2]);
	[{suspend, TS0}] ->
	    trace_clock(get(table), Pid, TS, Stack, #clocks.acc),
	    put(Pid, []);
	[{suspend, TS0} | [{Func1, _TS1} | Stack2] = Stack1] ->
	    Table = get(table),
	    trace_clock(Table, Pid, TS, Stack, #clocks.acc),
	    trace_clock(Table, Pid, TS, Stack1, #clocks.acc),
	    put(Pid, [{Func1, TS} | Stack2]);
	_ ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, Call]})
    end.



trace_gc_start(Pid, TS) ->    
    Stack = get(Pid),
    ?dbg(0, "trace_gc_start(~p, ~p) ~p~n", [Pid, TS, Stack]),
    case Stack of
	undefined ->
	    NewStack = [{garbage_collect, TS}],
	    trace_clock(get(table), Pid, 1, NewStack, #clocks.cnt),
	    put(Pid, NewStack);
	[] ->
	    NewStack = [{garbage_collect, TS}],
	    trace_clock(get(table), Pid, 1, NewStack, #clocks.cnt),
	    put(Pid, NewStack);
	[{suspend, TS0} | [_Call1 | _] = Stack1] ->
	    Call = {garbage_collect, TS},
	    trace_clock(get(table), Pid, 1, [Call | Stack1], #clocks.cnt),
	    put(Pid, [Call | Stack]);
	[{Func0, TS0} | _] ->
	    Table = get(table),
	    trace_clock(Table, Pid, TS, Stack, #clocks.own),
	    NewStack = [{garbage_collect, TS} | Stack],
	    trace_clock(Table, Pid, 1, NewStack, #clocks.cnt),
	    put(Pid, NewStack);
	_ ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, TS]})
    end.



trace_gc_end(Pid, TS) ->	    
    Stack = get(Pid),
    ?dbg(0, "trace_gc_end(~p, ~p) ~p~n", [Pid, TS, Stack]),
    case Stack of
	undefined ->
	    put(Pid, []);
	[] ->
	    put(Pid, []);
	[{garbage_collect, TS0}] ->
	    Table = get(table),
	    trace_clock(Table, Pid, TS, Stack, #clocks.own),
	    trace_clock(Table, Pid, TS, Stack, #clocks.acc),
	    put(Pid, []);
	[{garbage_collect, TS0} = Call0 | [{suspend, TS1} 
					   | Stack2] = Stack1] ->
	    Table = get(table),
	    trace_clock(Table, Pid, TS, [Call0 | Stack2], #clocks.own),
	    trace_clock(Table, Pid, TS, [Call0 | Stack2], #clocks.acc),
	    trace_clock(Table, Pid, TS, Stack1, #clocks.acc),
	    put(Pid, [{suspend, TS} | Stack2]);
	[{garbage_collect, TS0} | [{Func1, TS1} | Stack2] = Stack1] ->
	    Table = get(table),
	    trace_clock(Table, Pid, TS, Stack, #clocks.own),
	    trace_clock(Table, Pid, TS, Stack, #clocks.acc),
	    trace_clock(Table, Pid, TS, Stack1, #clocks.acc),
	    put(Pid, [{Func1, TS} | Stack2]);
	_ ->
	    exit({inconsistent_trace_data, ?MODULE, ?LINE,
		  [Stack, Pid, TS]})
    end.



end_of_trace(_N, FirstTS, LastTS) ->
%    exit(done),
    %%
    %% Close all process stacks, as if the processes exited.
    %%
    Procs = get(),
    ?dbg(2, "get() -> ~p~n", [Procs]),
    lists:map(
      fun ({Pid, _}) when pid(Pid) ->
	      trace_exit(Pid, LastTS);
	  (_) ->
	      ok
      end,
      Procs),
    %%
    %% Clean out the process dictionary before the next step
    %%
    Table = erase(table),
    Parent = erase(parent),
    Flags = erase(flags),
    GroupLeader = erase(group_leader),
    Erase = erase(),
    ?dbg(2, "erase() -> ~p~n", [Erase]),
    #opts{out = Io} = scan_flags(Flags),
    %%
    %% Process the collected data and spread it to 3 places:
    %% * Per {process, caller, func}. Stored in the process dictionary.
    %% * Sum per process. Stored in an ets table.
    %% * Extra info per process. Stored in another ets table.
    %%
    io:format(GroupLeader, "Processing data...~n", []),
    Data = ets:tab2list(Table),
    ?dbg(2, "Data = ~p~n", [Data]),
    PidTable = ets:new(?MODULE, [set, private, {keypos, #clocks.id}]),
    ProcTable = ets:new(?MODULE, [set, private, {keypos, #proc.id}]),
    lists:foreach(
      fun (#clocks{id = {Pid, Caller, Func}} = Clocks) ->
	      funcstat_pd_pid_func(Pid, Caller, Func, Clocks),
	      funcstat_pd_pid_caller(Pid, Caller, Func, Clocks),
	      clocks_add(PidTable, Clocks#clocks{id = Pid});
	  (#proc{} = Proc) ->
	      ets:insert(ProcTable, Proc)
      end,
      Data),
    ?dbg(3, "get() -> ~p~n", [get()]),
    PidList = ets:tab2list(PidTable),
    ProcList = ets:tab2list(ProcTable),
    ?dbg(3, "PidList =  ~p~n", [PidList]),
    ?dbg(3, "ProcList =  ~p~n", [ProcList]),
    %%
    %% Reorganize the process dictionary by Pid.
    %%
    lists:foreach(
      fun ({{Pid, _Func}, Funcstat}) ->
	      put(Pid, [Funcstat | case get(Pid) of
				       undefined -> [];
				       Other -> Other
				   end])
      end,
      erase()),
    ?dbg(4, "get() -> ~p~n", [get()]),
    %%
    %% Sort the processes.
    %%
    io:format(GroupLeader, "Calculating totals...~n", []),
    PidSorted = clocks_sort_ownR(PidList),
    ?dbg(4, "PidSorted = ~p~n", [PidSorted]),
    %%
    %% Calculate totals
    %%
    Totals0 = lists:foldl(fun(A, B) -> clocks_sum(A, B, totals) end,
			 #clocks{id = totals},
			 PidList),
    Totals = Totals0#clocks{acc = ts_sub(LastTS, FirstTS)},
    ?dbg(4, "Totals = ~p~n", [Totals]),
    %%
    %% Print the functions per process
    %%
    io:format(GroupLeader, "Creating output...~n", []),
    io:format(Io, "~nDetailed statistics:~n", []),
    print_head(Io),
    print_totals(Io, Totals),
    lists:foreach(
      fun ({#clocks{} = Clocks, ProcOrPid, FuncstatList}) ->
	      print_head(Io),
	      print_pid(Io, Clocks),
	      print_proc(Io, ProcOrPid),
	      lists:foreach(
		fun (#funcstat{callers_sum = CallersSum, 
%			       called_sum = CalledSum, 
			       callers = Callers, 
			       called = Called}) ->
			lists:foreach(
			  fun(C) ->
				  print_caller(Io, C)
			  end,
			  Callers),
			print_func(Io, CallersSum),
%			print_func(Io, CalledSum),
			lists:foreach(
			  fun(C) ->
				  print_called(Io, C)
			  end,
			  Called),
			print_nl(Io),
			ok
		end,
		%% Sort the functions within the process, 
		%% and the callers and called within the function.
		funcstat_sort_accR(FuncstatList, fun clocks_sort_accR/1))
      end,
      %% Look up the processes in sorted order
      lists:map(
	fun (#clocks{id = Pid} = Clocks) -> 
		Proc = case ets:lookup(ProcTable, Pid) of
			   [] -> Pid;
			   [ProcX] -> ProcX
		       end,
		FuncstatList = get(Pid),
		{Clocks, Proc, FuncstatList}
	end, 
	PidSorted)),
%     %%
%     %% Sort and print the functions per process
%     %%
%     io:format(GroupLeader, "Creating output...~n", []),
%     Io = group_leader(),
%     io:format(Io, "~nDetailed statistics:~n", []),
%     lists:foreach(
%       fun (Times) ->
% 	      print_head(Io),
% 	      Pid = times_id(Times),
% 	      print_entry(Io, Times, Totals),
% 	      case ets:lookup(ProcTable, Pid) of
% 		  [#proc{} = Proc] ->
% 		      print_entry(Io, Proc);
% 		  [] ->
% 		      ok
% 	      end,
% 	      lists:foreach(
% 		fun(Times1) ->
% 			print_entry(Io, Times1, Totals)
% 		end,
% 		lists:reverse(
% 		  lists:sort(
% 		    lists:map(fun clocks_pid_mfarity_presort/1,
% 			      getl(Pid)))))
%       end,
%       PidSorted),
%     %%
%     %% Sort and print the per process data
%     %%
%     FuncList = ets:tab2list(FuncTable),
%     ets:delete(FuncTable),
%     ?dbg(2, "FuncList = ~p~n", [FuncList]),
%     io:format(Io, "~n** Processes summarized~n", []),
%     print_head(Io),
%     lists:foreach(
%       fun (Times) ->
% 	      Func = times_id(Times),
% 	      print_entry(Io, Times, Totals)
%       end,
%       lists:reverse(
% 	lists:sort(
% 	  lists:map(fun clocks_mfarity_presort/1,
% 		    FuncList)))),
%     %%
%     %% Print the total totals
%     %%
%     print_head(Io),
%     print_entry(Io, Totals, Totals),
%     %%
%     %% Finish
%     %%
    ets:delete(PidTable),
    ets:delete(ProcTable),
    ets:delete(Table),
    io:format(GroupLeader, "Done!~n", []),
    Parent ! {self(), ok},
    ok.



scan_flags(Flags) ->
    scan_flags(Flags, #opts{}).

scan_flags([], #opts{out = Out}) ->
    #opts{out = case Out of
		    [] ->
			group_leader();
		    [Fd] ->
			Fd
		end};
scan_flags([outfile | Flags], #opts{} = Opts) ->
    scan_flags([{outfile, "fprof.txt"} | Flags], Opts);
scan_flags([{outfile, FName} | Flags], #opts{out = []} = Opts) ->
    case file:open(FName, [write]) of
	{ok, Fd} ->
	    scan_flags(Flags, Opts#opts{out = [Fd]});
	{error, Reason} ->
	    exit({could_not_open_file, ?MODULE, ?LINE, FName})
    end;
scan_flags([{out, Fd} | Flags], #opts{out = []} = Opts) ->
    scan_flags(Flags, Opts#opts{out = [Fd]}).



print_head(Io) ->
    io:format(Io, "~s~s~s~s  ~n~n",
	      [pad(" ", $ , ?FNAME_WIDTH),
	       pad($ , " CALLS", ?NR_WIDTH),
	       pad($ , " ACC", ?NR_WIDTH), 
	       pad($ , " OWN",  ?NR_WIDTH)]).

print_totals(Io, #clocks{cnt = Cnt, own = Own, acc = Acc}) ->
    io:format(Io, "~s~s~s~s  ~n~n",
	      [pad(flat_format("** Totals ", []), 
		   $ , ?FNAME_WIDTH),
	       pad($ , flat_format(" ~w",   [Cnt]),       ?NR_WIDTH),
	       pad($ , flat_format(" ~.3f", [Acc*0.001]), ?NR_WIDTH),
	       pad($ , flat_format(" ~.3f", [Own*0.001]), ?NR_WIDTH)]).

print_pid(Io, #clocks{id = Pid, cnt = Cnt, own = Own}) ->
    io:format(Io, "~s~s~s~s  ~n",
	      [pad(flat_format("** Process ~w ", [Pid]), 
		   $ , ?FNAME_WIDTH),
	       pad($ , flat_format(" ~w",   [Cnt]),       ?NR_WIDTH),
	       pad($ , " ",                               ?NR_WIDTH),
	       pad($ , flat_format(" ~.3f", [Own*0.001]), ?NR_WIDTH)]).

print_caller(Io, Clocks) ->
    print_func(Io, Clocks, "   ~s ", "  ").

print_called(Io, Clocks) ->
    print_func(Io, Clocks, "   ~s ", "  ").

print_func(Io, Clocks) ->
    print_func(Io, Clocks, "*  ~s ", " *").

print_func(Io, #clocks{id = Func, cnt = Cnt, own = Own, acc = Acc}, 
	   Format, Tail) ->
    io:format(Io, "~s~s~s~s~s~n",
	      [pad(flat_format(Format,
			       [format_func(Func)]), 
		   $ , ?FNAME_WIDTH),
	       pad($ , flat_format(" ~w",   [Cnt]),       ?NR_WIDTH),
	       pad($ , flat_format(" ~.3f", [Acc*0.001]), ?NR_WIDTH),
	       pad($ , flat_format(" ~.3f", [Own*0.001]), ?NR_WIDTH),
	       Tail]).

print_proc(Io, #proc{id = Pid, parent = undefined, init_log = []}) ->
    print_nl(Io);
print_proc(Io, #proc{id = Pid, 
		     parent = Parent, 
		     spawned_as = SpawnedAs,
		     init_log = InitLog}) ->
    io:put_chars(Io, 
		 ["** ",
		  case Parent of 
		      undefined ->
			  [];
		      _ ->
			  io_lib:format("Spawned by ~w~s.~n", 
					[Parent,
					 case SpawnedAs of
					     undefined ->
						 "";
					     _ ->
						 [" as ",
						  format_func(SpawnedAs)]
					 end])
		  end,
		  case InitLog of
		      [] ->
			  io_lib:nl();
		      _ ->
			  ["** Initial calls: " | io_lib:nl()]
		  end]),
    print_init_log(Io, InitLog),
    print_nl(Io);
print_proc(Io, Pid) ->
    print_nl(Io).

print_init_log(_Io, []) ->
    ok;
print_init_log(Io, [Func | Tail]) ->
    print_init_log(Io, Tail),
    io:format(Io, "**    ~s~n", [format_func(Func)]).

print_nl(Io) ->
    io:nl(Io).



funcstat_pd_pid_func(Pid, Caller, Func, Clocks) ->
    put({Pid, Func},
	case get({Pid, Func}) of
	    undefined ->
		#funcstat{callers_sum = Clocks#clocks{id = Func}, 
			  called_sum = #clocks{id = Func},
			  callers = [Clocks#clocks{id = Caller}]};
	    #funcstat{callers_sum = CallersSum,
		      callers = Callers} = Funcstat ->
		Funcstat#funcstat{
		  callers_sum = clocks_sum(CallersSum, Clocks, Func),
		  callers = [Clocks#clocks{id = Caller} | Callers]}
	end).

funcstat_pd_pid_caller(Pid, Caller, Func, Clocks) ->
    put({Pid, Caller},
	case get({Pid, Caller}) of
	    undefined ->
		#funcstat{callers_sum = #clocks{id = Caller}, 
			  called_sum = Clocks#clocks{id = Caller},
			  called = [Clocks#clocks{id = Func}]};
	    #funcstat{called_sum = CalledSum,
		      called = Called} = Funcstat ->
		Funcstat#funcstat{
		  called_sum = clocks_sum(CalledSum, Clocks, Caller),
		  called = [Clocks#clocks{id = Func} | Called]}
	end).



format_func({M, F, Arity}) when integer(Arity) ->
    io_lib:format("~w:~w/~w", [M, F, Arity]);
format_func({M, F, Args}) when list(Args) ->
    io_lib:format("~w:~w(~s)", 
		  [M, F, format_func_args(Args)]);
format_func(Func) ->
    io_lib:format("~w", [Func]).

format_func_args([A | [_|_] = T]) ->
    [io_lib:format("~p,", [A]) | format_func_args(T)];
format_func_args([A]) ->
    io_lib:format("~p", [A]);
format_func_args([]) ->
    [].



funcstat_sort_accR(FuncstatList, ClocksSort) ->
    funcstat_sort_accR_1(FuncstatList, ClocksSort, []).

funcstat_sort_accR_1([], _ClocksSort, R) ->
    postsortR(sort(R), []);
funcstat_sort_accR_1([#funcstat{callers_sum = #clocks{acc = K},
				callers = Callers,
				called = Called} = Funcstat
		      | L], 
		     ClocksSort,
		     R) ->
    funcstat_sort_accR_1(L, 
			 ClocksSort,
			 [[K | Funcstat#funcstat{callers = ClocksSort(Callers),
						 called = ClocksSort(Called)}]
			  | R]).



clocks_sort_accR(ClocksList) ->
    clocks_sort_accR_1(ClocksList, []).

clocks_sort_accR_1([], R) ->
    postsortR(sort(R), []);
clocks_sort_accR_1([#clocks{acc = K} = C | L], R) ->
    clocks_sort_accR_1(L, [[K | C] | R]).



clocks_sort_ownR(ClocksList) ->
    clocks_sort_ownR_1(ClocksList, []).

clocks_sort_ownR_1([], R) ->
    postsortR(sort(R), []);
clocks_sort_ownR_1([#clocks{own = K} = C | L], R) ->
    clocks_sort_ownR_1(L, [[K | C] | R]).



postsortR([], R) ->
    R;
postsortR([[_ | C] | L], R) ->
    postsortR(L, [C | R]).



init_log(_Table, _Id, suspend) ->
    ok;
init_log(_Table, _Id, void) ->
    ok;
init_log(Table, Id, Entry) ->
    case ets:lookup(Table, Id) of
	[#proc{init_cnt = 0}] ->
	    ok;
	[#proc{init_cnt = N, init_log = L} = Proc] ->
	    ets:insert(Table, 
		       Proc#proc{init_cnt = N-1, init_log = [Entry | L]});
	[] ->
	    ok
    end.



trace_clock(Table, Pid, TS, [{_Func0, TS0} = Call0], Clock) ->
    trace_clock(Table, Pid, TS, [Call0, {undefined, TS0}], Clock);
trace_clock(Table, Pid, T, [{Func0, TS0}, {Func1, TS1} | _], Clock)
  when integer(T) ->
    clock_add(Table, {Pid, Func1, Func0}, Clock, T);
trace_clock(Table, Pid, TS, [{Func0, TS0}, {Func1, TS1} | _], Clock) ->
    clock_add(Table, {Pid, Func1, Func0}, Clock, ts_sub(TS, TS0)).

clock_add(Table, Id, Clock, T) ->
    ?dbg(0, "        clock_add(Table, ~w, ~w, ~w)~n", [Id, Clock, T]),
    case (catch ets:update_counter(Table, Id, {Clock, T})) of
	{'EXIT', _} ->
	    ets:insert(Table, #clocks{id = Id}),
	    case ets:update_counter(Table, Id, {Clock, T}) of
		X when X >= 0 -> ok;
		X -> ?dbg(0, "Negative counter value ~p ~p ~p ~p~n",
			  [X, Id, Clock, T])
	    end;
	V ->
	    V
    end.

clocks_add(Table, #clocks{id = Id} = Clocks) ->
    case ets:lookup(Table, Id) of
	[Clocks0] ->
	    ets:insert(Table, clocks_sum(Clocks, Clocks0, Id));
	[] ->
	    ets:insert(Table, Clocks)
    end.



clocks_sum(#clocks{id = _Id1, 
		   cnt = Cnt1, 
		   own = Own1, 
		   acc = Acc1}, 
	   #clocks{id = _Id2, 
		   cnt = Cnt2, 
		   own = Own2, 
		   acc = Acc2}, 
	   Id) ->
    #clocks{id = Id,
	    cnt = Cnt1 + Cnt2,
	    own = Own1 + Own2,
	    acc = Acc1 + Acc2}.



ts_sub({A, B, C} = T, {A0, B0, C0} = T0) ->
    X = ((((A-A0)*1000000) + (B-B0))*1000000) + C - C0,
    if X >= 0 -> ok;
       true -> ?dbg(9, "Negative counter value ~p ~p ~p~n",
		    [X, T, T0])
    end,
    X;
ts_sub(_, _) ->
    undefined.



flat_format(Format, Args) ->
    flatten(io_lib:format(Format, Args)).

% sort_map(Fun, List) ->
%     sort(reverse_map(Fun, List)).

% reverse_sort_map(Fun, List) ->
%     reverse(sort(reverse_map(Fun, List))).

% reverse_sort(List) ->
%     reverse(sort(List)).

% map(Fun, List) ->
%     reverse(reverse_map(Fun, List)).

% reverse_map(Fun, List) ->
%     reverse_map(Fun, List, []).

% reverse_map(_Fun, [], R) ->
%     R;
% reverse_map(Fun, [X | L], R) ->
%     reverse_map(Fun, L, [Fun(X) | R]).



mfarity({M, F, Args}) when list(Args) ->
    {M, F, length(Args)};
mfarity(MFA) ->
    MFA.



pad(Char, L, Size) when integer(Char), list(L), integer(Size) ->
    List = lists:flatten(L),
    Length = length(List),
    if Length >= Size ->
	    List;
       true ->
	    lists:append(lists:duplicate(Size - Length, Char), List)
    end;
pad(L, Char, Size) when list(L), integer(Char), integer(Size) ->
    List = lists:flatten(L),
    Length = length(List),
    if Length >= Size ->
	    List;
       true ->
	    lists:append(List, lists:duplicate(Size - Length, Char))
    end.



burn() ->
    Self = self(),
    Pid = spawn_link(fun () -> burn(Self) end),
    receive Pid -> ok end.

burn(Parent) ->
    Self = self(),
    Pid = spawn_link(fun() -> foo(Self, {self(), Parent}) end),
    receive
	{Pid, Parent} ->
	    Parent ! Self;
	_ ->
	    failed
    end.

foo(Parent, Msg) ->    
    receive after 500 -> ok end,
    Parent ! Msg.
